package org.xbib.interlibrary.catalog.elasticsearch;

import java.io.IOException;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.elasticsearch.action.index.IndexRequest;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.elx.api.SearchClient;
import org.xbib.interlibrary.catalog.builder.Converter;
import org.xbib.interlibrary.catalog.builder.ConverterWorker;
import org.xbib.interlibrary.catalog.builder.ConverterWorkerPool;
import org.xbib.settings.Settings;

public abstract class ElasticsearchTool<R> extends Converter<R, ElasticsearchTool.ElasticsearchWorkerPool<R>> {

    @Override
    public abstract ElasticsearchWorkerPool<R> newPool(Settings settings) throws IOException;

    public static abstract class ElasticsearchWorkerPool<R> extends ConverterWorkerPool<R> {

        private final Logger logger = Logger.getLogger(ElasticsearchWorkerPool.class.getName());

        private final ElxInputHelper elasticsearchInput;

        private final ElxOutputHelper elasticsearchOutput;

        private final IRINamespaceContext namespaceContext;

        public ElasticsearchWorkerPool(String poolName, Settings settings) throws IOException {
            super(poolName, settings);
            Settings inputElasticsearchSettings = settings.getGroups("input").get("elasticsearch");
            boolean isInput = inputElasticsearchSettings != null && !inputElasticsearchSettings.getAsMap().isEmpty();
            Settings outputElasticsearchSettings = settings.getGroups("output").get("elasticsearch");
            boolean isOutput = outputElasticsearchSettings != null && !outputElasticsearchSettings.getAsMap().isEmpty();
            this.elasticsearchInput = isInput ? new ElxInputHelper(inputElasticsearchSettings) : null;
            this.elasticsearchOutput = isOutput ? new ElxOutputHelper(outputElasticsearchSettings) : null;
            this.namespaceContext = IRINamespaceContext.newInstance(ElasticsearchTool.class);
        }

        @Override
        protected void openInput() throws IOException, InterruptedException {
            if (elasticsearchInput != null) {
                Settings inputElasticsearch = settings.getGroups("input").get("elasticsearch");
                if (inputElasticsearch != null) {
                    elasticsearchInput.setIndexDefinitions(inputElasticsearch.getGroups("index"));
                    logger.log(Level.INFO, "input definitions set up: " + elasticsearchInput.getIndexDefinitions());
                } else {
                    logger.log(Level.WARNING, "no input definitions?");
                }
            } else {
                logger.log(Level.WARNING, "no elasticsearch input?");
            }
            super.openInput();
        }

        @Override
        protected void openOutput() throws IOException {
            if (elasticsearchOutput != null) {
                Settings outputElasticsearch = settings.getGroups("output").get("elasticsearch");
                if (outputElasticsearch != null) {
                    elasticsearchOutput.setIndexDefinitions(outputElasticsearch.getGroups("index"));
                    elasticsearchOutput.createIndices();
                    elasticsearchOutput.startBulk();
                } else {
                    logger.log(Level.WARNING, "no output definitions");
                }
            } else {
                logger.log(Level.WARNING, "no elasticsearch output");
            }
            super.openOutput();
        }

        @Override
        protected void closeInput(int returncode) throws IOException {
            try {
                if (elasticsearchInput != null) {
                    elasticsearchInput.close();
                }
            } finally {
                super.closeInput(returncode);
            }
        }

        @Override
        protected void closeOutput(int returncode) throws IOException {
            try {
                if (elasticsearchOutput != null) {
                    elasticsearchOutput.stopBulk();
                    if (returncode == 0) {
                        elasticsearchOutput.activateIndex();
                    }
                    elasticsearchOutput.close();
                }
            } finally {
                super.closeOutput(returncode);
            }
        }

        @Override
        protected void closeMetrics() throws IOException {
            super.closeMetrics();
        }

        public Map<String, ElasticsearchIndexDefinition> getInputIndexDefinitionMap() {
            if (elasticsearchInput != null) {
                return elasticsearchInput.getIndexDefinitions();
            }
            throw new IllegalStateException("no elasticsearch input found");
        }

        public Map<String, ElasticsearchIndexDefinition> getOutputIndexDefinitionMap() {
            if (elasticsearchOutput != null) {
                return elasticsearchOutput.getIndexDefinitions();
            }
            throw new IllegalStateException("no elasticsearch output found");
        }

        public SearchClient searchClient() {
            if (elasticsearchInput != null) {
                return elasticsearchInput.getSearchClient();
            }
            throw new IllegalStateException("no elasticsearch input found");
        }

        public void bulkIndex(IndexDefinition indexDefinition, String id, boolean create, String content) {
            elasticsearchOutput.index(indexDefinition, id, create, content);
        }

        public void bulkIndex(IndexRequest indexRequest) {
            elasticsearchOutput.index(indexRequest);
        }

        public void delete(IndexDefinition indexDefinition, String id) {
            elasticsearchOutput.delete(indexDefinition, id);
        }

        public IRINamespaceContext getNamespaceContext() {
            return namespaceContext;
        }
    }

    public static abstract class ElasticsearchWorker<R, P extends ElasticsearchWorkerPool<R>> extends ConverterWorker<R> {

        protected ElasticsearchWorker(Settings settings) {
            super(settings);
        }

        @Override
        protected abstract void doExecute(R request) throws IOException;
    }
}
