package org.xbib.interlibrary.catalog.elasticsearch;

import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.function.Predicate;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.elx.api.AdminClient;
import org.xbib.elx.api.SearchClient;
import org.xbib.elx.common.ClientBuilder;
import org.xbib.elx.common.MockAdminClientProvider;
import org.xbib.elx.common.MockSearchClientProvider;
import org.xbib.elx.transport.TransportAdminClientProvider;
import org.xbib.elx.transport.TransportSearchClientProvider;
import org.xbib.settings.Settings;
import org.xbib.settings.SettingsBuilder;

public class ElxInputHelper implements AutoCloseable {

    private static final Logger logger = Logger.getLogger(ElxInputHelper.class.getName());

    private final AdminClient adminClient;

    private final SearchClient searchClient;

    private Map<String, ElasticsearchIndexDefinition> indexDefinitionMap;

    public ElxInputHelper(Settings settings) throws IOException {
        logger.log(Level.FINE, "creating admin client, settings = " + settings.getAsMap());
        this.adminClient = createAdminClient(settings);
        logger.log(Level.FINE, "creating search client, settings = " + settings.getAsMap());
        this.searchClient = createSearchClient(settings);
    }

    @Override
    public void close() throws IOException {
        logger.log(Level.INFO, "closing elx input helper");
        if (adminClient != null) {
            logger.log(Level.INFO, "closing input admin client");
            adminClient.close();
        }
        if (searchClient != null) {
            logger.log(Level.INFO, "closing input search client");
            searchClient.close();
        }
        logger.log(Level.INFO, "elx helper closed");
    }

    public Map<String, ElasticsearchIndexDefinition> getIndexDefinitions() {
        return indexDefinitionMap;
    }

    public SearchClient getSearchClient() {
        return searchClient;
    }

    public void setIndexDefinitions(Map<String, Settings> indexDefinitions) throws IOException {
        this.indexDefinitionMap = new LinkedHashMap<>();
        for (Map.Entry<String, Settings> entry : indexDefinitions.entrySet()) {
            Settings settings = entry.getValue();
            String name = settings.get("name");
            ElasticsearchIndexDefinition def = new ElasticsearchIndexDefinition(adminClient,
                    name, settings.get("type", name),
                    createSettingsFrom(settings),
                    getClass().getClassLoader());
            indexDefinitionMap.put(entry.getKey(), def);
        }
    }

    private static org.elasticsearch.common.settings.Settings createSettingsFrom(Settings settings) {
        Map<String, String> map = settings.getAsMap();
        return org.elasticsearch.common.settings.Settings.builder()
                .putProperties(map, k -> k)
                .build();
    }

    private static AdminClient createAdminClient(Settings settings) throws IOException {
        ClientBuilder clientBuilder = ClientBuilder.builder();
        if (!settings.getAsBoolean("enabled", true)) {
            clientBuilder.setAdminClientProvider(MockAdminClientProvider.class);
        } else {
            clientBuilder.setAdminClientProvider(TransportAdminClientProvider.class);
        }
        return clientBuilder.put(filter(settings, s -> isSettingsDisallowed(s)).getAsMap()).build();
    }

    private static SearchClient createSearchClient(Settings settings) throws IOException {
        ClientBuilder clientBuilder = ClientBuilder.builder();
        if (!settings.getAsBoolean("enabled", true)) {
            clientBuilder.setSearchClientProvider(MockSearchClientProvider.class);
        } else {
            clientBuilder.setSearchClientProvider(TransportSearchClientProvider.class);
        }
        clientBuilder.put(org.xbib.elx.common.Parameters.SEARCH_METRIC_ENABLED.getName(), Boolean.TRUE);
        return clientBuilder.put(filter(settings, ElxInputHelper::isSettingsDisallowed).getAsMap()).build();
    }

    private static Settings filter(Settings settings, Predicate<String> predicate) {
        SettingsBuilder builder = Settings.settingsBuilder();
        for (Map.Entry<String, String> entry : settings.getAsMap().entrySet()) {
            if (predicate.test(entry.getKey())) {
                builder.put(entry.getKey(), entry.getValue());
            }
        }
        return builder.build();
    }

    private static boolean isSettingsDisallowed(String key) {
        if (key.startsWith("index")) {
            return false;
        }
        for (Parameters p : Parameters.values()) {
            if (key.equals(p.getName())) {
                return false;
            }
        }
        return true;
    }
}
