package org.xbib.interlibrary.catalog.elasticsearch;

import java.io.IOException;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.function.Predicate;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.elasticsearch.action.index.IndexRequest;
import org.xbib.datastructures.api.TimeValue;
import org.xbib.elx.api.AdminClient;
import org.xbib.elx.api.BulkClient;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.elx.common.ClientBuilder;
import org.xbib.elx.common.MockAdminClientProvider;
import org.xbib.elx.common.MockBulkClientProvider;
import org.xbib.elx.transport.TransportAdminClientProvider;
import org.xbib.elx.transport.TransportBulkClientProvider;
import org.xbib.settings.Settings;
import org.xbib.settings.SettingsBuilder;

public class ElxOutputHelper implements AutoCloseable {

    private static final Logger logger = Logger.getLogger(ElxOutputHelper.class.getName());

    private final Settings settings;

    private final AdminClient adminClient;

    private final BulkClient bulkClient;

    private Map<String, ElasticsearchIndexDefinition> indexDefinitionMap;

    public ElxOutputHelper(Settings settings) throws IOException {
        this.settings = settings;
        logger.log(Level.FINE, "creating admin client, settings = " + settings.getAsMap());
        this.adminClient = createAdminClient(settings);
        logger.log(Level.FINE, "creating bulk client, settings = " + settings.getAsMap());
        this.bulkClient = createBulkClient(settings);
    }

    @Override
    public void close() throws IOException {
        logger.log(Level.INFO, "closing elx helper");
        if (adminClient != null) {
            // default purge time = 7 days
            TimeValue purgeTime = settings.getAsTime("purgetime", TimeValue.timeValueHours(24 * 7));
            logger.log(Level.INFO, "purging all closed indices, purgetime in days = " + purgeTime.days());
            adminClient.purgeAllClosedIndicesOlderThan(Instant.now().minus(purgeTime.days(), ChronoUnit.DAYS));
            logger.log(Level.INFO, "closing admin client");
            adminClient.close();
        }
        if (bulkClient != null) {
            logger.log(Level.INFO, "closing bulk client");
            bulkClient.close();
        }
        logger.log(Level.INFO, "elx helper closed");
    }

    public void setIndexDefinitions(Map<String, Settings> indexDefinitions) throws IOException {
        this.indexDefinitionMap = new LinkedHashMap<>();
        for (Map.Entry<String, Settings> entry : indexDefinitions.entrySet()) {
            Settings settings = entry.getValue();
            String name = settings.get("name");
            ElasticsearchIndexDefinition def = new ElasticsearchIndexDefinition(adminClient,
                    name, settings.get("type", name),
                    createSettingsFrom(settings),
                    getClass().getClassLoader());
            indexDefinitionMap.put(entry.getKey(), def);
        }
    }

    private static org.elasticsearch.common.settings.Settings createSettingsFrom(Settings settings) {
        return org.elasticsearch.common.settings.Settings.builder()
                .putProperties(settings.getAsMap(), k -> k)
                .build();
    }

    public Map<String, ElasticsearchIndexDefinition> getIndexDefinitions() {
        return indexDefinitionMap;
    }

    public void index(IndexDefinition indexDefinition, String id, boolean create, String content) {
        if (indexDefinition.isEnabled()) {
            bulkClient.index(indexDefinition, id, create, content);
        } else {
            logger.log(Level.INFO, "indexing " + id + " content = " + content);
        }
    }

    public void index(IndexRequest indexRequest) {
        bulkClient.index(indexRequest);
    }

    public void delete(IndexDefinition indexDefinition, String id) {
        bulkClient.delete(indexDefinition, id);
    }

    public void createIndices() throws IOException {
        for (Map.Entry<String, ElasticsearchIndexDefinition> entry : indexDefinitionMap.entrySet()) {
            ElasticsearchIndexDefinition def = entry.getValue();
            try {
                bulkClient.newIndex(def);
            } catch (Exception e) {
                if (def.isShiftEnabled()) {
                    throw new IOException(e);
                } else {
                    logger.log(Level.WARNING, "ignoring index creation error for index " + def.getFullIndexName() + " because index is not in shift mode");
                }
            }
        }
    }

    public void createIndices(String... defs) throws IOException {
        for (Map.Entry<String, ElasticsearchIndexDefinition> entry : indexDefinitionMap.entrySet()) {
            for (String def : defs) {
                if (def.equals(entry.getKey())) {
                    ElasticsearchIndexDefinition elasticsearchIndexDefinition = entry.getValue();
                    try {
                        bulkClient.newIndex(elasticsearchIndexDefinition);
                    } catch (Exception e) {
                        if (elasticsearchIndexDefinition.isShiftEnabled()) {
                            throw new IOException(e);
                        } else {
                            logger.log(Level.WARNING, "ignoring index creation error for index " + elasticsearchIndexDefinition.getFullIndexName() + " because index is not in shift mode");
                        }
                    }
                }
            }
        }
    }

    public void startBulk() {
        for (Map.Entry<String, ElasticsearchIndexDefinition> entry : indexDefinitionMap.entrySet()) {
            ElasticsearchIndexDefinition def = entry.getValue();
            bulkClient.startBulk(def);
        }
    }

    public void startBulk(String... defs) {
        for (Map.Entry<String, ElasticsearchIndexDefinition> entry : indexDefinitionMap.entrySet()) {
            for (String def : defs) {
                if (def.equals(entry.getKey())) {
                    ElasticsearchIndexDefinition elasticsearchIndexDefinition = entry.getValue();
                    bulkClient.startBulk(elasticsearchIndexDefinition);
                }
            }
        }
    }

    public void stopBulk() {
        for (Map.Entry<String, ElasticsearchIndexDefinition> entry : indexDefinitionMap.entrySet()) {
            ElasticsearchIndexDefinition def = entry.getValue();
            bulkClient.stopBulk(def);
        }
    }

    public void stopBulk(String... defs) {
        for (Map.Entry<String, ElasticsearchIndexDefinition> entry : indexDefinitionMap.entrySet()) {
            for (String def : defs) {
                if (def.equals(entry.getKey())) {
                    ElasticsearchIndexDefinition elasticsearchIndexDefinition = entry.getValue();
                    bulkClient.stopBulk(elasticsearchIndexDefinition);
                }
            }
        }
    }

    public void activateIndex() {
        for (Map.Entry<String, ElasticsearchIndexDefinition> entry : getIndexDefinitions().entrySet()) {
            activateIndex(entry.getValue());
        }
    }

    public void activateIndex(String... defs) {
        for (Map.Entry<String, ElasticsearchIndexDefinition> entry : indexDefinitionMap.entrySet()) {
            for (String def : defs) {
                if (def.equals(entry.getKey())) {
                    activateIndex(entry.getValue());
                }
            }
        }
    }

    public void activateIndex(ElasticsearchIndexDefinition indexDefinition) {
        if (indexDefinition.isEnabled()) {
            logger.log(Level.INFO, "now shifting index: " + indexDefinition);
            adminClient.shiftIndex(indexDefinition, Collections.singletonList(indexDefinition.getIndex()), null);
        }
        if (indexDefinition.isEnabled()) {
            logger.log(Level.INFO, "now pruning index: " + indexDefinition);
            adminClient.pruneIndex(indexDefinition);
        }
        if (indexDefinition.isEnabled()) {
            logger.log(Level.INFO, "now merging (force): " + indexDefinition);
            adminClient.forceMerge(indexDefinition, 1);
        }
        if (indexDefinition.isEnabled()) {
            logger.log(Level.INFO, "now updating replica of index: " + indexDefinition);
            adminClient.updateReplicaLevel(indexDefinition);
        }
        if (indexDefinition.isEnabled()) {
            logger.log(Level.INFO, "now checking mapping  of index: " + indexDefinition);
            adminClient.checkMapping(indexDefinition);
        }
    }

    private static AdminClient createAdminClient(Settings settings) throws IOException {
        ClientBuilder clientBuilder = ClientBuilder.builder();
        if (!settings.getAsBoolean("enabled", true)) {
            clientBuilder.setAdminClientProvider(MockAdminClientProvider.class);
        } else {
            clientBuilder.setAdminClientProvider(TransportAdminClientProvider.class);
        }
        return clientBuilder.put(filter(settings, ElxOutputHelper::isSettingsDisallowed).getAsMap()).build();
    }

    private static BulkClient createBulkClient(Settings settings) throws IOException {
        ClientBuilder clientBuilder = ClientBuilder.builder();
        if (!settings.getAsBoolean("enabled", true)) {
            clientBuilder.setBulkClientProvider(MockBulkClientProvider.class);
        } else {
            clientBuilder.setBulkClientProvider(TransportBulkClientProvider.class);
        }
        clientBuilder.put(org.xbib.elx.common.Parameters.BULK_METRIC_ENABLED.getName(), Boolean.TRUE);
        return clientBuilder.put(filter(settings, ElxOutputHelper::isSettingsDisallowed).getAsMap()).build();
    }

    private static Settings filter(Settings settings, Predicate<String> predicate) {
        SettingsBuilder builder = Settings.settingsBuilder();
        for (Map.Entry<String, String> entry : settings.getAsMap().entrySet()) {
            if (predicate.test(entry.getKey())) {
                builder.put(entry.getKey(), entry.getValue());
            }
        }
        return builder.build();
    }

    private static boolean isSettingsDisallowed(String key) {
        if (key.startsWith("index")) {
            return false;
        }
        for (Parameters p : Parameters.values()) {
            if (key.equals(p.getName())) {
                return false;
            }
        }
        return true;
    }
}
