package org.xbib.interlibrary.catalog.elasticsearch.bib;

import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.marc.Marc;
import org.xbib.marc.dialects.pica.PicaXMLContentHandler;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.settings.Settings;

import java.io.IOException;
import java.io.InputStream;
import java.text.Normalizer;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Addresses extends ElasticsearchTool<String> {

    public Addresses() {
    }

    @Override
    public AddressesWorkerPool newPool(Settings settings) throws IOException {
        return new AddressesWorkerPool(settings);
    }

    public class AddressesWorkerPool extends ElasticsearchWorkerPool<String> {

        public AddressesWorkerPool(Settings settings) throws IOException {
            super("addresses", settings);
        }

        @Override
        public AddressesWorker newWorker(String name, int count) throws IOException {
            return new AddressesWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }

    public class AddressesWorker extends ElasticsearchWorker<String, AddressesWorkerPool> {

        private final Logger logger = Logger.getLogger(AddressesWorker.class.getName());

        private final CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        private final AddressesWorkerPool pool;

        public AddressesWorker(Settings settings, AddressesWorkerPool pool) throws IOException {
            super(settings);
            this.pool = pool;
            this.catalogEntityWorkerPool = new CatalogEntityWorkerPool<>(settings,
                    new DefaultWorkerPoolMarcRecordListener(), AddressesCatalogEntityBuilder::new, pool);
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings)) {
                PicaXMLContentHandler contentHandler = new PicaXMLContentHandler();
                contentHandler.setFormat("Pica");
                contentHandler.setType("XML");
                contentHandler.setMarcListener(catalogEntityWorkerPool);
                MarcValueTransformers marcValueTransformers = new MarcValueTransformers();
                marcValueTransformers.setMarcValueTransformer(value ->
                        Normalizer.normalize(value, Normalizer.Form.NFKC));
                Marc.builder()
                        .disableControlFields()
                        .setInputStream(inputStream)
                        .setContentHandler(contentHandler)
                        .setMarcValueTransformers(marcValueTransformers)
                        .build()
                        .xmlReader()
                        .parse();
            } catch (Exception e) {
                logger.log(Level.WARNING, "uri = " + uri + " error =" + e.getMessage(), e);
            }
        }

        protected class AddressesCatalogEntityBuilder extends CatalogEntityBuilder {

            protected AddressesCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                ElasticsearchIndexDefinition elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
                if (elasticsearchIndexDefinition == null) {
                    throw new IOException("no 'bib' index definition configured");
                }
                RouteRdfXContentParams routeRdfXContentParams =
                        new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                                elasticsearchIndexDefinition.getFullIndexName(),
                                elasticsearchIndexDefinition.getType());
                routeRdfXContentParams.setHandler((content, p) ->
                        pool.bulkIndex(elasticsearchIndexDefinition, state.getRecordIdentifier(), false, content));
                try (RdfContentBuilder<?> builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(state.getResource());
                }
            }
        }
    }
}
