package org.xbib.interlibrary.catalog.elasticsearch.biomedcentral;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.Literal;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.internal.DefaultAnonymousResource;
import org.xbib.content.rdf.internal.DefaultLiteral;
import org.xbib.interlibrary.catalog.entities.ArticleVocabulary;
import org.xbib.interlibrary.catalog.matching.AuthoredWorkKey;
import org.xbib.interlibrary.catalog.matching.EncoderException;

public class BioMedCentralMapper implements ArticleVocabulary {

    private static final Logger logger = Logger.getLogger(BioMedCentralMapper.class.getName());

    private final StringBuilder title = new StringBuilder();

    private final Set<Author> authors = new LinkedHashSet<>();

    private Author author;

    private String date;

    private String volume;

    private String issue;

    private String pagination;

    private String journal;

    private String issn;

    private String doi;

    public Resource map(Map<String, ?> map) throws IOException {
        Resource r = new DefaultAnonymousResource();
        map(r, null, map);
        r.add(DC_TITLE, title.toString());
        String year = date.substring(0, 4);
        r.add(DC_DATE, new DefaultLiteral(year).type(Literal.GYEAR));
        r.add(PRISM_DOI, doi);
        r.newResource(FRBR_EMBODIMENT)
                .a(FABIO_PERIODICAL_ISSUE)
                .add(PRISM_NUMBER, issue);
        r.newResource(FRBR_EMBODIMENT)
                .a(FABIO_PERIODICAL_VOLUME)
                .add(PRISM_VOLUME, volume);
        r.newResource(FRBR_EMBODIMENT)
                .a(FABIO_PRINT_OBJECT)
                .add(PRISM_STARTING_PAGE, pagination);
        r.newResource(FRBR_PARTOF)
                .a(FABIO_JOURNAL)
                .add(PRISM_PUBLICATION_NAME, journal)
                .add(PRISM_ISSN, issn);
        AuthoredWorkKey wa = new AuthoredWorkKey()
                .workName(title)
                .chronology(year);
        for (Author author : authors) {
            wa.authorNameWithForeNames(author.lastName, author.foreName);
            r.newResource(DC_CREATOR)
                    .a(FOAF_AGENT)
                    .add(FOAF_FAMILYNAME, author.lastName)
                    .add(FOAF_GIVENNAME, author.foreName);
        }
        try {
            r.add(XBIB_KEY, wa.createKey());
        } catch (EncoderException e) {
            throw new IOException(e);
        }
        return r;
    }

    @SuppressWarnings("unchecked")
    private void map(Resource r, String p, Map<String, ?> map) throws IOException {
        for (Map.Entry<String, ?> entry : map.entrySet()) {
            String key = entry.getKey();
            Object value = entry.getValue();
            String path = p != null ? p + "." + key : key;
            if (value instanceof Map) {
                map(r, path, (Map<String, Object>) value);
            } else if (value instanceof List) {
                for (Object o : (List) value) {
                    if (o instanceof Map) {
                        map(r, path, (Map<String, Object>) o);
                    } else {
                        map(r, path, o.toString());
                    }
                }
            } else {
                if (value != null) {
                    map(r, path, value.toString());
                }
            }
        }
    }

    public void map(Resource r, String path, String value) throws IOException {
        logger.log(Level.INFO, MessageFormat.format("path={0} value={1}", path, value));
        switch (path) {
            case "bibl.title.p.it":
                if (title.length() > 0) {
                    title.append(" ");
                }
                title.append(value);
                break;
            case "bibl.aug.au.snm":
                if (author != null) {
                    if (author.lastName != null) {
                        authors.add(author);
                        author = new Author();
                    }
                } else {
                    author = new Author();
                }
                author = new Author();
                author.setLastName(value);
                break;
            case "bibl.aug.au.fnm":
                if (author != null) {
                    if (author.foreName != null) {
                        authors.add(author);
                        author = new Author();
                    }
                } else {
                    author = new Author();
                }
                author.setForeName(value);
                break;
            case "bibl.source":
                journal = value;
                break;
            case "bibl.issn":
                issn = value;
                break;
            case "bibl.pubdate":
                date = value;
                break;
            case "bibl.volume":
                volume = value;
                break;
            case "bibl.issue":
                issue = value;
                break;
            case "bibl.fpage":
                pagination = value;
                break;
            case "bibl.xrefbib.pubidlist.pubid":
                // doi
                doi = value;
                break;
            case "bibl.aug.au.cnm":
            case "bibl.insg.ins.p":
            default:
                break;
        }
    }

    static class Author implements Comparable<Author> {
        private String lastName;
        private String foreName;
        private String normalized;

        void setLastName(String lastName) {
            this.lastName = lastName;
            this.normalized = normalize();
        }

        void setForeName(String foreName) {
            this.foreName = foreName;
            this.normalized = normalize();
        }

        private String normalize() {
            StringBuilder sb = new StringBuilder();
            if (lastName != null) {
                sb.append(lastName);
            }
            if (foreName != null) {
                sb.append(' ').append(foreName);
            }
            return sb.toString();
        }

        @Override
        public int compareTo(Author o) {
            return normalized.compareTo(o.normalized);
        }

        @Override
        public boolean equals(Object o) {
            return o instanceof Author && normalized.equals(((Author) o).normalized);
        }

        @Override
        public int hashCode() {
            return normalized.hashCode();
        }
    }

}
