package org.xbib.interlibrary.catalog.elasticsearch.brs;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.StandardCharsets;
import java.text.MessageFormat;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import org.xbib.content.rdf.Literal;
import org.xbib.content.rdf.RdfConstants;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.internal.DefaultAnonymousResource;
import org.xbib.content.rdf.internal.DefaultLiteral;
import org.xbib.content.resource.IRI;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.api.util.CharacterEntities;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.ArticleVocabulary;
import org.xbib.interlibrary.catalog.matching.AuthoredWorkKey;
import org.xbib.settings.Settings;

public class BRSSearch extends ElasticsearchTool<String> {

    public BRSSearch() {
    }

    @Override
    public BRSSearchWorkerPool newPool(Settings settings) throws IOException {
        return new BRSSearchWorkerPool(settings);
    }

    public class BRSSearchWorkerPool extends ElasticsearchWorkerPool<String> {

        private final IRINamespaceContext namespaceContext;

        public BRSSearchWorkerPool(Settings settings) throws IOException {
            super("brs", settings);
            this.namespaceContext = createNamespaceContext();
        }

        @Override
        public BRSSearchWorker newWorker(String name, int count) {
            return new BRSSearchWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        private IRINamespaceContext createNamespaceContext() {
            IRINamespaceContext namespaceContext = IRINamespaceContext.newInstance(BRSSearch.class);
            namespaceContext.addNamespace(RdfConstants.NS_PREFIX, RdfConstants.NS_URI);
            namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
            namespaceContext.addNamespace("dcterms", "http://purl.org/dc/terms/");
            namespaceContext.addNamespace("foaf", "http://xmlns.com/foaf/0.1/");
            namespaceContext.addNamespace("frbr", "http://purl.org/vocab/frbr/core#");
            namespaceContext.addNamespace("fabio", "http://purl.org/spar/fabio/");
            namespaceContext.addNamespace("prism", "http://prismstandard.org/namespaces/basic/3.0/");
            return namespaceContext;
        }
    }

    private static final Pattern datepattern = Pattern.compile("^\\d\\d\\d\\d.*");

    private static final Pattern volissuepattern1 =
            Pattern.compile("^(\\d\\d\\d\\d), Vol\\.\\s*(.*?), Nr\\.\\s*(.*?), S\\.\\s*(.*)$");

    private static final Pattern volissuepattern2 =
            Pattern.compile("^(\\d\\d\\d\\d), Vol\\.\\s*(.*?), T\\.\\s*(.*?), S\\.\\s*(.*)$");

    private static final Pattern volissuepattern3 =
            Pattern.compile("^(\\d\\d\\d\\d), Vol\\.\\s*(.*?), Vol\\.\\s*(.*?), S\\.\\s*(.*)$");

    private static final Pattern volissuepattern4 =
            Pattern.compile("^(\\d\\d\\d\\d), Vol\\.\\s*(.*?), S\\.\\s*(.*)$");

    private static final Pattern volissuepattern5 =
            Pattern.compile("^(\\d\\d\\d\\d), T\\.\\s*(.*?), S\\.\\s*(.*)$");

    private static final Pattern volissuepattern6 =
            Pattern.compile("^(\\d\\d\\d\\d), Nr\\.\\s*(.*?), S\\.\\s*(.*)$");

    private static final Pattern volissuepattern7 =
            Pattern.compile("^(\\d\\d\\d\\d), \\s*(.*?), S\\.\\s*(.*)$");

    private static final Pattern partpattern = Pattern.compile(",\"Part(.*?)\"");

    private static final AtomicInteger counter = new AtomicInteger();

    public class BRSSearchWorker extends ElasticsearchWorker<String, BRSSearchWorkerPool> implements ArticleVocabulary {

        private final Logger logger = Logger.getLogger(BRSSearchWorker.class.getName());

        private final Set<Author> authors;

        private final BRSSearchWorkerPool pool;

        private String title;

        private String year;

        private String journal;

        private String issn;

        private String volume;

        private String issue;

        private String pagination;

        private String citation;

        public BRSSearchWorker(Settings settings, BRSSearchWorkerPool pool) {
            super(settings);
            this.pool = pool;
            authors = new LinkedHashSet<>();
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            try (InputStream in = ConverterInput.getInputStream(uri, settings)) {
                LinkedList<String> lines = new LinkedList<>();
                Resource resource = new DefaultAnonymousResource();
                Reader r = new InputStreamReader(in, StandardCharsets.ISO_8859_1);
                BufferedReader reader = new BufferedReader(r);
                String line = reader.readLine();
                while (line != null) {
                    if (line.startsWith("*** BRS DOCUMENT BOUNDARY ***")) {
                        complete(resource);
                        resource = new DefaultAnonymousResource();
                        authors.clear();
                        title = null;
                        year = null;
                        journal = null;
                        issn = null;
                        volume = null;
                        issue = null;
                        pagination = null;
                        citation = null;
                        line = reader.readLine();
                    } else {
                        lines.add(line);
                        line = reader.readLine();
                        while (line != null && !isKey(line) && !isBoundary(line)) {
                            lines.add(line.trim());
                            line = reader.readLine();
                        }
                        process(uri, resource, lines);
                        lines.clear();
                    }
                }
                reader.close();
                complete(resource);
            }
            logger.log(Level.INFO, MessageFormat.format("end of {0}, counter = {1}",
                    uri, counter.get()));
        }

        private void process(String uri, Resource resource, LinkedList<String> lines) {
            StringBuilder sb = new StringBuilder();
            Iterator<String> it = lines.iterator();
            sb.append(it.next());
            while (it.hasNext()) {
                sb.append(" ").append(it.next());
            }
            String line = sb.toString();
            int pos = line.indexOf(':');
            String key = pos > 0 ? line.substring(0, pos) : line;
            String value = pos > 0 ? line.substring(pos + 1) : line;
            process(uri, resource, key, value);
        }

        private void process(String uri, Resource resource, String key, String valueString) {
            String value = valueString;
            switch (key) {
                case "..IDNR", "..Document-Number", "..QUEL" -> {
                }
                // ignore
                case "..AUTR" -> {
                    value = clean(value);
                    Author author = new Author(value);
                    // forename?
                    authors.add(author);
                }
                case "..TITL" -> {
                    if (value.endsWith(".")) {
                        value = value.substring(0, value.length() - 1);
                    }
                    value = clean(value);
                    if (value.endsWith("(Book Review)")) {
                        value = value.substring(0, value.length() - 13);
                        if (value.isEmpty()) {
                            value = "Review";
                        }
                        resource.a(FABIO_REVIEW);
                    } else {
                        resource.a(FABIO_ARTICLE);
                    }
                    resource.add(DC_TITLE, value);
                    title = value;
                }
                case "..ISSN" -> {
                    if (!"NONE-XXXX".equals(value)) {
                        issn = value;
                    }
                }
                case "..AZIT" -> {
                    citation = value;
                    resource.add(DCTERMS_BIBLIOGRAPHIC_CITATION, value);
                    parseCitation(value);
                    if (volume != null) {
                        // there still can be clutter which troubles match key building
                        volume = volume.replaceAll("\"", "").replaceAll("/", "");
                        resource.newResource(FRBR_EMBODIMENT)
                                .a(FABIO_PERIODICAL_VOLUME)
                                .add(PRISM_VOLUME, volume);
                    }
                    if (issue != null) {
                        // fix clutter
                        Matcher matcher = partpattern.matcher(issue);
                        StringBuilder sb = new StringBuilder();
                        if (matcher.find()) {
                            matcher.appendReplacement(sb, Matcher.quoteReplacement(matcher.group(1)));
                            issue = sb.toString();
                        } else if (",\"Complete\"".equalsIgnoreCase(issue)) {
                            issue = "Complete";
                        } else if (",\"Supplement\"".equalsIgnoreCase(issue) || ",\"Supl.1\"".equalsIgnoreCase(issue)) {
                            issue = "Supplement";
                        } else if (",\"Index\"".equalsIgnoreCase(issue)) {
                            issue = "Index";
                        }
                        // there still can be clutter which troubles match key building
                        issue = issue.replaceAll("\"", "").replaceAll("/", "");
                        resource.newResource(FRBR_EMBODIMENT)
                                .a(FABIO_PERIODICAL_ISSUE)
                                .add(PRISM_NUMBER, issue);
                    }
                    if (pagination != null) {
                        int pos = pagination.indexOf('-');
                        if (pos > 0) {
                            String spage = pagination.substring(0, pos).trim();
                            String epage = pagination.substring(pos + 1).trim();
                            int l = spage.length() - epage.length();
                            if (l > 0) {
                                epage = spage.substring(0, l) + epage;
                            }
                            if (spage.equals(epage)) {
                                epage = null; // do not set starting page to end page
                            }
                            resource.newResource(FRBR_EMBODIMENT)
                                    .a(FABIO_PRINT_OBJECT)
                                    .add(PRISM_STARTING_PAGE, spage)
                                    .add(PRISM_ENDING_PAGE, epage);
                        } else {
                            resource.newResource(FRBR_EMBODIMENT)
                                    .a(FABIO_PRINT_OBJECT)
                                    .add(PRISM_PAGERANGE, value);
                        }
                    }
                }
                case "..AZST" -> journal = value;
                case "..JAHR" -> {
                    resource.add(PRISM_PUBLICATIONDATE, value);
                    Matcher matcher = datepattern.matcher(value);
                    if (matcher.matches()) {
                        try {
                            resource.add(DC_DATE, new DefaultLiteral(value.substring(0, 4)).type(Literal.GYEAR));
                            year = value.substring(0, 4);
                        } catch (NumberFormatException e) {
                            logger.log(Level.SEVERE, e.getMessage());
                        }
                    }
                }
                case "..ABST" -> {
                    value = clean(value);
                    resource.add(DCTERMS_ABSTRACT, value);
                }
                default -> logger.log(Level.WARNING, MessageFormat.format("{0}: unknown key: {1}", uri, key));
            }
        }

        private void complete(Resource resource) throws IOException {
            if (journal != null || issn != null) {
                resource.newResource(FRBR_PARTOF)
                        .a(FABIO_JOURNAL)
                        .add(PRISM_PUBLICATION_NAME, journal)
                        .add(PRISM_ISSN, issn);
            }
            AuthoredWorkKey wa = new AuthoredWorkKey()
                    .workName(title)
                    .authorName(authors.stream().map(Author::name).collect(Collectors.toList()))
                    .chronology(year)
                    .chronology(volume)
                    .chronology(issue);
            for (Author author : authors) {
                resource.newResource(DC_CREATOR)
                        .a(FOAF_AGENT)
                        .add(FOAF_NAME, author.name);
            }
            String key = null;
            try {
                key = wa.createKey();
                if (key != null) {
                    resource.add(XBIB_KEY, key);
                    resource.setId(IRI.create("http://xbib.info/key/" + key));
                    flush(resource);
                }
            } catch (Exception e) {
                logger.log(Level.SEVERE, "wrong key = " + key + " citation = " + citation + " : " + e.getMessage(), e);
            }
        }

        private void flush(Resource resource) {
            counter.incrementAndGet();
            try {
                ElasticsearchIndexDefinition elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
                RouteRdfXContentParams params = new RouteRdfXContentParams(pool.namespaceContext,
                        elasticsearchIndexDefinition.getFullIndexName(),
                        elasticsearchIndexDefinition.getType());
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                    builder.receive(resource);
                }
                if (!resource.isEmpty() && resource.id() != null) {
                    pool.bulkIndex(elasticsearchIndexDefinition,
                            resource.id().toString(), false, params.getGenerator().get());
                }
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        private boolean isKey(String line) {
            boolean b = false;
            if (line.length() > 2) {
                b = line.charAt(0) == '.' && line.charAt(1) == '.' && line.charAt(2) != '.';
            }
            return b;
        }

        private boolean isBoundary(String line) {
            if (line == null) {
                return false;
            }
            boolean b = false;
            if (line.length() > 2) {
                b = line.charAt(0) == '*' && line.charAt(1) == '*' && line.charAt(2) == '*';
            }
            return b;
        }

        private String clean(String s) {
            String v = s;
            v = CharacterEntities.HTML40.unescape(v); // HTML entities
            v = v.replaceAll("\\<[^>]*>", ""); // XML tags
            v = v.replaceAll("amp;", "").replaceAll("lt;", "").replaceAll("gt;", "");
            v = v.replaceAll("\\[non\\-\\s*Roman script word\\]", "");
            v = v.replaceAll("\\(non\\-\\s*Roman script word\\)", "");
            v = v.replaceAll("\\s{2,}", " ");
            v = v.replaceAll("Ã„", "Ä");
            v = v.replaceAll("Ã¤", "ä");
            v = v.replaceAll("Ã–", "Ö");
            v = v.replaceAll("Ã¶", "ö");
            v = v.replaceAll("Ãœ", "Ü");
            v = v.replaceAll("Ã¼", "ü");
            v = v.replaceAll("ÃŸ", "ß");
            return v;
        }

        private void parseCitation(String value) {
            Matcher matcher = volissuepattern1.matcher(value);
            if (matcher.matches()) {
                volume = matcher.group(2) != null ? removeBlanks(matcher.group(2)) : null;
                issue = matcher.group(3) != null ? removeBlanks(matcher.group(3)) : null;
                pagination = matcher.group(4) != null ? removeBlanks(matcher.group(4)) : null;
                return;
            }
            matcher = volissuepattern2.matcher(value);
            if (matcher.matches()) {
                volume = matcher.group(2) != null ? removeBlanks(matcher.group(2)) : null;
                issue = matcher.group(3) != null ? removeBlanks(matcher.group(3)) : null;
                pagination = matcher.group(4) != null ? removeBlanks(matcher.group(4)) : null;
                return;
            }
            matcher = volissuepattern3.matcher(value);
            if (matcher.matches()) {
                volume = matcher.group(2) != null ? removeBlanks(matcher.group(2)) : null;
                issue = matcher.group(3) != null ? removeBlanks(matcher.group(3)) : null;
                pagination = matcher.group(4) != null ? removeBlanks(matcher.group(4)) : null;
                return;
            }
            matcher = volissuepattern4.matcher(value);
            if (matcher.matches()) {
                volume = matcher.group(2) != null ? removeBlanks(matcher.group(2)) : null;
                pagination = matcher.group(3) != null ? removeBlanks(matcher.group(3)) : null;
                return;
            }
            matcher = volissuepattern5.matcher(value);
            if (matcher.matches()) {
                issue = matcher.group(2) != null ? removeBlanks(matcher.group(2)) : null;
                pagination = matcher.group(3) != null ? removeBlanks(matcher.group(3)) : null;
                return;
            }
            matcher = volissuepattern6.matcher(value);
            if (matcher.matches()) {
                issue = matcher.group(2) != null ? removeBlanks(matcher.group(2)) : null;
                pagination = matcher.group(3) != null ? removeBlanks(matcher.group(3)) : null;
                return;
            }
            matcher = volissuepattern7.matcher(value);
            if (matcher.matches()) {
                pagination = matcher.group(3) != null ? removeBlanks(matcher.group(3)) : null;
            }
        }

        private String removeBlanks(String s) {
            return s.replaceAll("\\s+", "");
        }

        private static class Author implements Comparable<Author> {
            private final String name;

            Author(String name) {
                this.name = name;
            }

            String name() {
                return name;
            }

            @Override
            public int compareTo(Author o) {
                return name().compareTo(o.name());
            }

            @Override
            public boolean equals(Object o) {
                return o instanceof Author && name.equals(((Author) o).name);
            }

            @Override
            public int hashCode() {
                return name.hashCode();
            }
        }
    }
}
