package org.xbib.interlibrary.catalog.elasticsearch.ezb;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.net.URI;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.time.LocalDate;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilders;
import org.xbib.content.rdf.RdfConstants;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfContentFactory;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.internal.DefaultAnonymousResource;
import org.xbib.content.rdf.io.turtle.TurtleContentParams;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.elx.api.SearchClient;
import org.xbib.elx.api.SearchDocument;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.settings.Settings;

/**
 * Indexing EZB web service of german inter library loan codes (ILL) into Elasticsearch.
 */
public class EZBWeb extends ElasticsearchTool<String> {

    private static final Logger logger = Logger.getLogger(EZBWeb.class.getName());

    private static final String EMPTY = "";

    public EZBWeb() {
    }

    @Override
    public EZBWebWorkerPool newPool(Settings settings) throws IOException {
        return new EZBWebWorkerPool(settings);
    }

    public class EZBWebWorkerPool extends ElasticsearchWorkerPool<String> {

        public EZBWebWorkerPool(Settings settings) throws IOException {
            super("ezb-web", settings);
            String[] beforeCommand = settings.getAsArray("beforeCommand");
            if (beforeCommand != null && beforeCommand.length > 0) {
                executeCommand(beforeCommand);
            }
        }

        @Override
        public void close() throws IOException {
            super.close();
            String[] afterCommand = settings.getAsArray("afterCommand");
            if (afterCommand != null && afterCommand.length > 0) {
                executeCommand(afterCommand);
            }
        }

        @Override
        public EZBWebWorker newWorker(String name, int count) {
            return new EZBWebWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        @SuppressWarnings("unchecked")
        @Override
        protected void openInput() throws IOException, InterruptedException {
            super.openInput();
            String baseURL = getConverterInput().getFileMap().get("ezb").getFirst();
            String identifier = settings.get("identifier");
            if (identifier != null) {
                StringBuilder sb = new StringBuilder(baseURL);
                sb.append(identifier).insert(sb.length() - 1, '-');
                submit(newWorkerRequest(sb.toString(), baseURL));
            } else {
                ElasticsearchIndexDefinition def = getInputIndexDefinitionMap().get("zdb");
                try (SearchClient searchClient = searchClient();
                     Stream<SearchDocument> stream = searchClient.search(srb ->
                                srb.setIndices(def.getIndex())
                                        .setSize(100)
                                        .setScroll(TimeValue.timeValueMinutes(1))
                                        .setQuery(QueryBuilders.matchQuery("xbib.ezb", "true")),
                        TimeValue.timeValueMinutes(1), 100)) {
                    stream.forEach(doc -> {
                        Object object = doc.getFields().get("IdentifierZDB");
                        if (object instanceof List) {
                            object = ((List<Object>) object).getFirst();
                        }
                        Map<String, Object> map = (Map<String, Object>) object;
                        String zdbid = (String) map.get("identifierZDB");
                        StringBuilder sb = new StringBuilder(baseURL);
                        sb.append(zdbid).insert(sb.length() - 1, '-');
                        submit(newWorkerRequest(sb.toString(), baseURL));
                    });
                }
            }
        }
    }

    public class EZBWebWorker extends ElasticsearchWorker<String, EZBWebWorkerPool> {

        private final Logger logger = Logger.getLogger(EZBWebWorker.class.getName());

        private final EZBWebWorkerPool pool;

        public EZBWebWorker(Settings settings, EZBWebWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            IRINamespaceContext namespaceContext = createNamespaceContext();
            ElasticsearchIndexDefinition indexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
            RouteRdfXContentParams params = new RouteRdfXContentParams(namespaceContext,
                    indexDefinition.getFullIndexName(), "type");
            // optional turtle output
            RdfContentBuilder<?> turtleBuilder = null;
            if (pool.getConverterOutput().getMap().containsKey("turtle")) {
                OutputStream outputStream = pool.getConverterOutput().getMap().get("turtle").getOut();
                TurtleContentParams turtleParams = new TurtleContentParams(namespaceContext, true);
                turtleBuilder = RdfContentFactory.turtleBuilder(turtleParams, outputStream);
            }
            logger.log(Level.FINE, "received URI " + uri);
            int pos = uri.lastIndexOf('/');
            String zdbidWithHyphen = pos > 0 ? uri.substring(pos + 1) : uri;
            // remove hyphen from ZDB-ID
            StringBuilder sb = new StringBuilder().append(zdbidWithHyphen);
            sb.deleteCharAt(sb.length() - 2);
            String zdbid = sb.toString();
            try (BufferedReader br = new BufferedReader(new StringReader(getContent(uri)))) {
                String zdbidLine = br.readLine(); // ZDB-Id: ...
                String hitLine = br.readLine(); // Treffer: ...
                String emptyLine = br.readLine(); // empty line
                int count = 0;
                String line;
                while ((line = br.readLine()) != null) {
                    String[] s = line.split("\\t");
                    try {
                        int i = 0;
                        String sigel = i < s.length ? s[i++] : EMPTY;
                        String isil = i < s.length ? s[i++] : EMPTY;
                        if (i < s.length) {
                            i++;
                        }
                        String code1 = i < s.length ? s[i++] : EMPTY;
                        String code2 = i < s.length ? s[i++] : EMPTY;
                        String code3 = i < s.length ? s[i++] : EMPTY;
                        String comment = i < s.length ? s[i++] : EMPTY;
                        String firstDate = i < s.length ? s[i++] : EMPTY;
                        String firstVolume = i < s.length ? s[i++] : EMPTY;
                        String firstIssue = i < s.length ? s[i++] : EMPTY;
                        String lastDate = i < s.length ? s[i++] : EMPTY;
                        String lastVolume = i < s.length ? s[i++] : EMPTY;
                        String lastIssue = i < s.length ? s[i++] : EMPTY;
                        String movingWall = i < s.length ? s[i++] : EMPTY;
                        String licenseType = i < s.length ? s[i] : EMPTY;

                        // AAAAA entry ("free serials")
                        if ("AAAAA".equals(sigel)) {
                            isil = "DE-ALL";
                        }
                        String key = zdbid + "." + isil + "."
                                + firstDate + "." + firstVolume + "." + firstIssue + "."
                                + lastDate + "." + lastVolume + "." + lastIssue + "."
                                + (movingWall.isEmpty() ? "0" : movingWall);
                        // fix to null for resource add() method
                        if ("0".equals(firstVolume)) {
                            firstVolume = null;
                        }
                        if ("0".equals(firstIssue)) {
                            firstIssue = null;
                        }
                        if ("0".equals(lastVolume)) {
                            lastVolume = null;
                        }
                        if ("0".equals(lastIssue)) {
                            lastIssue = null;
                        }
                        if (firstDate.isEmpty()) {
                            firstDate = null;
                        }
                        if (lastDate.isEmpty()) {
                            lastDate = null;
                        }
                        if (movingWall.isEmpty()) {
                            movingWall = null;
                        }
                        Resource resource = new DefaultAnonymousResource();
                        resource.add("dc:identifier", key)
                                .add("xbib:identifier", zdbid)
                                .add("xbib:isil", isil)
                                .add("xbib:firstDate", firstDate)
                                .add("xbib:firstVolume", firstVolume)
                                .add("xbib:firstIssue", firstIssue)
                                .add("xbib:lastDate", lastDate)
                                .add("xbib:lastVolume", lastVolume)
                                .add("xbib:lastIssue", lastIssue)
                                .add("xbib:movingWall", movingWall)
                                .add("xbib:interlibraryloanCode",
                                        // replace empty value with "x" (unknown)
                                        (code1.isEmpty() ? "x" : code1)
                                                + (code2.isEmpty() ? "x" : code2)
                                                + (code3.isEmpty() ? "x" : code3))
                                .add("xbib:comment", comment)
                                .add("xbib:licenseType", licenseType)
                                .add("xbib:modified", LocalDate.now().toString());
                        params.setHandler((content, p) ->
                                pool.bulkIndex(indexDefinition, key, false, content));
                        try (RdfContentBuilder<?> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                            builder.receive(resource);
                        }
                        count++;
                        if (turtleBuilder != null) {
                            turtleBuilder.receive(resource);
                        }
                    } catch (NoSuchElementException e) {
                        logger.log(Level.SEVERE, uri + " " + e.getMessage(), e);
                    }
                }
                if (count == 0) {
                    logger.log(Level.WARNING, "really, no indicators in {} ?", uri);
                }
            } catch (Exception e) {
                throw new IOException(e);
            }
        }
    }

    private String getContent(String urlString) throws Exception {
        URL url = URI.create(urlString).toURL();
        try (InputStream inputStream = url.openStream()) {
            return new String(inputStream.readAllBytes(), StandardCharsets.UTF_8);
        }
    }

    private void executeCommand(String[] command) {
        ProcessBuilder processBuilder = new ProcessBuilder();
        processBuilder.command(command);
        try {
            Process process = processBuilder.start();
            int exitCode = process.waitFor();
            logger.log(Level.INFO, "command " + Arrays.asList(command) + " exited with " + exitCode);
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    private static IRINamespaceContext createNamespaceContext() {
        IRINamespaceContext namespaceContext = IRINamespaceContext.newInstance();
        namespaceContext.addNamespace(RdfConstants.NS_PREFIX, RdfConstants.NS_URI);
        namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
        namespaceContext.addNamespace("xbib", "http://xbib.org/elements/1.0/");
        return namespaceContext;
    }
}
