package org.xbib.interlibrary.catalog.elasticsearch.ezb;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.CodingErrorAction;
import java.nio.charset.MalformedInputException;
import java.nio.charset.StandardCharsets;
import java.nio.charset.UnmappableCharacterException;
import java.text.MessageFormat;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.time.temporal.WeekFields;
import java.util.Locale;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.namespace.QName;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.index.VersionType;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfContentParams;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RdfXContentParams;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.io.xml.AbstractXmlResourceHandler;
import org.xbib.content.rdf.io.xml.XmlContentParser;
import org.xbib.content.rdf.io.xml.XmlHandler;
import org.xbib.content.resource.IRI;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.content.resource.NamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.net.URL;
import org.xbib.net.URLSyntaxException;
import org.xbib.settings.Settings;
import org.xml.sax.SAXException;

/**
 * Elasticsearch indexer for "Elektronische Zeitschriftenbibliothek" (EZB) XMl formatted files.
 * The EZB XML files are encoded in ISO 8859-1.
 * Format documentation:
 * <a href="http://www.zeitschriftendatenbank.de/fileadmin/user_upload/ZDB/pdf/services/Datenlieferdienst_ZDB_EZB_Lizenzdatenformat.pdf">...</a>
 */
public class EZBXML extends ElasticsearchTool<String> {

    public EZBXML() {
    }

    @Override
    public EZBXMLWorkerPool newPool(Settings settings) throws IOException {
        return new EZBXMLWorkerPool(settings);
    }

    public class EZBXMLWorkerPool extends ElasticsearchWorkerPool<String> {

        public EZBXMLWorkerPool(Settings settings) throws IOException {
            super("ezb-xml", settings);
        }

        @Override
        public EZBXMLWorker newWorker(String name, int count) {
            return new EZBXMLWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }

    private static final DateTimeFormatter DATE_TIME_FORMATTER = new DateTimeFormatterBuilder()
            .appendPattern("YYYY ww")
            .parseDefaulting(WeekFields.ISO.dayOfWeek(), 1)
            .toFormatter().withLocale(Locale.GERMANY);


    public class EZBXMLWorker extends ElasticsearchWorker<String, EZBXMLWorkerPool> {

        private final Logger logger = Logger.getLogger(EZBXMLWorker.class.getName());

        private final EZBXMLWorkerPool pool;

        public EZBXMLWorker(Settings settings, EZBXMLWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            // EZB XML is encoded in ISO-8859-1
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings);
                 BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(inputStream,
                         StandardCharsets.ISO_8859_1))) {
                IRINamespaceContext namespaceContext = IRINamespaceContext.newInstance();
                RdfContentParams params = new RdfXContentParams(namespaceContext);
                // minimum must be 6 digits
                Pattern pattern = Pattern.compile("(\\d{6,})");
                Matcher matcher = pattern.matcher(uri);
                long version = matcher.find() ? Long.parseLong(matcher.group(1)) : 1L;
                logger.log(Level.INFO, MessageFormat.format("version of {0} -> {1}",
                        uri, version));
                EZBHandler handler = new EZBHandler(params, namespaceContext, uri, version);
                handler.setDefaultNamespace("ezb", "http://ezb.uni-regensburg.de/ezeit/");
                try {
                    XmlContentParser<RdfContentParams> parser = new XmlContentParser<RdfContentParams>(params, bufferedReader);
                    parser.setNamespaces(false)
                            .setHandler(handler)
                            .parse();
                } catch (Exception e) {
                    logger.log(Level.SEVERE, "file " + uri + ": " + e.getMessage(), e);
                }
            }
        }

        private class EZBHandler extends AbstractXmlResourceHandler<RdfContentParams> {

            private final long version;

            private final String uri;

            private final RouteRdfXContentParams routeRdfXContentParams;

            private NamespaceContext namespaceContext;

            EZBHandler(RdfContentParams params, IRINamespaceContext namespaceContext, String uri, long version) {
                super(params);
                this.uri = uri;
                this.version = version;
                setNamespaceContext(namespaceContext);
                this.routeRdfXContentParams = new RouteRdfXContentParams(namespaceContext,
                        pool.getOutputIndexDefinitionMap().get("bib").getFullIndexName(), "type");
            }

            @Override
            public void endElement(String uri, String localName, String qName) throws SAXException {
                super.endElement(uri, localName, qName);
            }

            @Override
            public void identify(QName name, String value, IRI identifier) {
                if ("license_entry_id".equals(name.getLocalPart())) {
                    getResource().setId(IRI.create(value));
                }
            }

            @Override
            public boolean isResourceDelimiter(QName name) {
                return "license_set".equals(name.getLocalPart());
            }

            @Override
            public void closeResource() throws IOException {
                super.closeResource();
                getResource().add("uri", uri);
                if (settings.get("collection") != null) {
                    getResource().add("collection", settings.get("collection"));
                }
                if (Long.toString(version).length() >= 6) {
                    // EZB weekly date stamps: yyyywwccc where ccc is a counter, normally 001
                    String timestamp = Long.toString(version).substring(0, 6);
                    // add a space, otherweise date pattern does not work
                    LocalDate localDate = LocalDate.parse(timestamp.substring(0, 4) + " " + timestamp.substring(4, 6),
                            DATE_TIME_FORMATTER);
                    // add ISO 8601 format
                    getResource().add("modified", localDate.toString());
                }
                routeRdfXContentParams.setHandler((content, p) -> {
                    if (!getResource().isEmbedded()) {
                        // use versioning to ensure docs in most recent files win
                        IndexRequest indexRequest = new IndexRequest()
                                .index(p.getIndex())
                                .id(getResource().id().toString())
                                .versionType(VersionType.EXTERNAL)
                                .version(version)
                                .source(content);
                        pool.bulkIndex(indexRequest);
                    }
                });
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(getResource());
                }
            }

            @Override
            public boolean skip(QName name) {
                return "ezb-export".equals(name.getLocalPart())
                        || "release".equals(name.getLocalPart())
                        || "version".equals(name.getLocalPart())
                        || name.getLocalPart().startsWith("@");
            }

            @Override
            public Object toObject(QName name, String content) {
                switch (name.getLocalPart()) {
                    case "reference_url":
                        // fall-through
                    case "readme_url":
                        try {
                            return URL.create(content);
                        } catch (Exception e) {
                            try {
                                return URL.parser(StandardCharsets.ISO_8859_1, CodingErrorAction.REPLACE).parse(content);
                            } catch (URLSyntaxException | MalformedInputException | UnmappableCharacterException ex) {
                                logger.log(Level.WARNING, "unable to create README URL from " + content);
                                return content;
                            }
                        }
                    case "zdbid":
                        return content.replaceAll("\\-", "").toLowerCase();
                    case "type_id":
                        switch (Integer.parseInt(content)) {
                            case 1:
                                return "online"; // Volltext nur online
                            case 2:
                                return "online-and-print"; // Volltext online und Druckausgabe
                            case 9:
                                return "self-hosted"; // lokale Zeitschrift
                            case 11:
                                return "digitalization"; // retrodigitalisiert
                            default:
                                break;
                        }
                        throw new IllegalArgumentException("unknown type_id: " + content);
                    case "license_type_id":
                        switch (Integer.parseInt(content)) {
                            case 1:
                                return "solitary"; // Einzellizenz
                            case 2:
                                return "consortial"; // Konsortiallizenz
                            case 4:
                                return "national"; // Nationallizenz
                            default:
                                break;
                        }
                        throw new IllegalArgumentException("unknown license_type_id: " + content);
                    case "price_type_id":
                        switch (Integer.parseInt(content)) {
                            case 1:
                                return "no"; // lizenzfrei
                            case 2:
                                return "no-with-print"; // Kostenlos mit Druckausgabe
                            case 3:
                                return "yes"; // Kostenpflichtig
                            default:
                                break;
                        }
                        throw new IllegalArgumentException("unknown price_type_id: " + content);
                    case "ill_code":
                        switch (content) {
                            case "n":
                                return "no"; // nein
                            case "l":
                                return "copy-loan"; // ja, Leihe und Kopie
                            case "k":
                                return "copy"; // ja, nur Kopie
                            case "e":
                                return "copy-electronic";  // ja, auch elektronischer Versand an Nutzer
                            case "ln":
                                return "copy-loan-domestic";  // ja, Leihe und Kopie (nur Inland)
                            case "kn":
                                return "copy-domestic";  // ja, nur Kopie (nur Inland)
                            case "en":
                                return "copy-electronic-domestic";  // ja, auch elektronischer Versand an Nutzer (nur Inland)
                            default:
                                break;
                        }
                        throw new IllegalArgumentException("unknown ill_code: " + content);
                    default:
                        break;
                }
                return super.toObject(name, content);
            }

            @Override
            public XmlHandler<RdfContentParams> setNamespaceContext(NamespaceContext namespaceContext) {
                this.namespaceContext = namespaceContext;
                return this;
            }

            @Override
            public NamespaceContext getNamespaceContext() {
                return namespaceContext;
            }
        }
    }
}
