package org.xbib.interlibrary.catalog.elasticsearch.ezb;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.xbib.elx.api.SearchDocument;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.marc.MarcRecord;
import org.xbib.settings.Settings;

public class ZDBEZB extends ElasticsearchTool<String> {

    private static final Logger logger = Logger.getLogger(ZDBEZB.class.getName());

    public ZDBEZB() {
    }

    @Override
    public ElasticsearchWorkerPool<String> newPool(Settings settings) throws IOException {
        return new ZDBEZBWorkerPool(settings);
    }

    public class ZDBEZBWorkerPool extends ElasticsearchWorkerPool<String> {

        public ZDBEZBWorkerPool(Settings settings) throws IOException {
            super("zdb-ezb", settings);
        }

        @Override
        public ZDBEZBWorker newWorker(String name, int count) {
            return new ZDBEZBWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        @Override
        protected void openInput() throws IOException, InterruptedException {
            super.openInput();
            submit(newWorkerRequest("dummy", "unknown origin"));
        }
    }

    public class ZDBEZBWorker extends ElasticsearchWorker<String, ZDBEZBWorkerPool> {

        private final ZDBEZBWorkerPool pool;

        public ZDBEZBWorker(Settings settings, ZDBEZBWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String request) throws IOException {
            Path path = pool.getConverterOutput().getMap().get("zdb-id-file").getPath();
            try (BufferedWriter bufferedWriter = Files.newBufferedWriter(path)) {
                logger.log(Level.INFO, "start writing to " + path);
                QueryBuilder query = QueryBuilders
                        .termQuery("856.4_.x", "EZB");
                Stream<SearchDocument> stream = pool.searchClient().search(srb ->
                                srb.setIndices(pool.getInputIndexDefinitionMap().get("zdb").getFullIndexName())
                                        .setSize(1000)
                                        .setScroll(TimeValue.timeValueMinutes(1))
                                        .setQuery(query),
                        TimeValue.timeValueMinutes(1), 1000);
                stream.forEach(doc -> {
                    MarcRecord marcRecord = MarcRecord.from(doc.getFields());
                    marcRecord.getAll("016").stream()
                            .filter(mf -> "DE-600".equals(mf.getFirstSubfieldValue("2")))
                            .map(mf -> mf.getFirstSubfieldValue("a"))
                            .filter(Objects::nonNull)
                            .findAny()
                            .ifPresent(zdbid -> {
                                try {
                                    bufferedWriter.write(zdbid);
                                    bufferedWriter.newLine();
                                } catch (IOException e) {
                                    throw new UncheckedIOException(e);
                                }
                            });
                });
                logger.log(Level.INFO, "stop writing to " + path);
            }
        }
    }
}
