package org.xbib.interlibrary.catalog.elasticsearch.gnd;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.util.zip.GZIPOutputStream;
import org.xbib.content.rdf.RdfConstants;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.Triple;
import org.xbib.content.rdf.io.rdfxml.RdfXmlContentParser;
import org.xbib.content.rdf.io.turtle.TurtleContentParams;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.settings.Settings;

import static org.xbib.content.rdf.RdfContentFactory.turtleBuilder;

/**
 * Indexing DNB GND from RDF/XML files into Elasticsearch.
 */
public class GNDRdfXml extends ElasticsearchTool<String> {

    public GNDRdfXml() {
    }

    @Override
    public GNDRdfXmlWorkerPool newPool(Settings settings) throws IOException {
        return new GNDRdfXmlWorkerPool(settings);
    }

    public class GNDRdfXmlWorkerPool extends ElasticsearchWorkerPool<String> {

        private final IRINamespaceContext namespaceContext;

        public GNDRdfXmlWorkerPool(Settings settings) throws IOException {
            super("gnd-rdf-xml", settings);
            this.namespaceContext = createNamespaceContext();
        }

        @Override
        public GNDRdfXmlWorker newWorker(String name, int count) {
            return new GNDRdfXmlWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        private static IRINamespaceContext createNamespaceContext() {
            IRINamespaceContext namespaceContext = IRINamespaceContext.newInstance();
            namespaceContext.addNamespace(RdfConstants.NS_PREFIX, RdfConstants.NS_URI);
            namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
            namespaceContext.addNamespace("geo", "http://rdvocab.info/");
            namespaceContext.addNamespace("rda", "http://purl.org/dc/elements/1.1/");
            namespaceContext.addNamespace("foaf", "http://xmlns.com/foaf/0.1/");
            namespaceContext.addNamespace("sf", "http://www.opengis.net/ont/sf#");
            namespaceContext.addNamespace("isbd", "http://iflastandards.info/ns/isbd/elements/");
            namespaceContext.addNamespace("gnd", "http://d-nb.info/standards/elementset/gnd#");
            namespaceContext.addNamespace("dcterms", "http://purl.org/dc/terms/");
            namespaceContext.addNamespace("rdfs", "http://www.w3.org/2000/01/rdf-schema#");
            namespaceContext.addNamespace("marcRole", "http://id.loc.gov/vocabulary/relators/");
            namespaceContext.addNamespace("lib", "http://purl.org/library/");
            namespaceContext.addNamespace("umbel", "http://umbel.org/umbel#");
            namespaceContext.addNamespace("bibo", "http://purl.org/ontology/bibo/");
            namespaceContext.addNamespace("owl", "http://www.w3.org/2002/07/owl#");
            namespaceContext.addNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
            namespaceContext.addNamespace("skos", "http://www.w3.org/2004/02/skos/core#");
            namespaceContext.addNamespace("geosparql", "http://www.opengis.net/ont/geosparql#");
            namespaceContext.addNamespace("frbr", "http://purl.org/vocab/frbr/core#");
            namespaceContext.addNamespace("fabio", "http://purl.org/spar/fabio/");
            namespaceContext.addNamespace("prism", "http://prismstandard.org/namespaces/basic/3.0/");
            return namespaceContext;
        }
    }

    public class GNDRdfXmlWorker extends ElasticsearchWorker<String, GNDRdfXmlWorkerPool> {

        private final GNDRdfXmlWorkerPool pool;

        public GNDRdfXmlWorker(Settings settings, GNDRdfXmlWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            TurtleContentParams params = new TurtleContentParams(pool.namespaceContext, false);
            RouteRdfXContentParams routeRdfXContentParams = new RouteRdfXContentParams(pool.namespaceContext, "index", "type");
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings)) {
                GNDRdfXmlContentParser reader = new GNDRdfXmlContentParser(params, routeRdfXContentParams, inputStream);
                if (pool.getConverterOutput().getMap().containsKey("turtle")) {
                    OutputStream outputStream = pool.getConverterOutput().getMap().get("turtle").getOut();
                    final Writer writer = new OutputStreamWriter(new GZIPOutputStream(outputStream),
                            StandardCharsets.UTF_8);
                    reader.setRdfContentBuilderProvider(() -> turtleBuilder(params));
                    reader.setRdfContentBuilderHandler(builder -> writer.write(builder.string()));
                    reader.parse();
                    writer.close();
                } else {
                    reader.parse();
                    reader.flush();
                }
            }
        }

        /*
         * For reasons I don't find, a normal routeRdfXContentBuilder did not work.
         * This is a workaround.
         */
        private class GNDRdfXmlContentParser extends RdfXmlContentParser<TurtleContentParams> {

            private final RouteRdfXContentParams routeRdfXContentParams;

            private Resource lastSubject;

            private RdfContentBuilder<?> builder;

            private String id;

            private GNDRdfXmlContentParser(TurtleContentParams params,
                                           RouteRdfXContentParams routeRdfXContentParams,
                                           InputStream inputStream) {
                super(params, inputStream);
                this.routeRdfXContentParams = routeRdfXContentParams;
            }

            @Override
            protected void yieldTriple(Triple t) throws IOException {
                if (lastSubject == null || !t.subject().equals(lastSubject)) {
                    if (builder != null) {
                        builder.endStream();
                        IndexDefinition indexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
                        pool.bulkIndex(indexDefinition, id, false, routeRdfXContentParams.getGenerator().get());
                        builder.close();
                    }
                    builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams);
                    builder.startStream();
                    builder.receive(t.subject().id());
                    lastSubject = t.subject();
                }
                if ("http://d-nb.info/standards/elementset/gnd#gndIdentifier".equals(t.predicate().toString())) {
                    id = t.object().toString();
                }
                builder.receive(t);
            }

            void flush() throws IOException {
                if (builder != null && lastSubject != null) {
                    builder.endStream();
                    IndexDefinition indexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
                    pool.bulkIndex(indexDefinition, id, false, routeRdfXContentParams.getGenerator().get());
                    builder.close();
                }
            }
        }
    }
}
