package org.xbib.interlibrary.catalog.elasticsearch.hbz;

import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;
import java.time.Instant;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.interlibrary.catalog.entities.ErrorFields;
import org.xbib.interlibrary.catalog.entities.Facet;
import org.xbib.interlibrary.catalog.entities.ISILTransformer;
import org.xbib.interlibrary.catalog.entities.SkipFields;
import org.xbib.interlibrary.catalog.entities.TransformFields;
import org.xbib.marc.Marc;
import org.xbib.marc.transformer.field.MarcFieldTransformers;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.settings.Settings;

public class TitleHoldings extends ElasticsearchTool<String> {

    @Override
    public TitleHoldingsWorkerPool newPool(Settings settings) throws IOException {
        return new TitleHoldingsWorkerPool(settings);
    }

    public class TitleHoldingsWorkerPool extends ElasticsearchWorkerPool<String> {

        private final IRINamespaceContext namespaceContext;

        protected ISILTransformer isilTransformer;

        protected TitleHoldingsWorkerPool(Settings settings) throws IOException {
            super("title-holdings", settings);
            this.namespaceContext = IRINamespaceContext.newInstance(TitleHoldings.class);
            this.isilTransformer = new ISILTransformer(settings);
        }

        @Override
        public TitleHoldingsWorker newWorker(String name, int count) throws IOException {
            return new TitleHoldingsWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        public IRINamespaceContext getNamespaceContext() {
            return namespaceContext;
        }
    }

    public class TitleHoldingsWorker extends ElasticsearchWorker<String, TitleHoldingsWorkerPool> {

        private final Logger logger = Logger.getLogger(TitleHoldingsWorker.class.getName());

        protected final MarcFieldTransformers marcFieldTransformers;

        protected final MarcValueTransformers marcValueTransformers;

        private final TitleHoldingsWorkerPool pool;

        private CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        protected TitleHoldingsWorker(Settings settings, TitleHoldingsWorkerPool pool) {
            super(settings);
            this.pool = pool;
            this.marcFieldTransformers = new MarcFieldTransformers();
            this.marcValueTransformers = new MarcValueTransformers();
            ErrorFields errorFields = new ErrorFields();
            errorFields.install(settings, marcFieldTransformers);
            SkipFields skipFields = new SkipFields();
            skipFields.install(settings, marcFieldTransformers);
            TransformFields transformFields = new TransformFields();
            transformFields.install(settings, marcFieldTransformers, marcFieldTransformers, marcFieldTransformers);
            pool.isilTransformer.createTransformer(settings, marcValueTransformers);
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings)) {
                catalogEntityWorkerPool = new CatalogEntityWorkerPool<>(settings,
                                 new DefaultWorkerPoolMarcRecordListener(), TitleHoldingsCatalogEntityBuilder::new, pool);
                catalogEntityWorkerPool.setRequestOrigin(uri);
                doExecute(inputStream, catalogEntityWorkerPool);
                catalogEntityWorkerPool.close();
            }
        }

        protected void doExecute(InputStream inputStream,
                                 CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool) throws IOException {
            Marc.builder()
                    .setInputStream(inputStream)
                    .setMarcFieldTransformers(marcFieldTransformers)
                    .setMarcValueTransformers(marcValueTransformers)
                    .setMarcRecordListener(catalogEntityWorkerPool)
                    .build()
                    .writeRecords();
        }

        protected class TitleHoldingsCatalogEntityBuilder extends CatalogEntityBuilder {

            private TitleHoldingsCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
                for (Map.Entry<String, List<String>> entry : pool.getConverterInput().getFileMap().entrySet()) {
                    String key = entry.getKey();
                    if (key.startsWith("taxonomy-")) {
                        String isil = key.substring("taxonomy-".length());
                        for (String taxFile : entry.getValue()) {
                            logger.log(Level.INFO, MessageFormat.format("adding classifier file from {0} for ISIL {1}",
                                    taxFile, isil));
                            addClassifier("(" + settings.get("catalogid") + ")", isil, taxFile);
                        }
                    }
                }
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                // write title resource
                final ElasticsearchIndexDefinition titleDef = pool.getOutputIndexDefinitionMap().get("title");
                if (titleDef == null) {
                    throw new IllegalArgumentException("no 'title' index definition");
                }
                RouteRdfXContentParams params = new RouteRdfXContentParams(pool.getNamespaceContext(),
                        titleDef.getFullIndexName(),
                        titleDef.getType());
                params.setHandler((content, p) ->
                        pool.bulkIndex(titleDef, state.getIdentifier(), false, content));
                if (settings.get("format") != null) {
                    state.getResource().add("format", settings.get("format"));
                }
                if (settings.get("type") != null) {
                    state.getResource().add("type", settings.get("type"));
                }
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("xbib")
                        .add("uid", state.getRecordIdentifier());
                state.getResource().newResource("rec")
                        .add("modified", Instant.now().toString())
                        .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                if (state.getResource().isDeleted()) {
                    pool.delete(titleDef, state.getIdentifier());
                } else {
                    try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                        builder.receive(state.getResource());
                    }
                    // write holdings resources
                    final ElasticsearchIndexDefinition holDef = pool.getOutputIndexDefinitionMap().get("holdings");
                    if (holDef == null) {
                        throw new IllegalArgumentException("no 'holdings' index definition");
                    }
                    String holdingsIndex = holDef.getFullIndexName();
                    String holdingsIndexType = holDef.getType();
                    Iterator<Resource> it = state.getResourceIterator();
                    while (it.hasNext()) {
                        Resource resource = it.next();
                        if (resource.equals(state.getResource())) {
                            // skip main resource
                            continue;
                        }
                        params = new RouteRdfXContentParams(pool.getNamespaceContext(), holdingsIndex,
                                holdingsIndexType);
                        params.setHandler((content, p) -> pool.bulkIndex(holDef,
                                state.getIdentifier() + "." + resource.id(), false, content));
                        resource.newResource("xbib")
                                .add("uid", state.getRecordIdentifier());
                        state.getResource().newResource("rec")
                                .add("modified", Instant.now().toString())
                                .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                        try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)){
                            builder.receive(resource);
                        }
                    }
                    // create more indices
                    for (Map.Entry<String, ElasticsearchIndexDefinition> entry : pool.getOutputIndexDefinitionMap().entrySet()) {
                        if (!"title".equals(entry.getKey()) && !"holdings".equals(entry.getKey())) {
                            createIndexByKey(entry.getKey(), entry.getValue(), state);
                        }
                    }
                }
            }

            protected void createIndexByKey(String key,
                                            ElasticsearchIndexDefinition indexDefinition,
                                            CatalogEntityBuilderState state)
                    throws IOException {
                // check holdings facet for filtering ISIL
                Facet<String> holdingsFacet = state.getFacets().get("xbib.identifier");
                if (holdingsFacet == null) {
                    logger.log(Level.WARNING, "no holdings facet found: {}", state.getFacets());
                    return;
                }
                if (holdingsFacet.getValues().contains(key)) {
                    // join main resource with matching index name
                    Resource resource = state.getResource();
                    Iterator<Resource> it = state.getResourceIterator();
                    while (it.hasNext()) {
                        Resource r = it.next();
                        if (r.equals(resource)) {
                            // skip main resource
                            continue;
                        }
                        // holdings document ID starts with ISIL
                        if (r.id().toString().startsWith(key + "/")) {
                            resource.add("Holdings", r);
                        }
                    }
                    RouteRdfXContentParams params = new RouteRdfXContentParams(pool.getNamespaceContext(),
                            indexDefinition.getFullIndexName(),
                            indexDefinition.getType());
                    params.setHandler((content, p) ->
                            pool.bulkIndex(indexDefinition, resource.id().toString(), false, content));
                    try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                        builder.receive(resource);
                    }
                }
            }
        }
    }
}
