package org.xbib.interlibrary.catalog.elasticsearch.hbz;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.io.UncheckedIOException;
import java.nio.charset.StandardCharsets;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.interlibrary.catalog.entities.ErrorFields;
import org.xbib.interlibrary.catalog.entities.ISILTransformer;
import org.xbib.interlibrary.catalog.entities.SkipFields;
import org.xbib.interlibrary.catalog.entities.TransformFields;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecordAdapter;
import org.xbib.marc.transformer.field.MarcFieldTransformers;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.settings.Settings;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class TitleHoldingsAlephXMLClobs extends TitleHoldings {

    @Override
    public TitleHoldingsAlephXMLClobsPool newPool(Settings settings) throws IOException {
        return new TitleHoldingsAlephXMLClobsPool(settings);
    }

    public class TitleHoldingsAlephXMLClobsPool extends TitleHoldingsWorkerPool {

        protected TitleHoldingsAlephXMLClobsPool(Settings settings) throws IOException {
            super(settings);
        }

        @Override
        public TitleHoldingsAlephXMLClobsWorker newWorker(String name, int count) throws IOException {
            return new TitleHoldingsAlephXMLClobsWorker(settings, this);
        }
    }

    public class TitleHoldingsAlephXMLClobsWorker extends TitleHoldingsWorker {

        private final MarcContentHandler contentHandler;

        private final MarcFieldTransformers marcFieldTransformers;

        private final MarcValueTransformers marcValueTransformers;

        private final int bufferSize;

        private final SAXParser saxParser;

        protected TitleHoldingsAlephXMLClobsWorker(Settings settings, TitleHoldingsWorkerPool pool) throws IOException {
            super(settings, pool);
            this.contentHandler = new MarcContentHandler();
            contentHandler.addNamespace("http://www.ddb.de/professionell/mabxml/mabxml-1.xsd");
            contentHandler.setFormat("AlephXML");
            contentHandler.setType("Bibliographic");
            // set up field transformers
            this.marcFieldTransformers = new MarcFieldTransformers();
            ErrorFields errorFields = new ErrorFields();
            errorFields.install(settings, marcFieldTransformers);
            SkipFields skipFields = new SkipFields();
            skipFields.install(settings, marcFieldTransformers);
            TransformFields transformFields = new TransformFields();
            transformFields.install(settings, marcFieldTransformers, marcFieldTransformers, marcFieldTransformers);
            // set up field value transformers
            this.marcValueTransformers = new MarcValueTransformers();
            ISILTransformer isilTransformer = new ISILTransformer(settings);
            isilTransformer.createTransformer(settings, marcValueTransformers);
            bufferSize = settings.getAsInt("buffersize", 65536);
            try {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                factory.setNamespaceAware(true);
                saxParser = factory.newSAXParser();
            } catch (ParserConfigurationException | SAXException e) {
                throw new IOException(e);
            }
        }

        @Override
        protected void doExecute(InputStream inputStream,
                                 CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool)
                throws IOException {
            // the MARCRecordAdapter does the magic
            contentHandler.setMarcListener(new MarcRecordAdapter(catalogEntityWorkerPool, null));
            Marc marc = Marc.builder()
                    .setMarcFieldTransformers(marcFieldTransformers)
                    .setMarcValueTransformers(marcValueTransformers)
                    .setContentHandler(contentHandler)
                    .build();
            try (BufferedReader bufferedReader =
                         new BufferedReader(new InputStreamReader(inputStream, StandardCharsets.UTF_8), bufferSize)) {
                bufferedReader.lines().forEach(line -> {
                    try {
                        saxParser.reset();
                        marc.xmlReader().setSaxParser(saxParser).parse(new InputSource(new StringReader(line)));
                    } catch (IOException e) {
                        throw new UncheckedIOException(e);
                    }
                });
            }
        }
    }
}
