package org.xbib.interlibrary.catalog.elasticsearch.hbz;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.nio.charset.StandardCharsets;
import java.text.MessageFormat;
import java.time.Instant;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.XMLConstants;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.interlibrary.catalog.entities.ErrorFields;
import org.xbib.interlibrary.catalog.entities.Facet;
import org.xbib.interlibrary.catalog.entities.ISILTransformer;
import org.xbib.interlibrary.catalog.entities.SkipFields;
import org.xbib.interlibrary.catalog.entities.TransformFields;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecordAdapter;
import org.xbib.marc.transformer.field.MarcFieldTransformers;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.settings.Settings;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class XMLClobs extends ElasticsearchTool<String> {

    private static final Logger logger = Logger.getLogger(XMLClobs.class.getName());

    private static final AtomicLong creations = new AtomicLong();

    private static final AtomicLong deletions = new AtomicLong();

    public XMLClobs() {
    }

    @Override
    public XMLClobsWorkerPool newPool(Settings settings) throws IOException {
        return new XMLClobsWorkerPool(settings);
    }

    @Override
    public void close() throws IOException {
        super.close();
        logger.log(Level.INFO, "creations = " + creations.get() + " deletions = " + deletions.get());
    }

    public static class XMLClobsWorkerPool extends ElasticsearchWorkerPool<String> {

        private final Logger logger = Logger.getLogger(XMLClobsWorkerPool.class.getName());

        protected ISILTransformer isilTransformer;

        protected MarcValueTransformers marcValueTransformers;

        protected CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        public XMLClobsWorkerPool(Settings settings) throws IOException {
            super("xml-clobs", settings);
            try {
                isilTransformer = new ISILTransformer(settings);
                marcValueTransformers = new MarcValueTransformers();
                isilTransformer.createTransformer(settings, marcValueTransformers);
                catalogEntityWorkerPool = new CatalogEntityWorkerPool<>(settings,
                        new DefaultWorkerPoolMarcRecordListener(), XmlClobsEntityBuilder::new, this);
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        @Override
        public void openInput() throws IOException, InterruptedException {
            super.openInput();
            for (Map.Entry<String, List<String>> me : getConverterInput().getFileMap().entrySet()) {
                if ("lines".equals(me.getKey())) {
                    for (String uri : me.getValue()) {
                        logger.log(Level.INFO, "reading JSON lines from input file: " + uri);
                        InputStream inputStream = ConverterInput.getInputStream(uri, settings);
                        int bufferSize = settings.getAsInt("buffersize", 65536);
                        try (BufferedReader bufferedReader =
                                     new BufferedReader(new InputStreamReader(inputStream, StandardCharsets.UTF_8), bufferSize)) {
                            bufferedReader.lines().forEach(line -> submit(newWorkerRequest(line, uri)));
                        }
                    }
                }
            }
        }

        @Override
        public XMLClobsWorker newWorker(String name, int count) throws IOException {
            return new XMLClobsWorker(settings);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        public class XMLClobsWorker extends ElasticsearchWorker<String, XMLClobsWorkerPool> {

            protected MarcFieldTransformers marcFieldTransformers;

            protected MarcContentHandler contentHandler;

            protected SAXParser saxParser;

            public XMLClobsWorker(Settings settings) throws IOException {
                super(settings);
                this.contentHandler = new MarcContentHandler();
                contentHandler.addNamespace("http://www.ddb.de/professionell/mabxml/mabxml-1.xsd");
                contentHandler.setFormat("AlephXML");
                contentHandler.setType("Bibliographic");
                this.marcFieldTransformers = new MarcFieldTransformers();
                ErrorFields errorFields = new ErrorFields();
                errorFields.install(settings, marcFieldTransformers);
                SkipFields skipFields = new SkipFields();
                skipFields.install(settings, marcFieldTransformers);
                TransformFields transformFields = new TransformFields();
                transformFields.install(settings, marcFieldTransformers, marcFieldTransformers, marcFieldTransformers);
                // the MARCRecordAdapter does the magic
                contentHandler.setMarcListener(new MarcRecordAdapter(catalogEntityWorkerPool, null));
                try {
                    SAXParserFactory parserFactory = SAXParserFactory.newInstance();
                    parserFactory.setFeature("http://xml.org/sax/features/namespaces", true);
                    parserFactory.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
                    parserFactory.setFeature("http://xml.org/sax/features/validation", false);
                    parserFactory.setFeature("http://xml.org/sax/features/external-general-entities", false);
                    parserFactory.setFeature("http://xml.org/sax/features/external-parameter-entities", false);
                    parserFactory.setFeature("http://apache.org/xml/features/nonvalidating/load-dtd-grammar", false);
                    parserFactory.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);
                    parserFactory.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
                    parserFactory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);
                    parserFactory.setXIncludeAware(false);
                    parserFactory.setNamespaceAware(true);
                    saxParser = parserFactory.newSAXParser();
                } catch (ParserConfigurationException | SAXException e) {
                    throw new IOException(e);
                }
            }

            @Override
            protected void doExecute(String xmlClob) throws IOException {
                if (xmlClob != null) {
                    saxParser.reset();
                    Marc marc = Marc.builder()
                            .setMarcFieldTransformers(marcFieldTransformers)
                            .setMarcValueTransformers(marcValueTransformers)
                            .setContentHandler(contentHandler)
                            .build();
                    marc.xmlReader().setSaxParser(saxParser).parse(new InputSource(new StringReader(xmlClob)));
                }
            }
        }

        protected class XmlClobsEntityBuilder extends CatalogEntityBuilder {

            private final Logger logger = Logger.getLogger(XmlClobsEntityBuilder.class.getName());

            private final boolean create;

            private XmlClobsEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
                this.create = settings.getAsBoolean("create", false);
                for (Map.Entry<String, List<String>> entry : getConverterInput().getFileMap().entrySet()) {
                    String key = entry.getKey();
                    if (key.startsWith("taxonomy-")) {
                        String isil = key.substring("taxonomy-".length());
                        for (String taxFile : entry.getValue()) {
                            logger.log(Level.INFO, MessageFormat.format("adding classifier file from {0} for ISIL {1}",
                                    taxFile, isil));
                            addClassifier("(" + settings.get("catalogid") + ")", isil, taxFile);
                        }
                    }
                }
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                // write title resource
                final ElasticsearchIndexDefinition titleDef = getOutputIndexDefinitionMap().get("title");
                if (titleDef == null) {
                    throw new IllegalArgumentException("no 'title' index definition");
                }
                RouteRdfXContentParams params = new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                        titleDef.getFullIndexName(),
                        titleDef.getType());
                params.setHandler((content, p) -> {
                    if (state.getResource().isDeleted()) {
                        delete(titleDef, state.getRecordIdentifier());
                        deletions.incrementAndGet();
                    } else {
                        bulkIndex(titleDef, state.getRecordIdentifier(), create, content);
                        creations.incrementAndGet();
                    }
                });
                if (settings.get("format") != null) {
                    state.getResource().add("format", settings.get("format"));
                }
                if (settings.get("type") != null) {
                    state.getResource().add("type", settings.get("type"));
                }
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("xbib")
                        .add("uid", state.getRecordIdentifier());
                state.getResource().newResource("rec")
                        .add("created", catalogEntityWorkerPool.getCreatedFormatted())
                        .add("modified", catalogEntityWorkerPool.getLastModifiedFormatted())
                        .add("indexed", Instant.now().toString())
                        .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                    builder.receive(state.getResource());
                }
                // write holdings resources
                final ElasticsearchIndexDefinition holDef = getOutputIndexDefinitionMap().get("holdings");
                if (holDef == null) {
                    // silently ignore, useful when indexing MARC
                    return;
                }
                String holdingsIndex = holDef.getFullIndexName();
                String holdingsIndexType = holDef.getType();
                Iterator<Resource> it = state.getResourceIterator();
                while (it.hasNext()) {
                    Resource resource = it.next();
                    if (resource.equals(state.getResource())) {
                        // skip main resource
                        continue;
                    }
                    params = new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                            holdingsIndex,
                            holdingsIndexType);
                    params.setHandler((content, p) ->
                            bulkIndex(holDef, state.getRecordIdentifier() + "." + resource.id(), false, content));
                    resource.newResource("xbib")
                            .add("uid", state.getRecordIdentifier());
                    state.getResource().newResource("rec")
                            .add("created", catalogEntityWorkerPool.getCreatedFormatted())
                            .add("modified", catalogEntityWorkerPool.getLastModifiedFormatted())
                            .add("indexed", Instant.now().toString())
                            .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                    try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                        builder.receive(resource);
                    }
                }
                // create more indices
                for (Map.Entry<String, ElasticsearchIndexDefinition> entry : getOutputIndexDefinitionMap().entrySet()) {
                    if (!"title".equals(entry.getKey()) && !"holdings".equals(entry.getKey())) {
                        createIndexByKey(entry.getKey(), entry.getValue(), state);
                    }
                }
            }

            protected void createIndexByKey(String key,
                                            ElasticsearchIndexDefinition indexDefinition,
                                            CatalogEntityBuilderState state)
                    throws IOException {
                // check holdings facet for filtering ISIL
                Facet<String> holdingsFacet = state.getFacets().get("xbib.identifier");
                if (holdingsFacet == null) {
                    logger.log(Level.WARNING, "no holdings facet found: {}", state.getFacets());
                    return;
                }
                if (holdingsFacet.getValues().contains(key)) {
                    // join main resource with matching index name
                    Resource resource = state.getResource();
                    Iterator<Resource> it = state.getResourceIterator();
                    while (it.hasNext()) {
                        Resource r = it.next();
                        if (r.equals(resource)) {
                            // skip main resource
                            continue;
                        }
                        // holdings document ID starts with ISIL
                        if (r.id().toString().startsWith(key + "/")) {
                            resource.add("Holdings", r);
                        }
                    }
                    RouteRdfXContentParams params = new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                            indexDefinition.getFullIndexName(),
                            indexDefinition.getType());
                    params.setHandler((content, p) ->
                            bulkIndex(indexDefinition, resource.id().toString(), false, content));
                    try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                        builder.receive(resource);
                    }
                }
            }
        }
    }
}
