package org.xbib.interlibrary.catalog.elasticsearch.hbzalmagvi;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;
import java.time.Instant;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.io.archive.tar.TarArchiveEntry;
import org.xbib.io.archive.tar.TarArchiveInputStream;
import org.xbib.marc.Marc;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.settings.Settings;

public class HbzAlmaGvi extends ElasticsearchTool<String> {

    private static final Logger logger = Logger.getLogger(HbzAlmaGvi.class.getName());

    private static final AtomicLong creations = new AtomicLong();

    private static final AtomicLong deletions = new AtomicLong();

    public HbzAlmaGvi() {
    }

    @Override
    public HbzAlmaGviWorkerPool newPool(Settings settings) throws IOException {
        return new HbzAlmaGviWorkerPool(settings);
    }

    @Override
    public void close() throws IOException {
        super.close();
        logger.log(Level.INFO, "creations = " + creations.get() + " deletions = " + deletions.get());
    }

    public class HbzAlmaGviWorkerPool extends ElasticsearchWorkerPool<String> {

        public HbzAlmaGviWorkerPool(Settings settings) throws IOException {
            super("hbz-alma-gvi", settings);
        }

        @Override
        public HbzAlmaGviWorker newWorker(String name, int count) throws IOException {
            return new HbzAlmaGviWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }

    public class HbzAlmaGviWorker extends ElasticsearchWorker<String, HbzAlmaGviWorkerPool> {

        private final Logger logger = Logger.getLogger(HbzAlmaGviWorker.class.getName());

        private final HbzAlmaGviWorkerPool pool;

        private final Set<String> logIdentifiers;

        private CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        public HbzAlmaGviWorker(Settings settings, HbzAlmaGviWorkerPool pool) {
            super(settings);
            this.pool = pool;
            this.logIdentifiers = settings.getAsArray("logids") != null ? Set.of(settings.getAsArray("logids")) : null;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            catalogEntityWorkerPool = new CatalogEntityWorkerPool<>(settings, new DefaultWorkerPoolMarcRecordListener(),
                    HbzAlmaCatalogEntityBuilder::new, pool);
            InputStream tarInputStream = ConverterInput.getInputStream(uri, settings);
            Objects.requireNonNull(tarInputStream);
            TarArchiveInputStream tarArchiveInputStream = new TarArchiveInputStream(tarInputStream);
            TarArchiveEntry entry;
            while ((entry = tarArchiveInputStream.getNextEntry()) != null) {
                logger.log(Level.INFO, "processing entry = " + entry.getName());
                catalogEntityWorkerPool.setRequestOrigin(entry.getName());
                Path tempFile = Files.createTempFile(entry.getName(), "xml");
                Files.copy(tarArchiveInputStream, tempFile, StandardCopyOption.REPLACE_EXISTING);
                long size = Files.size(tempFile);
                logger.log(Level.INFO, "size = " + size);
                try (InputStream inputStream = Files.newInputStream(tempFile)) {
                    MarcContentHandler marcContentHandler = new MarcContentHandler();
                    marcContentHandler.setMarcRecordListener(catalogEntityWorkerPool);
                    Marc.builder()
                            .setInputStream(inputStream)
                            .setCharset(StandardCharsets.UTF_8)
                            .setContentHandler(marcContentHandler)
                            .build()
                            .xmlReader()
                            .parse();
                } finally {
                    Files.delete(tempFile);
                }
            }
            tarArchiveInputStream.close();
            catalogEntityWorkerPool.close();
        }

        protected class HbzAlmaCatalogEntityBuilder extends CatalogEntityBuilder {

            protected HbzAlmaCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                ElasticsearchIndexDefinition elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
                if (elasticsearchIndexDefinition == null) {
                    throw new IOException("no 'bib' index definition configured");
                }
                RouteRdfXContentParams routeRdfXContentParams =
                        new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                                elasticsearchIndexDefinition.getFullIndexName(),
                                elasticsearchIndexDefinition.getType());
                routeRdfXContentParams.setHandler((content, p) -> {
                    if (state.getRecordIdentifier() != null) {
                        if (logIdentifiers != null && logIdentifiers.contains(state.getRecordIdentifier())) {
                            logger.log(Level.INFO, "id = " + state.getRecordIdentifier() + " content = " + content);
                        }
                        String origin = catalogEntityWorkerPool.getRequestOrigin();
                        if (origin != null && origin.contains("delete")) {
                            pool.delete(elasticsearchIndexDefinition, state.getRecordIdentifier());
                            deletions.incrementAndGet();
                        } else {
                            pool.bulkIndex(elasticsearchIndexDefinition, state.getRecordIdentifier(), false, content);
                            creations.incrementAndGet();
                        }
                    }
                });
                if (settings.get("format") != null) {
                    state.getResource().add("format", settings.get("format"));
                }
                if (settings.get("type") != null) {
                    state.getResource().add("type", settings.get("type"));
                }
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("rec")
                        .add("created", catalogEntityWorkerPool.getCreatedFormatted())
                        .add("modified", catalogEntityWorkerPool.getLastModifiedFormatted())
                        .add("indexed", Instant.now().toString())
                        .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(state.getResource());
                }
            }
        }
    }
}

