package org.xbib.interlibrary.catalog.elasticsearch.hbzfix;

import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.MessageFormat;
import java.time.Instant;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.xml.util.XMLUtil;
import org.xbib.interlibrary.api.action.worker.WorkerPool;
import org.xbib.interlibrary.api.action.worker.WorkerPoolListener;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.Finder;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderProvider;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecord;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.marc.xml.MarcXmlWriter;
import org.xbib.settings.Settings;

public class HbzFix extends ElasticsearchTool<String> {

    private static final AtomicLong creations = new AtomicLong();

    private static final AtomicLong deletions = new AtomicLong();

    private static final Set<String> recordedIdentifiers = new ConcurrentSkipListSet<>();

    public HbzFix() {
    }

    @Override
    public HbzFixWorkerPool newPool(Settings settings) throws IOException {
        return new HbzFixWorkerPool(settings);
    }

    public class HbzFixWorkerPool extends ElasticsearchWorkerPool<String> {

        private final Logger logger = Logger.getLogger(HbzFixWorkerPool.class.getName());

        private MarcXmlWriter marcWriter;

        public HbzFixWorkerPool(Settings settings) throws IOException {
            super("hbz-fix", settings);
        }

        @Override
        public HbzFixWorker newWorker(String name, int count) throws IOException {
            return new HbzFixWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        @Override
        protected void openOutput() throws IOException {
            super.openOutput();
            Map<String, Settings> outputMap = settings.getGroups("output");
            for (Map.Entry<String, Settings> entry : outputMap.entrySet()) {
                String entryKey = entry.getKey();
                Settings entrySettings = entry.getValue();
                if ("file".equals(entryKey)) {
                    if ("true".equals(entrySettings.get("enabled"))) {
                        String archivePath = entrySettings.get("path");
                        if (archivePath != null) {
                            logger.log(Level.INFO, "archive path = " + archivePath);
                            performRetention(entrySettings, Paths.get(archivePath));
                            boolean compressed = archivePath.endsWith(".gz");
                            MarcValueTransformers marcValueTransformers = new MarcValueTransformers();
                            marcValueTransformers.setMarcValueTransformer(XMLUtil::sanitizeXml10);
                            int splitLimit = entrySettings.getAsInt("splitlimit", -1);
                            int bufferSize = entrySettings.getAsInt("buffersize", 65536);
                            boolean indent = entrySettings.getAsBoolean("indent", true);
                            marcWriter = new MarcXmlWriter(archivePath, splitLimit, bufferSize, compressed, indent);
                            marcWriter.setMarcValueTransformers(marcValueTransformers);
                            marcWriter.startDocument();
                            marcWriter.beginCollection();
                        }
                    }
                }
            }
        }

        @Override
        protected void closeOutput(int returnCode) throws IOException {
            super.closeOutput(returnCode);
            if (marcWriter != null) {
                marcWriter.endCollection();
                marcWriter.endDocument();
                marcWriter.close();
            }
        }

        private void performRetention(Settings settings, Path path) throws IOException {
            settings = settings.getAsSettings("retention");
            String basepattern = settings.get("basepattern", "*");
            String pattern = settings.get("pattern", "*");
            long minkeep = settings.getAsLong("minkeep", 1L);
            logger.log(Level.INFO, MessageFormat.format("found retention policy: parent={0} basepattern={1} pattern={2} minkeep={3}",
                    path.getParent(), basepattern, pattern, minkeep));
            // sort by last modified date in descending order and skip min-to-keep entries
            Queue<Finder.PathFile> paths = new Finder()
                    .find(path.getParent(), basepattern, path, pattern)
                    .sortBy(settings.get("sort_by", "lastmodified"))
                    .order(settings.get("order", "desc"))
                    .skipPathFiles(minkeep);
            logger.log(Level.INFO, MessageFormat.format("files to delete due to retention policy: {0}", paths));
            for (Finder.PathFile pf : paths) {
                try {
                    Files.deleteIfExists(pf.getPath());
                } catch (IOException e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                }
            }
        }
    }

    public class HbzFixWorker extends ElasticsearchWorker<String, HbzFixWorkerPool> {

        CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        private final Set<String> logIdentifiers;

        private final HbzFixWorkerPool pool;

        public HbzFixWorker(Settings settings, HbzFixWorkerPool pool) {
            super(settings);
            this.pool = pool;
            this.logIdentifiers = settings.getAsArray("logids") != null ? Set.of(settings.getAsArray("logids")) : null;
        }

        @Override
        protected void doExecute(String request) throws IOException {
            try (InputStream inputStream = ConverterInput.getInputStream(request, settings)) {
                 catalogEntityWorkerPool = new HbzFixCatalogEntityWorkerPool(settings, new DefaultWorkerPoolMarcRecordListener(),
                                 HbzFixCatalogEntityBuilder::new, pool);
                MarcContentHandler marcContentHandler = new MarcContentHandler();
                marcContentHandler.setMarcRecordListener(catalogEntityWorkerPool);
                Marc.builder()
                        .setInputStream(inputStream)
                        .setCharset(StandardCharsets.UTF_8)
                        .setContentHandler(marcContentHandler)
                        .build()
                        .xmlReader()
                        .parse();
                catalogEntityWorkerPool.close();
            }
        }

        protected class HbzFixCatalogEntityWorkerPool extends CatalogEntityWorkerPool<CatalogEntityBuilder> {

            private final Logger logger = Logger.getLogger(HbzFixCatalogEntityWorkerPool.class.getName());

            public HbzFixCatalogEntityWorkerPool(Settings settings,
                                                 WorkerPoolListener<WorkerPool<MarcRecord>> listener,
                                                 CatalogEntityBuilderProvider<CatalogEntityBuilder> entityBuilderProvider,
                                                 Closeable closeable) throws IOException {
                super(settings, listener, entityBuilderProvider, closeable);
            }

            @Override
            public void record(final MarcRecord marcRecord) {
                archiveRecord(marcRecord);
                super.record(marcRecord);
            }

            private void archiveRecord(MarcRecord marcRecord) {
                try {
                    if (pool.marcWriter != null && !pool.marcWriter.isClosed()) {
                        pool.marcWriter.record(marcRecord);
                        Exception exception = pool.marcWriter.getException();
                        if (exception != null) {
                            logger.log(Level.SEVERE, "marc write error: " + exception, exception);
                            throw new IOException(exception);
                        }
                    }
                } catch (IOException e) {
                    logger.log(Level.WARNING, e.getMessage(), e);

                }
            }

            private static final LocalDate BASE_DATE = LocalDate.of(1970, 1, 1);

            private static final DateTimeFormatter dateFormat = DateTimeFormatter.ofPattern("yyyyMMdd");

        }

        protected class HbzFixCatalogEntityBuilder extends CatalogEntityBuilder {

            private final Logger logger = Logger.getLogger(HbzFixCatalogEntityBuilder.class.getName());

            protected HbzFixCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                ElasticsearchIndexDefinition elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
                if (elasticsearchIndexDefinition == null) {
                    throw new IOException("no 'bib' index definition configured");
                }
                RouteRdfXContentParams routeRdfXContentParams =
                        new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                                elasticsearchIndexDefinition.getFullIndexName(),
                                elasticsearchIndexDefinition.getType());
                String recordIdentifier = clean(state.getRecordIdentifier());
                if (recordedIdentifiers.contains(recordIdentifier)) {
                    logger.log(Level.WARNING, "double record ID = " + recordIdentifier);
                }
                recordedIdentifiers.add(recordIdentifier);
                routeRdfXContentParams.setHandler((content, p) -> {
                    if (!recordIdentifier.isEmpty()) {
                        if (logIdentifiers != null && logIdentifiers.contains(recordIdentifier)) {
                            logger.log(Level.INFO, "id = " + recordIdentifier + " content = " + content);
                        }
                        String origin = catalogEntityWorkerPool.getRequestOrigin();
                        if (origin != null && origin.contains("delete")) {
                            pool.delete(elasticsearchIndexDefinition, recordIdentifier);
                            deletions.incrementAndGet();
                        } else {
                            pool.bulkIndex(elasticsearchIndexDefinition, recordIdentifier, false, content);
                            creations.incrementAndGet();
                        }
                    }
                });
                if (settings.get("format") != null) {
                    state.getResource().add("format", settings.get("format"));
                }
                if (settings.get("type") != null) {
                    state.getResource().add("type", settings.get("type"));
                }
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("rec")
                        .add("created", catalogEntityWorkerPool.getCreatedFormatted())
                        .add("modified", catalogEntityWorkerPool.getLastModifiedFormatted())
                        .add("indexed", Instant.now().toString())
                        .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(state.getResource());
                }
            }
        }

        /**
         * Clean record identifier, accept only printable ASCII characters, at most 32.
         * This is sane for Elasticsearch document IDs.
         * @param string the record identifiervalue
         * @return the clean value
         */
        private static String clean(String string) {
            StringBuilder sb = new StringBuilder();
            if (string != null) {
                for (char ch : string.toCharArray()) {
                    if (ch < 32 || ch > 127) {
                        break;
                    }
                    if (sb.length() > 31) {
                        break;
                    }
                    sb.append(ch);
                }
            }
            return sb.toString();
        }
    }
}
