package org.xbib.interlibrary.catalog.elasticsearch.lvi;

import java.io.IOException;
import java.io.StringWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.MessageFormat;
import java.time.Instant;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilder;
import org.xbib.content.xml.util.XMLUtil;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.elx.api.SearchDocument;
import org.xbib.interlibrary.api.action.worker.AbstractWorkerPool;
import org.xbib.interlibrary.api.action.worker.Tool;
import org.xbib.interlibrary.api.action.worker.Worker;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.api.library.DefaultIndexLibrary;
import org.xbib.interlibrary.api.library.IndexLibrary;
import org.xbib.interlibrary.catalog.builder.Finder;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.ElxInputHelper;
import org.xbib.interlibrary.catalog.elasticsearch.ElxOutputHelper;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.ConfigurationArguments;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.DefaultConfigurationArguments;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.ElasticsearchConfiguration;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.EzbHoldings;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.HbzAlephHoldings;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.HbzAlephRecord;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.HbzAlmaHoldings;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.HbzAlmaItems;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.HbzAlmaOnlineResource;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.HbzAlmaRecord;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.HbzfixHoldings;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.HbzfixRecord;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.LviRecord;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.ZdbHoldings;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.record.ZdbRecord;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.service.ServiceManager;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics.ConfigurationTable;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics.ProductStatistics;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics.RecordStatistics;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics.ResourceStatistics;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics.ResourceStatisticsTable;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics.StatusCodeMapStatistics;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;
import org.xbib.marc.json.JsonBuilder;
import org.xbib.marc.json.MarcJsonWriter;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.marc.xml.MarcXmlWriter;
import org.xbib.settings.Settings;
import static org.elasticsearch.index.query.QueryBuilders.boolQuery;
import static org.elasticsearch.index.query.QueryBuilders.matchAllQuery;
import static org.elasticsearch.index.query.QueryBuilders.termQuery;

public class Lvi implements Tool {

    private static final Logger logger = Logger.getLogger(Lvi.class.getName());

    private static final Set<String> privateTags = Set.of("rec",
            "created", "modified", "indexed", "uid", "origin");

    private static final DateTimeFormatter dateFormat = DateTimeFormatter.ofPattern("yyyyMMdd");

    private static final LocalDate BASE_DATE = LocalDate.of(1970, 1, 1);

    private final AtomicBoolean closed;

    private final AtomicInteger workerCount;

    private LviWorkerPool workerPool;

    private boolean enabled;

    private volatile int returncode;

    public Lvi() {
        closed = new AtomicBoolean();
        workerCount = new AtomicInteger();
    }

    @Override
    public void run(Settings settings) throws IOException {
        this.enabled = settings.getAsBoolean("enabled", false);
        try {
            returncode = 0;
            workerPool = new LviWorkerPool(settings);
            // first: open output, input, then open worker pool
            workerPool.openOutput();
            workerPool.openInput();
            workerPool.open();
            workerPool.close();
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            returncode = 1;
        } finally {
            close();
        }
    }

    @Override
    public int getReturnCode() {
        return returncode;
    }

    @Override
    public void close() throws IOException {
        if (closed.compareAndSet(false, true)) {
            if (workerPool != null) {
                workerPool.close();
                workerPool.closeOutput(returncode);
                workerPool.closeInput();
                for (Map.Entry<Runnable, Throwable> entry : workerPool.getExceptions().entrySet()) {
                    Runnable r = entry.getKey();
                    Throwable t = entry.getValue();
                    logger.log(Level.SEVERE, MessageFormat.format("{0}: {1}", r, t.getMessage()), t);
                    returncode = 1;
                }
            }
        }
    }

    protected class LviWorkerPool extends AbstractWorkerPool<SearchDocument> {

        private final Logger logger = Logger.getLogger(LviWorkerPool.class.getName());

        private final Settings settings;

        private final ElxInputHelper elasticsearchInput;

        private final ElxOutputHelper elasticsearchOutput;

        private IndexDefinition sourceZdbIndexDefinition;

        private IndexDefinition sourceZdbHoldingsIndexDefinition;

        private IndexDefinition sourceEzbHoldingsIndexDefinition;

        private IndexDefinition sourceHbzAlmaIndexDefinition;

        private IndexDefinition sourceHbzAlephIndexDefinition;

        private IndexDefinition sourceHbzfixIndexDefinition;

        private IndexDefinition sourceDoajIndexDefinition;

        private IndexDefinition targetIndexDefinition;

        private IndexDefinition targetProductIndexDefinition;

        private IndexDefinition targetStatisticsIndexDefinition;

        private IndexDefinition targetConfigurationsIndexDefinition;

        private IndexDefinition targetServicesIndexDefinition;

        private final RecordStatistics recordStatistics;

        private final ResourceStatistics zdbResourceStatistics;

        private final ResourceStatistics hbzalmaResourceStatistics;

        private final ResourceStatistics hbzalephResourceStatistics;

        private final ResourceStatistics hbzfixResourceStatistics;

        private final ResourceStatistics lviResourceStatistics;

        private final ProductStatistics lviProductStatistics;

        private final StatusCodeMapStatistics hbzalmaStatusCodeMapStatistics;

        private long sourceZdbIndexCount;

        private long sourceHbzAlephIndexCount;

        private long sourceHbzAlmaIndexCount;

        private long sourceHbzFixIndexCount;

        private long sourceDoajIndexCount;

        private final AtomicInteger maxCount;

        private final AtomicInteger zdbAlephMatch;

        private final AtomicInteger zdbAlmaMatch;

        private final AtomicInteger zdbAlephHCount;

        private final AtomicInteger zdbAlephUCount;

        private final AtomicInteger zdbDeleted;

        private final AtomicInteger hbzfixDeleted;

        private final AtomicInteger hbzalmaDeleted;

        private Configuration configuration;

        private MarcXmlWriter marcWriter;

        protected LviWorkerPool(Settings settings) throws IOException {
            super(settings.get("name", "lvipool"),
                    settings.getAsInt("workers", Runtime.getRuntime().availableProcessors()), null);
            this.settings = settings;
            Settings inputElasticsearchSettings = settings.getGroups("input").get("elasticsearch");
            boolean isInput = inputElasticsearchSettings != null && !inputElasticsearchSettings.getAsMap().isEmpty();
            this.elasticsearchInput = isInput ? new ElxInputHelper(inputElasticsearchSettings) : null;
            Settings outputElasticsearchSettings = settings.getGroups("output").get("elasticsearch");
            boolean isOutput = outputElasticsearchSettings != null && !outputElasticsearchSettings.getAsMap().isEmpty();
            this.elasticsearchOutput = isOutput ? new ElxOutputHelper(outputElasticsearchSettings) : null;
            Objects.requireNonNull(this.elasticsearchInput);
            Objects.requireNonNull(this.elasticsearchOutput);
            this.recordStatistics = new RecordStatistics(settings, "global");
            this.zdbResourceStatistics = new ResourceStatistics(settings, "zdb");
            this.hbzalephResourceStatistics = new ResourceStatistics(settings, "hbzaleph");
            this.hbzalmaResourceStatistics = new ResourceStatistics(settings, "hbzalma");
            this.hbzfixResourceStatistics = new ResourceStatistics(settings, "hbzfix");
            this.lviResourceStatistics = new ResourceStatistics(settings, "lvi");
            this.lviProductStatistics = new ProductStatistics(settings);
            this.hbzalmaStatusCodeMapStatistics = new StatusCodeMapStatistics();
            this.maxCount = new AtomicInteger();
            this.zdbAlephMatch = new AtomicInteger();
            this.zdbAlephHCount = new AtomicInteger();
            this.zdbAlephUCount = new AtomicInteger();
            this.zdbAlmaMatch = new AtomicInteger();
            this.zdbDeleted = new AtomicInteger();
            this.hbzfixDeleted = new AtomicInteger();
            this.hbzalmaDeleted = new AtomicInteger();
        }

        protected void openInput() throws IOException {
            logger.log(Level.INFO, "opening input definitions");
            Map<String, Settings> inputMap = settings.getGroups("input");
            for (Map.Entry<String, Settings> entry : inputMap.entrySet()) {
                if ("elasticsearch".equals(entry.getKey())) {
                    elasticsearchInput.setIndexDefinitions(entry.getValue().getGroups("index"));
                }
            }
            if (elasticsearchInput.getIndexDefinitions().isEmpty()) {
                throw new IllegalArgumentException("no input definitions found");
            } else {
                logger.log(Level.INFO, "input definitions = " + elasticsearchInput.getIndexDefinitions());
            }
            IndexDefinition sourceBibIndexDefinition = elasticsearchInput.getIndexDefinitions().get("bib");
            checkIndex(sourceBibIndexDefinition);
            ConfigurationArguments arguments =
                    new DefaultConfigurationArguments(settings.getAsStructuredMap(),
                            elasticsearchInput.getSearchClient().getClient());
            this.configuration = new ElasticsearchConfiguration(arguments,
                    elasticsearchInput.getSearchClient());
            logger.log(Level.INFO, "index bib found and " + configuration.getLibraries().size() + " libraries configured");

            this.sourceZdbIndexDefinition = elasticsearchInput.getIndexDefinitions().get("zdb");
            this.sourceZdbIndexCount = checkIndex(sourceZdbIndexDefinition);
            logger.log(Level.INFO, "index zdb found with " + sourceZdbIndexCount + " docs");

            this.sourceZdbHoldingsIndexDefinition = elasticsearchInput.getIndexDefinitions().get("zdbholdings");
            long sourceZdbHoldingsIndexCount = checkIndex(sourceZdbHoldingsIndexDefinition);
            logger.log(Level.INFO, "index zdbholdings found with " + sourceZdbHoldingsIndexCount + " docs");

            this.sourceEzbHoldingsIndexDefinition = elasticsearchInput.getIndexDefinitions().get("ezb");
            long sourceEzbHoldingsIndexCount = checkIndex(sourceEzbHoldingsIndexDefinition);
            logger.log(Level.INFO, "index ezb found with " + sourceEzbHoldingsIndexCount + " docs");

            this.sourceHbzAlmaIndexDefinition = elasticsearchInput.getIndexDefinitions().get("hbzalma");
            this.sourceHbzAlmaIndexCount = checkIndex(sourceHbzAlmaIndexDefinition);
            logger.log(Level.INFO, "index hbzalma found with " + sourceHbzAlmaIndexCount + " docs");

            this.sourceHbzAlephIndexDefinition = elasticsearchInput.getIndexDefinitions().get("hbzaleph");
            this.sourceHbzAlephIndexCount = checkIndex(sourceHbzAlephIndexDefinition);
            logger.log(Level.INFO, "index hbzaleph found with " + sourceHbzAlephIndexCount + " docs");

            this.sourceHbzfixIndexDefinition = elasticsearchInput.getIndexDefinitions().get("hbzfix");
            this.sourceHbzFixIndexCount = checkIndex(sourceHbzfixIndexDefinition);
            logger.log(Level.INFO, "index hbzfix found with " + sourceHbzFixIndexCount + " docs");

            this.sourceDoajIndexDefinition = elasticsearchInput.getIndexDefinitions().get("doaj");
            this.sourceDoajIndexCount = checkIndex(sourceDoajIndexDefinition);
            logger.log(Level.INFO, "index doaj found with " + sourceDoajIndexCount + " docs");

            logger.log(Level.INFO, "all inputs found and open");
        }

        private long checkIndex(IndexDefinition indexDefinition) {
            if (indexDefinition == null) {
                throw new IllegalArgumentException("no index definition present");
            }
            long l = elasticsearchInput.getSearchClient().getSearchableDocs(indexDefinition);
            if (l <=0 ) {
                throw new IllegalStateException("no index or no documents in index: " + indexDefinition.getFullIndexName());
            } else {
                logger.log(Level.INFO, "found searchable docs in " + indexDefinition.getFullIndexName() + " = " + l);
            }
            return l;
        }

        protected void openOutput() throws IOException {
            Map<String, Settings> outputMap = settings.getGroups("output");
            for (Map.Entry<String, Settings> entry : outputMap.entrySet()) {
                String entryKey = entry.getKey();
                Settings entrySettings = entry.getValue();
                if ("elasticsearch".equals(entryKey)) {
                    elasticsearchOutput.setIndexDefinitions(entrySettings.getGroups("index"));
                }
                if ("file".equals(entryKey)) {
                    if ("true".equals(entrySettings.get("enabled"))) {
                        String archivePath = entrySettings.get("path");
                        if (archivePath != null) {
                            logger.log(Level.INFO, "archive path = " + archivePath);
                            performRetention(entrySettings, Paths.get(archivePath));
                            boolean compressed = archivePath.endsWith(".gz");
                            MarcValueTransformers marcValueTransformers = new MarcValueTransformers();
                            marcValueTransformers.setMarcValueTransformer(XMLUtil::sanitizeXml10);
                            int splitLimit = entrySettings.getAsInt("splitlimit", -1);
                            int bufferSize = entrySettings.getAsInt("buffersize", 65536);
                            boolean indent = entrySettings.getAsBoolean("indent", true);
                            marcWriter = new MarcXmlWriter(archivePath, splitLimit, bufferSize, compressed, indent);
                            marcWriter.setMarcValueTransformers(marcValueTransformers);
                            marcWriter.startDocument();
                            marcWriter.beginCollection();
                        }
                    }
                }
            }
            if (elasticsearchOutput.getIndexDefinitions().isEmpty()) {
                throw new IllegalArgumentException("no output definitions found");
            }
            Map<String, ElasticsearchIndexDefinition> map = elasticsearchOutput.getIndexDefinitions();

            this.targetIndexDefinition = map.get("lvi");
            Objects.requireNonNull(this.targetIndexDefinition);

            this.targetServicesIndexDefinition = map.get("services");
            Objects.requireNonNull(this.targetServicesIndexDefinition);

            this.targetStatisticsIndexDefinition = map.get("statistics");
            Objects.requireNonNull(this.targetStatisticsIndexDefinition);

            this.targetProductIndexDefinition = map.get("products");
            Objects.requireNonNull(this.targetProductIndexDefinition);

            this.targetConfigurationsIndexDefinition = map.get("configurations");
            Objects.requireNonNull(this.targetConfigurationsIndexDefinition);

            elasticsearchOutput.createIndices("lvi", "services");
            elasticsearchOutput.startBulk("lvi", "services");
            logger.log(Level.INFO, "outputs created and open");
            maxCount.set(settings.getAsInt("max", 0));
        }

        @Override
        public void open() throws IOException {
            logger.log(Level.INFO, "opening");
            // spawn worker threads
            super.open();
            // process
            if (settings.getAsBoolean("zdb.enabled", false)) {
                processZdb();
            }
            if (settings.getAsBoolean("hbzalma.enabled", false)) {
                processHbzAlma();
            }
            if (settings.getAsBoolean("hbzfix.enabled", false)) {
                processHbzfix();
            }
        }

        protected void closeInput() throws IOException {
            elasticsearchInput.close();
            if (configuration != null && configuration.getLibraries() != null) {
                for (IndexLibrary library : configuration.getLibraries().values()) {
                    if (library instanceof DefaultIndexLibrary defaultLibrary) {
                        if (!defaultLibrary.getUnmappedStatusTexts().isEmpty()) {
                            logger.log(Level.WARNING, "unmapped status codes for " + library.getISIL() + ": " +
                                    defaultLibrary.getUnmappedStatusTexts());
                        }
                    }
                }
            }
        }

        protected void closeOutput(int returncode) throws IOException {
            if (marcWriter != null) {
                marcWriter.endCollection();
                marcWriter.endDocument();
                marcWriter.close();
            }
            elasticsearchOutput.stopBulk("lvi", "services");
            if (returncode == 0) {
                elasticsearchOutput.activateIndex("lvi", "services");
            }
            elasticsearchOutput.createIndices("statistics", "products", "configurations");
            elasticsearchOutput.startBulk("statistics", "products", "configurations");
            writeStatistics();
            writeDefaultConfiguration();
            elasticsearchOutput.stopBulk("statistics", "products", "configurations");
            if (returncode == 0) {
                elasticsearchOutput.activateIndex("statistics", "products", "configurations");
            }

            elasticsearchOutput.close();

            logStatistics();

            logger.log(Level.INFO, "unknown Alma MARC relator codes = " + HbzAlmaRecord.getUnknownRelatorCodes());
            logger.log(Level.INFO, "unknown Alma RDA codes = " + HbzAlmaRecord.getUnknownRdaCodes());
            logger.log(Level.INFO, "unknown Alma sublibrary codes = " +
                    HbzAlmaItems.getUnknownSublibraries() + "; " + HbzAlmaOnlineResource.getUnknownSublibraries());
        }

        private void performRetention(Settings settings, Path path) throws IOException {
            settings = settings.getAsSettings("retention");
            String basepattern = settings.get("basepattern", "*");
            String pattern = settings.get("pattern", "*");
            long minkeep = settings.getAsLong("minkeep", 1L);
            logger.log(Level.INFO, MessageFormat.format("found retention policy: parent={0} basepattern={1} pattern={2} minkeep={3}",
                    path.getParent(), basepattern, pattern, minkeep));
            // sort by last modified date in descending order and skip min-to-keep entries
            Queue<Finder.PathFile> paths = new Finder()
                    .find(path.getParent(), basepattern, path, pattern)
                    .sortBy(settings.get("sort_by", "lastmodified"))
                    .order(settings.get("order", "desc"))
                    .skipPathFiles(minkeep);
            logger.log(Level.INFO, MessageFormat.format("files to delete due to retention policy: {0}", paths));
            for (Finder.PathFile pf : paths) {
                try {
                    Files.deleteIfExists(pf.getPath());
                } catch (IOException e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                }
            }
        }

        private void writeStatistics() throws IOException {
            if (!settings.getAsBoolean("statistics.enabled", false)) {
                return;
            }
            recordStatistics.index(settings, elasticsearchOutput, targetStatisticsIndexDefinition);
            recordStatistics.write("records", "json");

            lviProductStatistics.index(settings, elasticsearchOutput, targetProductIndexDefinition);
            lviProductStatistics.write("products", "json");

            ResourceStatisticsTable resourceStatisticsTable = new ResourceStatisticsTable(configuration, settings,
                    lviResourceStatistics, zdbResourceStatistics, hbzfixResourceStatistics,
                    hbzalephResourceStatistics, hbzalmaResourceStatistics, hbzalmaStatusCodeMapStatistics);
            resourceStatisticsTable.index(settings, elasticsearchOutput, targetStatisticsIndexDefinition);
            resourceStatisticsTable.write("statistics.csv");

            zdbResourceStatistics.write("zdb", "json");
            hbzalmaResourceStatistics.write("hbzalma", "json");
            hbzalephResourceStatistics.write("hbzaleph", "json");
            hbzfixResourceStatistics.write("hbzfix", "json");
            lviResourceStatistics.write("lvi", "json");
        }

        private void writeDefaultConfiguration() {
            ConfigurationTable configurationTable = new ConfigurationTable(settings,
                    lviResourceStatistics, lviProductStatistics, hbzalmaStatusCodeMapStatistics);
            configurationTable.write("config");
            configurationTable.index(settings, elasticsearchOutput, targetConfigurationsIndexDefinition);
        }

        private void logStatistics() throws IOException {
            if (!settings.getAsBoolean("statistics.enabled", false)) {
                return;
            }

            logger.log(Level.INFO, "record counts = " + recordStatistics.getCountMap().toString());
            logger.log(Level.INFO, "record summary = " + recordStatistics.getCountSummary().toString());
            logger.log(Level.INFO, "record deleted counts = " + recordStatistics.getDeletedCountMap().toString());
            logger.log(Level.INFO, "record deleted summary = " + recordStatistics.getDeletedCountSummary().toString());
            logger.log(Level.INFO, "record material counts = " + recordStatistics.getMaterialCountMap().toString());
            logger.log(Level.INFO, "record material summary = " + recordStatistics.getMaterialCountSummary().toString());
            logger.log(Level.INFO, "record green counts = " + recordStatistics.getGreenCountMap().toString());
            logger.log(Level.INFO, "record green summary = " + recordStatistics.getGreenCountSummary().toString());
            logger.log(Level.INFO, "record open access counts = " + recordStatistics.getOpenAccessCountMap().toString());
            logger.log(Level.INFO, "record open access summary = " + recordStatistics.getOpenAccessCountSummary().toString());

            logger.log(Level.INFO, "zdb resources counts = " + zdbResourceStatistics.getCountMap().toString());
            logger.log(Level.INFO, "zdb resources collections = " + zdbResourceStatistics.getCollectionMapAsJson());
            logger.log(Level.INFO, "zdb resources statuscodes = " + zdbResourceStatistics.getStatusCodeMapAsJson());
            logger.log(Level.INFO, "zdb deletions = " + zdbDeleted.get());
            logger.log(Level.INFO, "zdb -> hbzaleph match count = " + zdbAlephMatch.get());
            logger.log(Level.INFO, "zdb -> hbzaleph H count = " + zdbAlephHCount.get());
            logger.log(Level.INFO, "zdb -> hbzaleph U count = " + zdbAlephUCount.get());
            logger.log(Level.INFO, "zdb -> hbzalma match count = " + zdbAlmaMatch.get());

            logger.log(Level.INFO, "hbzalma resources counts = " + hbzalmaResourceStatistics.getCountMap().toString());
            logger.log(Level.INFO, "hbzalma resources collections = " + hbzalmaResourceStatistics.getCollectionMapAsJson());
            logger.log(Level.INFO, "hbzalma resources statuscodes = " + hbzalmaResourceStatistics.getStatusCodeMapAsJson());
            logger.log(Level.INFO, "hbzalma resources processcodes = " + hbzalmaResourceStatistics.getItemProcessCodeMapAsJson());
            logger.log(Level.INFO, "hbzalma/hbzaleph statuscode mapping = " + hbzalmaStatusCodeMapStatistics.getCodeMaps().toString());

            logger.log(Level.INFO, "hbzfix deletions = " + hbzfixDeleted.get());

            logger.log(Level.INFO, "lvi resources counts = " + lviResourceStatistics.getCountMap().toString());
            logger.log(Level.INFO, "lvi resources count summary = " + lviResourceStatistics.getCountSummary().toString());
            logger.log(Level.INFO, "lvi deleted resources counts = " + lviResourceStatistics.getDeletedCountMap().toString());
            logger.log(Level.INFO, "lvi deleted resources count summary = " + lviResourceStatistics.getDeletedCountSummary().toString());
            logger.log(Level.INFO, "lvi material resources counts = " + lviResourceStatistics.getMaterialCountMap().toString());
            logger.log(Level.INFO, "lvi material resources count summary = " + lviResourceStatistics.getMaterialCountSummary().toString());
            logger.log(Level.INFO, "lvi product counts = " + lviProductStatistics.getCountMap().toString());
            logger.log(Level.INFO, "lvi product counts summary = " + lviProductStatistics.getCountMapSummary().toString());
            logger.log(Level.INFO, "lvi product code maps = " + lviProductStatistics.getCodeMap().toString());
            logger.log(Level.INFO, "lvi product code maps summary = " + lviProductStatistics.getCodeMapSummary().toString());
        }

        @Override
        public Worker<SearchDocument> newWorker(String name, int count) {
            return new LviWorker(settings, this, workerCount.incrementAndGet());
        }

        @Override
        public Iterator<WorkerRequest<SearchDocument>> iterator() {
            return getQueue().iterator();
        }

        @Override
        public WorkerRequest<SearchDocument> newWorkerRequest(Object input, String origin) {
            return new WorkerRequest<>() {
                @Override
                public SearchDocument getRequest() {
                    return (SearchDocument) input;
                }

                @Override
                public boolean isPoison() {
                    return false;
                }

                @Override
                public String getRequestOrigin() {
                    return origin;
                }
            };
        }

        @Override
        public void afterExecute(Worker<SearchDocument> worker, WorkerRequest<SearchDocument> request) {
        }

        Configuration getLibraryConfiguration() {
            return configuration;
        }

        private void processZdb() {
            logger.log(Level.INFO, "before zdb processing");
            QueryBuilder queryBuilder = matchAllQuery();
            String identifier = settings.get("zdb.identifier");
            if (identifier != null) {
                queryBuilder = termQuery("016.7_.a", identifier);
            }
            int size = settings.getAsInt("main.scrollsize", 100);
            TimeValue timeValue = TimeValue.timeValueMinutes(settings.getAsTime("main.scrolltime",
                    org.xbib.datastructures.api.TimeValue.timeValueMinutes(10)).getMinutes());
            final QueryBuilder qb = queryBuilder;
            Stream<SearchDocument> stream = elasticsearchInput.getSearchClient().search(srb -> srb
                    .setIndices(sourceZdbIndexDefinition.getFullIndexName()).setQuery(qb), timeValue, size);
            getExecutionCounter().set(0L);
            stream.forEach(doc -> {
                submit(newWorkerRequest(doc, "zdb"));
                if (maxCount.get() > 0 && getExecutionCounter().get() > maxCount.get()) {
                    throw new IllegalArgumentException("max count reached: " + maxCount.get());
                }
                if (elasticsearchInput.getSearchClient().isSearchMetricEnabled()) {
                    if (elasticsearchInput.getSearchClient().getSearchMetric().getFailedQueries().getCount() > 0L) {
                        throw new IllegalStateException("detected failed query while search after " + doc.getIndex() + "/" + doc.getId());
                    }
                    if (elasticsearchInput.getSearchClient().getSearchMetric().getTimeoutQueries().getCount() > 0L) {
                        throw new IllegalStateException("detected timed out query while search after " + doc.getIndex() + "/" + doc.getId());
                    }
                }
            });
            logger.log(Level.INFO, MessageFormat.format("after zdb processing: index count = {0} executions = {1}",
                    this.sourceZdbIndexCount, getExecutionCounter().get()));
        }

        private void processHbzAlma() {
            logger.log(Level.INFO, "before hbzalma processing");
            QueryBuilder queryBuilder = matchAllQuery();
            String identifier = settings.get("hbzalma.identifier");
            if (identifier != null) {
                queryBuilder = boolQuery()
                        .should(termQuery("_id", identifier))
                        .should(termQuery("035.__.a", "(DE-605)" + identifier));
            }
            int size = settings.getAsInt("main.scrollsize", 100);
            TimeValue timeValue = TimeValue.timeValueMinutes(settings.getAsTime("main.scrolltime",
                    org.xbib.datastructures.api.TimeValue.timeValueMinutes(10)).getMinutes());
            final QueryBuilder qb = queryBuilder;
            Stream<SearchDocument> stream = elasticsearchInput.getSearchClient().search(srb -> srb
                    .setIndices(sourceHbzAlmaIndexDefinition.getFullIndexName()).setQuery(qb), timeValue, size);
            getExecutionCounter().set(0L);
            stream.forEach(doc -> {
                submit(newWorkerRequest(doc, "hbzalma"));
                if (maxCount.get() > 0 && getExecutionCounter().get() > maxCount.get()) {
                    throw new IllegalArgumentException("max count reached: " + maxCount.get());
                }
                if (elasticsearchInput.getSearchClient().isSearchMetricEnabled()) {
                    if (elasticsearchInput.getSearchClient().getSearchMetric().getFailedQueries().getCount() > 0L) {
                        throw new IllegalStateException("detected failed query while search after " + doc.getIndex() + "/" + doc.getId());
                    }
                    if (elasticsearchInput.getSearchClient().getSearchMetric().getTimeoutQueries().getCount() > 0L) {
                        throw new IllegalStateException("detected timed out query while search after " + doc.getIndex() + "/" + doc.getId());
                    }
                }
            });
            logger.log(Level.INFO, MessageFormat.format("after hbzalma processing: index count = {0} executions = {1}",
                    this.sourceHbzAlmaIndexCount, getExecutionCounter().get()));
        }

        private void processHbzfix() {
            logger.log(Level.INFO, "before hbzfix processing");
            QueryBuilder queryBuilder = matchAllQuery();
            String identifier = settings.get("hbzfix.identifier");
            if (identifier != null) {
                queryBuilder = termQuery("_id", identifier);
            }
            int size = settings.getAsInt("main.scrollsize", 100);
            TimeValue timeValue = TimeValue.timeValueMinutes(settings.getAsTime("main.scrolltime",
                    org.xbib.datastructures.api.TimeValue.timeValueMinutes(10)).getMinutes());
            final QueryBuilder qb = queryBuilder;
            Stream<SearchDocument> stream = elasticsearchInput.getSearchClient().search(srb -> srb
                    .setIndices(sourceHbzfixIndexDefinition.getFullIndexName()).setQuery(qb), timeValue, size);
            getExecutionCounter().set(0L);
            stream.forEach(doc -> {
                submit(newWorkerRequest(doc, "hbzfix"));
                if (maxCount.get() > 0 && getExecutionCounter().get() > maxCount.get()) {
                    throw new IllegalArgumentException("max count reached: " + maxCount.get());
                }
                if (elasticsearchInput.getSearchClient().isSearchMetricEnabled()) {
                    if (elasticsearchInput.getSearchClient().getSearchMetric().getFailedQueries().getCount() > 0L) {
                        throw new IllegalStateException("detected failed query while search after " + doc.getIndex() + "/" + doc.getId());
                    }
                    if (elasticsearchInput.getSearchClient().getSearchMetric().getTimeoutQueries().getCount() > 0L) {
                        throw new IllegalStateException("detected timed out query while search after " + doc.getIndex() + "/" + doc.getId());
                    }
                }
            });
            logger.log(Level.INFO, MessageFormat.format("after hbzfix processing: index count = {0} executions = {1}",
                    this.sourceHbzFixIndexCount, getExecutionCounter().get()));
        }

        private void validateAndIndex(IndexDefinition indexDefinition, String id, String content) {
            if (id == null) {
                return;
            }
            if (enabled) {
                // id may be null but that would be ridiculous for repeatable indexing
                // true = enforce creation, false = override old documents
                elasticsearchOutput.index(indexDefinition, id, true, content);
            } else {
                if (content == null || content.isEmpty()) {
                    logger.log(Level.WARNING, "content is empty?");
                }
                logger.log(Level.FINE, MessageFormat.format("not indexing {0}/{1}/{2} {3}",
                        indexDefinition.getFullIndexName(), indexDefinition.getType(), id, content));
            }
        }
    }

    protected class LviWorker implements Worker<SearchDocument> {

        private final LviWorkerPool pool;

        private final Configuration configuration;

        private final String name;

        private final int scrollSize;

        private final long scrollMillis;

        protected LviWorker(Settings settings, LviWorkerPool pool, int n) {
            this.pool = pool;
            this.configuration = pool.getLibraryConfiguration();
            this.name = getClass().getSimpleName() + "." + n;
            this.scrollSize = settings.getAsInt("worker.scrollsize", 100); // per shard!
            this.scrollMillis = settings.getAsTime("worker.scrolltimeout",
                    org.xbib.datastructures.api.TimeValue.timeValueMinutes(10)).millis();
            // bail out early if fundamental input definitions are missing
            Map<String, ElasticsearchIndexDefinition> indexDefinitionMap = pool.elasticsearchInput.getIndexDefinitions();
            if (indexDefinitionMap.isEmpty()) {
                throw new IllegalArgumentException("pool elasticsearch input definition is empty");
            }
            if (indexDefinitionMap.get("zdb") == null) {
                throw new IllegalArgumentException("zdb elasticsearch input definition not found");
            }
            if (indexDefinitionMap.get("zdbholdings") == null) {
                throw new IllegalArgumentException("zdbholdings elasticsearch input definition not found");
            }
            if (indexDefinitionMap.get("ezb") == null) {
                throw new IllegalArgumentException("ezb elasticsearch input definition not found");
            }
            if (indexDefinitionMap.get("hbzalma") == null) {
                throw new IllegalArgumentException("hbzalma elasticsearch input definition not found");
            }
            if (indexDefinitionMap.get("hbzfix") == null) {
                throw new IllegalArgumentException("hbzfix elasticsearch input definition not found");
            }
            if (indexDefinitionMap.get("doaj") == null) {
                throw new IllegalArgumentException("doaj elasticsearch input definition not found");
            }
        }

        @Override
        public void execute(WorkerRequest<SearchDocument> workerRequest) throws IOException {
            SearchDocument searchDocument = workerRequest.getRequest();
            switch (workerRequest.getRequestOrigin()) {
                case "zdb" -> processZDB(searchDocument);
                case "hbzalma" -> processHbzAlma(searchDocument);
                case "hbzfix" -> processHbzFix(searchDocument);
                default -> throw new IllegalStateException("request origin not accepted for processing: " +
                        workerRequest.getRequestOrigin());
            }
        }

        @Override
        public void close() throws IOException {
            // do nothing
        }

        @Override
        public String toString() {
            return name;
        }

        private void processZDB(SearchDocument searchDocument) throws IOException {
            MarcRecord marcRecord = MarcRecord.from(searchDocument.getFields(), privateTags);
            final ZdbRecord zdbRecord = ZdbRecord.builder(configuration, marcRecord)
                    .build();
            pool.recordStatistics.count(zdbRecord);
            if (zdbRecord.isDeleted()) {
                pool.zdbDeleted.incrementAndGet();
                logger.log(Level.FINEST, "ZDB record skipped: " + searchDocument.getId());
            } else {
                String externalIdentifier = zdbRecord.getExternalIdentifier();
                if (externalIdentifier == null) {
                    // no ZDB-ID in ZDB? should not happen
                    logger.log(Level.WARNING, "no ZDB identifier found in zdb record, doc id = " +
                            searchDocument.getId());
                    return;
                }
                attachZDBHoldings(zdbRecord);
                attachEZBHoldings(zdbRecord);
                attachHbzAlma(zdbRecord);
                attachHbzAleph(zdbRecord);
                attachDOAJ(zdbRecord);
                List<MarcField> list = new LinkedList<>();
                zdbRecord.getHoldings().forEach(holding -> {
                    list.addAll(holding.asMarc());
                    pool.lviResourceStatistics.count(holding);
                    holding.getProducts().forEach(product -> pool.lviProductStatistics.count(product, holding.getISIL()));
                });
                // add alias identifiers (HT numbers from Aleph, MMS IDs from Alma)
                zdbRecord.getAlias().forEach(alias ->
                        list.add(MarcField.builder().tag("035").indicator("  ").subfield("a", alias).build()));
                MarcRecord cleanRecord = zdbRecord.getFilteredMarcRecord(list, fieldComparator);
                cleanRecord.rebuildMap();
                StringWriter stringWriter = new StringWriter();
                try (MarcJsonWriter writer = new MarcJsonWriter(stringWriter)
                        .setStyle(EnumSet.of(MarcJsonWriter.Style.ARRAY,
                                MarcJsonWriter.Style.EMBEDDED_RECORD))) {
                    writer.startDocument();
                    writer.record(cleanRecord);
                    JsonBuilder jsonBuilder = writer.getJsonBuilder();
                    jsonBuilder.buildKey("rec")
                            .beginMap()
                            .buildKey("created")
                            .buildValue(cleanRecord.getCreationDate(BASE_DATE).format(dateFormat))
                            .buildKey("modified")
                            .buildValue(cleanRecord.getLastModificationDate(BASE_DATE).format(dateFormat))
                            .buildKey("indexed")
                            .buildValue(Instant.now().toString())
                            .endMap();
                    writer.endDocument();
                }
                pool.validateAndIndex(pool.targetIndexDefinition,
                        zdbRecord.getIdentifier(), stringWriter.toString());
                if (pool.marcWriter != null && !pool.marcWriter.isClosed()) {
                    pool.marcWriter.record(cleanRecord);
                    Exception exception = pool.marcWriter.getException();
                    if (exception != null) {
                        logger.log(Level.SEVERE, "marc write error: " + exception, exception);
                        throw new IOException(exception);
                    }
                }
                createServices(zdbRecord);
                // pseudo LVI record for counting
                Record lviRecord = LviRecord.builder(configuration)
                        .setDeleted(zdbRecord.isDeleted())
                        .setMaterials(zdbRecord.getMaterials())
                        .setGreen(zdbRecord.isGreen())
                        .setOpenAccess(zdbRecord.isOpenAccess())
                        .build();
                pool.recordStatistics.count(lviRecord);
            }
        }

        private void processHbzAlma(SearchDocument searchDocument) throws IOException {
            MarcRecord marcRecord = MarcRecord.from(searchDocument.getFields(), privateTags);
            HbzAlmaRecord hbzAlmaRecord = HbzAlmaRecord.builder(configuration, marcRecord)
                    .build();
            pool.recordStatistics.count(hbzAlmaRecord);
            if (hbzAlmaRecord.isDeleted()) {
                logger.log(Level.FINEST, "hbzalma record deleted, skipped: " + searchDocument.getId());
                pool.hbzalmaDeleted.incrementAndGet();
            } else {
                String identifier = hbzAlmaRecord.getIdentifier();
                if (identifier == null) {
                    logger.log(Level.WARNING, "no identifier found in hbzalma record, doc id = " +
                            searchDocument.getId());
                } else {
                    String externalIdentifier = hbzAlmaRecord.getExternalIdentifier();
                    // - externalIdentifier can be null, means no HBZ-ID and not from ZDB (E-Book-Collections etc.)
                    // skip records with:
                    // - Alma 003 = "DE-605" plus Alma 001 "ZDB...", these are "mirrored ZDB records"
                    if (externalIdentifier != null && !externalIdentifier.startsWith("(DE-599)ZDB")) {
                        hbzAlmaRecord.setCreationDate(marcRecord.getCreationDate(BASE_DATE));
                        hbzAlmaRecord.setModificationDate(marcRecord.getLastModificationDate(BASE_DATE));
                        hbzAlmaRecord.handleItmResources(f -> {
                            // "ITM"
                            HbzAlmaItems.Builder hbzAlmaItems = HbzAlmaItems.builder(configuration, hbzAlmaRecord, f);
                            pool.hbzalmaResourceStatistics.count(hbzAlmaRecord.addHoldings(hbzAlmaItems));
                        });
                        hbzAlmaRecord.handleHolResources(f -> {
                            // "HOL"
                            HbzAlmaHoldings.Builder hbzAlmaHoldings = HbzAlmaHoldings.builder(configuration, hbzAlmaRecord, f);
                            pool.hbzalmaResourceStatistics.count(hbzAlmaRecord.addHoldings(hbzAlmaHoldings));
                        });
                        hbzAlmaRecord.handlePortfolioResources(f -> {
                            // "POR"
                            HbzAlmaOnlineResource.Builder almaResource = HbzAlmaOnlineResource.builder(configuration, hbzAlmaRecord, f);
                            pool.hbzalmaResourceStatistics.count(hbzAlmaRecord.addHoldings(almaResource));
                        });
                        hbzAlmaRecord.handlePortfolioCoverage(f -> {
                            // "POC"
                            String almaId = f.getFirstSubfieldValue("a");
                            hbzAlmaRecord.getHoldings().stream()
                                    .filter(h -> h.getIdentifier().contains(almaId))
                                    .findFirst()
                                    .ifPresent(h -> {
                                        String firstyear = f.getFirstSubfieldValue("b");
                                        String lastyear = f.getFirstSubfieldValue("c");
                                        h.addYears(firstyear, lastyear);
                                    });
                        });
                        List<MarcField> list = new LinkedList<>();
                        hbzAlmaRecord.getHoldings().forEach(holdings -> {
                            list.addAll(holdings.asMarc());
                            pool.lviResourceStatistics.count(holdings);
                            hbzAlmaRecord.getProducts().forEach(product -> pool.lviProductStatistics.count(product, holdings.getISIL()));
                        });
                        // use cleaned Alma record to conform with DNB specification plus our generated holdings
                        MarcRecord dnbRecord = hbzAlmaRecord.getFilteredMarcRecord(list, fieldComparator);
                        dnbRecord.rebuildMap();
                        StringWriter stringWriter = new StringWriter();
                        try (MarcJsonWriter writer = new MarcJsonWriter(stringWriter)
                                .setStyle(EnumSet.of(MarcJsonWriter.Style.ARRAY,
                                        MarcJsonWriter.Style.EMBEDDED_RECORD))) {
                            writer.startDocument();
                            writer.record(dnbRecord);
                            JsonBuilder jsonBuilder = writer.getJsonBuilder();
                            jsonBuilder.buildKey("rec")
                                    .beginMap()
                                    .buildKey("id")
                                    .buildValue(hbzAlmaRecord.getIdentifier())
                                    .buildKey("created")
                                    .buildValue(hbzAlmaRecord.getCreationDate().format(dateFormat))
                                    .buildKey("modified")
                                    .buildValue(hbzAlmaRecord.getModificationDate().format(dateFormat))
                                    .buildKey("indexed")
                                    .buildValue(Instant.now().toString())
                                    .endMap();
                            writer.endDocument();
                        }
                        pool.validateAndIndex(pool.targetIndexDefinition, identifier, stringWriter.toString());
                        if (pool.marcWriter != null && !pool.marcWriter.isClosed()) {
                            pool.marcWriter.record(dnbRecord);
                            Exception exception = pool.marcWriter.getException();
                            if (exception != null) {
                                logger.log(Level.SEVERE, "marc write error: " + exception, exception);
                                throw new IOException(exception);
                            }
                        }
                        // create LVI services and we are done
                        createServices(hbzAlmaRecord);
                        // build transient LVI record, just for counting the correct statistics
                        Record lviRecord = LviRecord.builder(configuration)
                                .setDeleted(hbzAlmaRecord.isDeleted())
                                .setGreen(hbzAlmaRecord.isGreen())
                                .setOpenAccess(hbzAlmaRecord.isOpenAccess())
                                .setMaterials(hbzAlmaRecord.getMaterials())
                                .build();
                        pool.recordStatistics.count(lviRecord);
                    }
                }
            }
        }

        private void processHbzFix(SearchDocument searchDocument) throws IOException {
            MarcRecord marcRecord = MarcRecord.from(searchDocument.getFields(), privateTags);
            HbzfixRecord hbzfixRecord = HbzfixRecord.builder(configuration, marcRecord)
                    .build();
            pool.recordStatistics.count(hbzfixRecord);
            if (hbzfixRecord.isDeleted()) {
                logger.log(Level.FINEST, "hbzfix record skipped: " + searchDocument.getId());
                pool.hbzfixDeleted.incrementAndGet();
            } else {
                String identifier = hbzfixRecord.getIdentifier();
                if (identifier == null) {
                    logger.log(Level.WARNING, "no identifier found in hbzfix record, id = " +
                            searchDocument.getId());
                } else {
                    hbzfixRecord.handleResources(f -> {
                        HbzfixHoldings.Builder hbzfixHoldings = HbzfixHoldings.builder(configuration, hbzfixRecord, f);
                        Holdings holdings = hbzfixRecord.addHoldings(hbzfixHoldings);
                        pool.hbzfixResourceStatistics.count(holdings);
                    });
                    List<MarcField> list = new LinkedList<>();
                    hbzfixRecord.getHoldings().forEach(holding -> {
                        list.addAll(holding.asMarc());
                        pool.lviResourceStatistics.count(holding);
                        hbzfixRecord.getProducts().forEach(product -> pool.lviProductStatistics.count(product, holding.getISIL()));
                    });
                    // delete existing MARC 924, we can not use it, and add our list of holdings
                    hbzfixRecord.getMarcRecord().filterFields(marcField -> !marcField.getTag().equals("924"),
                            list.stream(),
                            fieldComparator);
                    hbzfixRecord.getMarcRecord().rebuildMap();
                    StringWriter stringWriter = new StringWriter();
                    try (MarcJsonWriter writer = new MarcJsonWriter(stringWriter)
                            .setStyle(EnumSet.of(MarcJsonWriter.Style.ARRAY,
                                    MarcJsonWriter.Style.EMBEDDED_RECORD))) {
                        writer.startDocument();
                        writer.record(hbzfixRecord.getMarcRecord());
                        JsonBuilder jsonBuilder = writer.getJsonBuilder();
                        jsonBuilder.buildKey("rec")
                                .beginMap()
                                .buildKey("created")
                                .buildValue(hbzfixRecord.getMarcRecord().getCreationDate(BASE_DATE).format(dateFormat))
                                .buildKey("modified")
                                .buildValue(hbzfixRecord.getMarcRecord().getLastModificationDate(BASE_DATE).format(dateFormat))
                                .buildKey("indexed")
                                .buildValue(Instant.now().toString())
                                .endMap();
                        writer.endDocument();
                    }
                    pool.validateAndIndex(pool.targetIndexDefinition, identifier, stringWriter.toString());
                    if (pool.marcWriter != null && !pool.marcWriter.isClosed()) {
                        pool.marcWriter.record(hbzfixRecord.getMarcRecord());
                        Exception exception = pool.marcWriter.getException();
                        if (exception != null) {
                            logger.log(Level.SEVERE, "marc write error: " + exception, exception);
                            throw new IOException(exception);
                        }
                    }
                    createServices(hbzfixRecord);
                    // pseudo LVI record for counting
                    Record lviRecord = LviRecord.builder(configuration)
                            .setDeleted(hbzfixRecord.isDeleted())
                            .setMaterials(hbzfixRecord.getMaterials())
                            .setGreen(hbzfixRecord.isGreen())
                            .setOpenAccess(hbzfixRecord.isOpenAccess())
                            .build();
                    pool.recordStatistics.count(lviRecord);
                }
            }
        }

        private void attachZDBHoldings(ZdbRecord zdbRecord) {
            QueryBuilder qb = termQuery("016.7_.a", zdbRecord.getExternalIdentifier());
            search(pool.sourceZdbHoldingsIndexDefinition, qb, doc -> {
                ZdbHoldings.Builder zdbHoldings = ZdbHoldings.builder(configuration, zdbRecord, MarcRecord.from(doc.getFields()));
                pool.zdbResourceStatistics.count(zdbRecord.addHoldings(zdbHoldings));
            });
        }

        private void attachEZBHoldings(ZdbRecord zdbRecord) {
            QueryBuilder qb = termQuery("016.7_.a", zdbRecord.getExternalIdentifier());
            search(pool.sourceEzbHoldingsIndexDefinition, qb, doc -> {
                EzbHoldings.Builder ezbHoldings = EzbHoldings.builder(configuration, zdbRecord, MarcRecord.from(doc.getFields()));
                pool.zdbResourceStatistics.count(zdbRecord.addHoldings(ezbHoldings));
            });
        }

        private void attachHbzAleph(ZdbRecord zdbRecord) {
            QueryBuilder queryBuilder = termQuery("025.z1.a", zdbRecord.getExternalIdentifier()); // with hyphen
            search(pool.sourceHbzAlephIndexDefinition, queryBuilder, doc -> {
                MarcRecord alephMarcRecord = MarcRecord.from(doc.getFields());
                HbzAlephRecord hbzAlephRecord = HbzAlephRecord.builder(configuration, alephMarcRecord)
                        .build();
                // move HT number into ZDB record under 035
                zdbRecord.addAlias(hbzAlephRecord.getExternalIdentifier());
                if (hbzAlephRecord.isDeleted()) {
                    logger.log(Level.FINEST, "zdb -> aleph record skipped: " + doc.getId());
                } else {
                    pool.zdbAlephMatch.incrementAndGet();
                    if (hbzAlephRecord.getMarcRecord().getRecordLabel().toString().endsWith("h")) {
                        pool.zdbAlephHCount.incrementAndGet();
                    } else if (hbzAlephRecord.getMarcRecord().getRecordLabel().toString().endsWith("u")) {
                        pool.zdbAlephUCount.incrementAndGet();
                    }
                    hbzAlephRecord.handleResources(f -> {
                        HbzAlephHoldings.Builder alephHoldings = HbzAlephHoldings.builder(configuration, hbzAlephRecord, f);
                        zdbRecord.addHoldings(alephHoldings);
                    });
                }
            });
        }

        private void attachHbzAlma(ZdbRecord zdbRecord) {
            // Alma 035 $a format = "(DE-600)<ZDB-ID>"  "DE-600" = ZDB
            QueryBuilder queryBuilder = termQuery("035.__.a", "(DE-600)" + zdbRecord.getExternalIdentifier());
            search(pool.sourceHbzAlmaIndexDefinition, queryBuilder, doc -> {
                MarcRecord marcRecord = MarcRecord.from(doc.getFields());
                HbzAlmaRecord hbzAlmaRecord = HbzAlmaRecord.builder(configuration, marcRecord)
                        .build();
                if (hbzAlmaRecord.isDeleted()) {
                    logger.log(Level.FINEST, "zdb -> hbzalma record skipped becasue it is deleted: " + doc.getId());
                } else {
                    pool.zdbAlmaMatch.incrementAndGet();
                    zdbRecord.addAlias("(DE-655)" + stripPrefixIfPresent(hbzAlmaRecord.getIdentifier()));
                    // "ITM"
                    hbzAlmaRecord.handleItmResources(res -> {
                        HbzAlmaItems.Builder hbzAlmaItems = HbzAlmaItems.builder(configuration, hbzAlmaRecord, res);
                        // swap parent record identifier
                        hbzAlmaItems.setIdentifier(hbzAlmaItems.getIdentifier() + zdbRecord.getIdentifier());
                        hbzAlmaItems.setMerged(true);
                        zdbRecord.addHoldings(hbzAlmaItems);
                    });
                    // "HOL"
                    hbzAlmaRecord.handleHolResources(f -> {
                        HbzAlmaHoldings.Builder hbzAlmaHoldings = HbzAlmaHoldings.builder(configuration, hbzAlmaRecord, f);
                        // swap parent record identifier
                        hbzAlmaHoldings.setIdentifier(hbzAlmaHoldings.getIdentifier() + zdbRecord.getIdentifier());
                        hbzAlmaHoldings.setMerged(true);
                        zdbRecord.addHoldings(hbzAlmaHoldings);
                    });
                    // "POR"
                    hbzAlmaRecord.handlePortfolioResources(res -> {
                        HbzAlmaOnlineResource.Builder almaResource = HbzAlmaOnlineResource.builder(configuration, hbzAlmaRecord, res);
                        // swap parent record identifier
                        almaResource.setIdentifier(almaResource.getIdentifier() + zdbRecord.getIdentifier());
                        almaResource.setMerged(true);
                        zdbRecord.addHoldings(almaResource);
                    });
                    // "POC"
                    hbzAlmaRecord.handlePortfolioCoverage(res-> {
                        String almaId = res.getFirstSubfieldValue("a");
                        zdbRecord.getHoldings().stream()
                                .filter(h -> h.getIdentifier().contains(almaId))
                                .findFirst()
                                .ifPresent(h -> {
                                    String firstyear = res.getFirstSubfieldValue("b");
                                    String lastyear = res.getFirstSubfieldValue("c");
                                    h.addYears(firstyear, lastyear);
                                });
                    });
                }
            });
        }

        private void attachDOAJ(ZdbRecord zdbRecord) {
            for (String issn: zdbRecord.getISSNs()) {
                QueryBuilder qb = termQuery("dc:identifier", issn);
                search(pool.sourceDoajIndexDefinition, qb, doc -> {
                    zdbRecord.setOpenAccess(true);
                });
            }
        }

        private void search(IndexDefinition indexDefinition, QueryBuilder queryBuilder, Consumer<SearchDocument> consumer) {
            Stream<SearchDocument> stream = pool.elasticsearchInput.getSearchClient().search(srb -> srb
                    .setIndices(indexDefinition.getFullIndexName())
                    .setQuery(queryBuilder), TimeValue.timeValueMillis(scrollMillis), scrollSize);
            stream.forEach(consumer);
        }

        private void createServices(Record record) {
            ServiceManager serviceManager = new ServiceManager(pool.settings, configuration, record);
            serviceManager.index(pool.elasticsearchOutput, pool.targetServicesIndexDefinition);
        }

        private static final Map<String, Integer> domainOrder = new HashMap<>();

        static {
            domainOrder.put("NRW", 0);
            domainOrder.put("BAY", 1);
            domainOrder.put("GBV", 2);
            domainOrder.put("NIE", 2);
            domainOrder.put("HAM", 2);
            domainOrder.put("THU", 2);
            domainOrder.put("SAA", 2);
            domainOrder.put("BAW", 3);
            domainOrder.put("SAX", 3);
            domainOrder.put("HES", 4);
            domainOrder.put("BER", 5);
        }

        private static final Comparator<MarcField> fieldComparator = Comparator.comparing(LviWorker::getKey);

        private static String getKey(MarcField field) {
            return switch (field.getTag()) {
                // keep duplicate control tags and indicators ("repeatable")
                case "006", "007" -> field.getTag() + field.getFirstSubfield().getValue();
                // keep duplicate tags and indicators ("repeatable")
                case "016", "020", "022", "035", "264", "653" -> field.getTag() + field.getFirstSubfieldValue("a");
                // keep duplicate tags and indicators ("repeatable") with "9" subfield
                case "773", "776", "780" -> field.getTag() + field.getFirstSubfieldValue("9");
                // 856 = order by "u"
                case "856" -> field.getTag() + field.getFirstSubfieldValue("u");
                // 924 - order by domain + ISIL
                case "924" -> field.getTag() + domainOrder.get(field.getFirstSubfieldValue("c")) +
                        field.getFirstSubfieldValue("b") + // ISIL
                        field.getFirstSubfieldValue("g") + // Standort
                        field.getFirstSubfieldValue("h"); // Signatur
                // order fields by tag and indicator
                default -> field.toTagIndicatorKey();
            };
        }

        private static String stripPrefixIfPresent(String identifier) {
            if (identifier == null) {
                return null;
            }
            if (identifier.indexOf('(') == 0) {
                int pos = identifier.lastIndexOf(')');
                return identifier.substring(pos + 1);
            }
            return identifier;
        }
    }
}
