package org.xbib.interlibrary.catalog.elasticsearch.lvi.config;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.api.library.Domain;
import org.xbib.interlibrary.api.library.IndexLibrary;
import org.xbib.interlibrary.api.product.Product;
import org.xbib.interlibrary.api.library.DefaultIndexLibrary;
import org.xbib.settings.Settings;
import org.xbib.settings.SettingsLoader;
import org.xbib.settings.datastructures.json.JsonSettingsLoader;

import static org.xbib.interlibrary.api.library.LibraryHelper.mainISIL;

public class DefaultConfiguration implements Configuration {

    private static final Logger logger = Logger.getLogger(DefaultConfiguration.class.getName());

    protected final ConfigurationArguments arguments;

    protected final Settings settings;

    private final Set<String> domains;

    private final Map<String, IndexLibrary> libraries;

    private final Collection<String> blacklisted;

    private final Collection<String> unconfiguredDomain;

    private final Map<String, Object> mappedLibraries;

    private final Map<String, String> aliases;

    private final TabSigel tabSigel;

    private final ResourceLoader resourceLoader;

    private final Map<String, Product> products;

    private final Map<String, Object> itemProcesses;

    public DefaultConfiguration(ConfigurationArguments arguments) throws IOException {
        this.unconfiguredDomain = new ArrayList<>();
        this.blacklisted = new ArrayList<>();
        this.arguments = arguments;
        SettingsLoader settingsLoader = new JsonSettingsLoader();
        this.settings = Settings.settingsBuilder()
                .put(settingsLoader.load(arguments.getSettings()))
                .build();
        this.domains = new HashSet<>(Arrays.asList(settings.getAsArray("domains")));
        this.tabSigel = new TabSigel();
        this.libraries = new HashMap<>();
        this.aliases = new HashMap<>();
        this.resourceLoader = new ResourceLoader();
        this.mappedLibraries = resourceLoader.loadResourceMap("org/xbib/interlibrary/catalog/elasticsearch/lvi/config/isil-map.json");
        this.products = new HashMap<>();
        this.itemProcesses = resourceLoader.loadResourceMap("org/xbib/interlibrary/catalog/elasticsearch/lvi/config/item-process.json");
        logger.log(Level.INFO, "configured domains: " + domains);
        logger.log(Level.INFO, "tab sigel: " + tabSigel.getMap().size());
        logger.log(Level.INFO, "mapped libraries: " + mappedLibraries.size());
        logger.log(Level.INFO, "item processes: " + itemProcesses.size());
    }

    @Override
    public Map<String, IndexLibrary> getLibraries() {
        return libraries;
    }


    @Override
    public String normalizeLibraryIdentifier(String identifier) {
        if (identifier == null) {
            return null;
        }
        if (aliases.containsKey(identifier)) {
            Object object = aliases.get(identifier);
            return object == null ? null : object.toString();
        }
        // maybe illegal spacing?
        String spacedIdentifier = identifier.replaceAll("(\\p{L}+)(\\p{N}+)", "$1 $2")
                .replaceAll("(\\p{N}+)(\\p{L}+)", "$1 $2");
        if (aliases.containsKey(spacedIdentifier)) {
            Object object = aliases.get(spacedIdentifier);
            return object == null ? null : object.toString();
        }
        return identifier;
    }

    @SuppressWarnings("unchecked")
    @Override
    public List<String> mapLibraryIdentifier(String input) {
        String result = input;
        if (result == null) {
            return List.of();
        }
        if (tabSigel.containsKey(result)) {
            result = tabSigel.get(result);
        }
        result = normalizeLibraryIdentifier(result);
        if (result == null) {
            return List.of();
        }
        if (mappedLibraries.containsKey(result)) {
            Object mapped = mappedLibraries.get(result);
            if (mapped == null) {
                return List.of();
            }
            if (mapped instanceof List) {
                return (List<String>) mapped;
            }
            return List.of(mapped.toString());
        }
        return List.of(result);
    }

    @Override
    public Map<String, Product> getProducts() {
        return products;
    }

    public String mapItemProcess(String code) {
        return code != null ? itemProcesses.getOrDefault(code, code).toString() : null;
    }

    protected void addCustomLibraries() {
        for (Map.Entry<String, Settings> entry : settings.getGroups("libraries").entrySet()) {
            DefaultIndexLibrary.Builder builder = DefaultIndexLibrary.builder();
            builder.setISIL(entry.getKey());
            builder.addAlias(entry.getValue().get("sigel"));
            builder.setName(entry.getValue().get("name"));
            String domainString = entry.getValue().get("domain");
            if (domainString != null) {
                String organization = entry.getValue().get("organization");
                builder.setDomain(Domain.of(domainString, organization));
                if (domains.contains(domainString)) {
                    addLibrary(entry.getKey(), builder);
                    logger.log(Level.FINE, "added custom library: " + entry.getKey());
                } else {
                    unconfiguredDomain.add(entry.getKey());
                }
            }
        }
        logger.log(Level.INFO, "libraries configured = " + getLibraries().size());
        logger.log(Level.INFO, "libraries ignored because of unconfigured domain = " + unconfiguredDomain.size());
        logger.log(Level.INFO, "libraries blacklisted = " + blacklisted);
        logger.log(Level.INFO, "products configured = " + getProducts().size());
    }

    @SuppressWarnings("unchecked")
    protected void addLibrary(String isil, DefaultIndexLibrary.Builder builder) {
        if (isil == null) {
            return;
        }
        if (libraries.containsKey(isil)) {
            logger.log(Level.WARNING, "overriding " + isil);
        }
        if (isBlackListed(isil)) {
            blacklisted.add(isil);
            return;
        }
        // add local setup if present. If not, try the "main ISIL"
        Map<String, Object> map = resourceLoader.loadResourceMap("org/xbib/interlibrary/catalog/elasticsearch/lvi/config/" + isil + ".json");
        if (map == null || map.isEmpty()) {
            map = resourceLoader.loadResourceMap("org/xbib/interlibrary/catalog/elasticsearch/lvi/config/" + mainISIL(isil) + ".json");
        }
        if (map != null && !map.isEmpty()) {
            List<String> locations = (List<String>) map.get("locations");
            if (locations != null) {
                builder.setCollections(locations);
            }
            Map<String, Object> filter = (Map<String, Object>) map.get("filter");
            if (filter != null) {
                builder.setFilter(filter.entrySet().stream()
                    .map(e -> Map.entry(Pattern.compile(e.getKey()), e.getValue()))
                    .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue)));
            }
            TinyMap.Builder<String, Object> statusCodes = TinyMap.builder();
            statusCodes.put("a", "a");
            statusCodes.put("b", "b");
            statusCodes.put("c", "c");
            statusCodes.put("d", "d");
            statusCodes.put("e", "e");
            Map<String, Object> m = (Map<String, Object>) map.get("statuscodes");
            if (m != null) {
                statusCodes.putAll(m);
                builder.setStatusCodes(statusCodes.build());
            }
            m = (Map<String, Object>) map.get("sublibraries");
            if (m != null) {
                builder.setSubLibraries(m);
            }
        }
        DefaultIndexLibrary library = builder.build();
        library.getAliases().forEach(alias -> addAlias(alias, isil));
        libraries.put(isil, library);
    }

    protected void addAlias(String identifier, String normalizedIdentifier) {
        aliases.put(identifier, normalizedIdentifier);
    }

    protected void addProduct(String isil, Product product) {
        if (isil == null || product == null) {
            return;
        }
        products.put(isil, product);
    }

    @Override
    public void close() throws IOException {
        // nothing to close
    }

    private boolean isBlackListed(String isil) {
        if (mappedLibraries.containsKey(isil)) {
            String mappedIsil = (String) mappedLibraries.get(isil);
            // blacklisted = entry is set to null
            return mappedIsil == null;
        }
        return false;
    }
}
