package org.xbib.interlibrary.catalog.elasticsearch.lvi.record;

import java.time.Year;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xbib.interlibrary.api.library.Mode;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Holdings;
import org.xbib.interlibrary.api.library.IndexLibrary;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.api.product.Product;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Record;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Service;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;

import static org.xbib.interlibrary.api.library.Material.ELECTRONIC_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.GLOBE;
import static org.xbib.interlibrary.api.library.Material.KIT;
import static org.xbib.interlibrary.api.library.Material.MAP;
import static org.xbib.interlibrary.api.library.Material.MICROFORM;
import static org.xbib.interlibrary.api.library.Material.MOTION_PICTURE;
import static org.xbib.interlibrary.api.library.Material.NON_PROJECTED_GRAPHIC;
import static org.xbib.interlibrary.api.library.Material.NOTATED_MUSIC;
import static org.xbib.interlibrary.api.library.Material.ONLINE_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.PROJECTED_GRAPHIC;
import static org.xbib.interlibrary.api.library.Material.REMOTE_SENSING_IMAGE;
import static org.xbib.interlibrary.api.library.Material.SOUND_RECORDING;
import static org.xbib.interlibrary.api.library.Material.TACTILE;
import static org.xbib.interlibrary.api.library.Material.TEXT;
import static org.xbib.interlibrary.api.library.Material.UNSPECIFIED;
import static org.xbib.interlibrary.api.library.Material.VIDEO_RECORDING;
import static org.xbib.interlibrary.api.library.Mode.COPY;
import static org.xbib.interlibrary.api.library.Mode.LOAN;
import static org.xbib.interlibrary.api.library.Mode.LOAN_COPY;
import static org.xbib.interlibrary.api.library.Mode.NONE;
import static org.xbib.interlibrary.api.library.Mode.ELECTRONIC_COPY;

public abstract class AbstractHoldings implements Holdings {

    private static final Logger logger = Logger.getLogger(AbstractHoldings.class.getName());

    protected Builder builder;

    private final Collection<Integer> years;

    protected AbstractHoldings() {
        this.years = new LinkedHashSet<>();
    }

    /**
     * MARC 924
     * 9 = Sigel
     * b = ISIL
     * c = Region
     * d = Fernleihindikator (a,b,c,d,e)
     * e = Einschränkung zum Frnleihindikator ("p")
     * f = Kommentar zum Fernleihinidkator
     * g = Signatur
     * h = Standort (Kollektion)
     * i = Standortsignatur (?)
     * j = Kommentar zur Signatur, Lückenangaben usw.
     * k = URL
     * l = URL-Beschreibung
     * m,n,o,p,q,r,s,t = normierter Bestandsverlauf
     * y = Moving Wall
     * z = zusammenfassender Bestandsverlauf
     * @return the MARC field
     */

    @Override
    public Collection<MarcField> asMarc() {
        List<MarcField> list = new ArrayList<>();
        if (builder.isil == null) {
            return list;
        }
        for (Service service : builder.services) {
            MarcField.Builder marcFieldBuilder = MarcField.builder()
                    .tag("924")
                    .indicator(getIndicator())
                    .subfieldIfNotNull("9", builder.sigel)
                    .subfieldIfNotNull("b", builder.isil)
                    .subfieldIfNotNull("c", builder.domain);
            marcFieldBuilder.subfieldIfNotNull("d", service.getFirstStatusCode());
            marcFieldBuilder.subfieldIfNotNull("e", service.getSecondStatusCode());
            marcFieldBuilder.subfieldIfNotNull("f", service.getThirdStatusCode());
            marcFieldBuilder.subfieldIfNotNull("g", service.getCallnumber());
            if (!builder.isil.equals(service.getISIL())) {
                marcFieldBuilder.subfieldIfNotNull("h", service.getISIL());
            }
            marcFieldBuilder.subfieldIfNotNull("h", service.getCollection());
            marcFieldBuilder.subfieldIfNotNull("i", service.getCallnumber2());
            marcFieldBuilder.subfieldIfNotNull("j", service.getDescription());
            for (Service.LabeledUrl labeledUrl : service.getUrls()) {
                marcFieldBuilder.subfieldIfNotNull("k", labeledUrl.getUrl());
                marcFieldBuilder.subfieldIfNotNull("l", labeledUrl.getLabel());
            }
            customSubfields(marcFieldBuilder);
            marcFieldBuilder.subfieldIfNotNull("y", builder.movingWall);
            marcFieldBuilder.subfieldIfNotNull("z", service.getTextualHoldings());
            list.add(marcFieldBuilder.build());
        }
        return list;
    }

    protected abstract String getIndicator();

    protected abstract void customSubfields(MarcField.Builder marcFieldBuilder);

    @Override
    public boolean isDeleted() {
        return builder.isDeleted;
    }

    @Override
    public boolean isMerged() {
        return builder.isMerged;
    }

    @Override
    public boolean isGreen() {
        return builder.isGreen;
    }

    @Override
    public String getIdentifier() {
        return builder.identifier;
    }

    @Override
    public String getISIL() {
        return builder.isil;
    }

    @Override
    public String getCategory() {
        return builder.isil;
    }

    @Override
    public String getDomain() {
        return builder.domain;
    }

    /**
     * The holdings key.
     * It consists of three parts: the parent record ID, the ISIL, and the holdings record ID (which might be synthetic).
     * Attention: the first "(" ")" parenthesis pair must denote the ISIL.
     * Therefore, the parent record identifier is second.
     * The parent record identifier ensures that we have unqique keys and we can use "create = true" in ES indexing.
     * @return the holdings key
     */
    @Override
    public String getKey() {
        return builder.identifier + builder.record.getIdentifier();
    }

    @Override
    public Collection<Service> getServices() {
        return builder.services;
    }

    @Override
    public Collection<Product> getProducts() {
        return builder.products;
    }

    @Override
    public void merge(Holdings holdings) {
        holdings.getServices().forEach(this::merge);
    }

    @Override
    public void addYear(int year) {
        if (year != -1) {
            this.years.add(year);
        }
    }

    @Override
    public void addYears(String firstYear, String lastYear) {
        if (firstYear != null) {
            int year1 = Integer.parseInt(firstYear);
            int year2 = Year.now().getValue();
            if (lastYear != null) {
                year2 = Integer.parseInt(lastYear);
            }
            for (int i = year1; i <= year2; i++) {
                addYear(i);
            }
        }
    }

    @Override
    public Collection<Integer> getYears() {
        return years;
    }

    @Override
    public int hashCode() {
        return Objects.hash(builder.isil, builder.services);
    }

    @Override
    public boolean equals(Object obj) {
        return super.equals(obj);
    }

    @Override
    public int compareTo(Holdings o) {
        return toString().compareTo(o.toString());
    }

    @Override
    public String toString() {
        return "[" + builder.isil + "][" + builder.services + "]";
    }

    protected void merge(Service service) {
        Optional<Service> serviceOptional = builder.services.stream()
                .filter(s -> equivalent(service, s))
                .findAny();
        if (serviceOptional.isPresent()) {
            Service oldService = serviceOptional.get();
            collapse(service, oldService);
            builder.services.remove(oldService);
            builder.services.add(service);
        } else {
            builder.services.add(service);
        }
    }

    protected boolean equivalent(Service a, Service b) {
        // if service a is empty, use service b
        boolean b0 = a.getCollection() == null &&
                a.getCallnumber() == null &&
                a.getUrls().isEmpty() &&
                a.getTextualHoldings() == null;
        if (b0) {
            return true;
        }
        boolean b1 = compareCollection(a, b);
        boolean b2 = compareCallnumber(a, b);
        boolean b3 = compareLinks(a, b);
        boolean b4 = compareTextualHoldings(a,b);
        return b1 && b2 && b3 && b4;
    }

    private static boolean compareCollection(Service a, Service b) {
        return (a.getCollection() != null && a.getCollection().equals(b.getCollection())) ||
                a.getCollection() == null || b.getCollection() == null;
    }

    private static boolean compareCallnumber(Service a, Service b) {
        return (a.getCallnumber() != null && a.getCallnumber().equals(b.getCallnumber())) ||
                a.getCallnumber() == null || b.getCallnumber() == null;
    }

    private static boolean compareLinks(Service a, Service b) {
        return (a.getUrls().equals(b.getUrls()) || a.getUrls().isEmpty() || b.getUrls().isEmpty()) &&
                (a.getTextualHoldings() != null && a.getTextualHoldings().equals(b.getTextualHoldings()));
    }

    private static boolean compareTextualHoldings(Service a, Service b) {
        return (a.getTextualHoldings() != null && a.getTextualHoldings().equals(b.getTextualHoldings())) ||
                a.getTextualHoldings() == null || b.getTextualHoldings() == null;
    }

    private static void collapse(Service a, Service b) {
        // special case: if new ISIL is shorter, keep the old one (example: Alma-Aleph HT018367575 DE-6-182)
        String oldISIL = b.getISIL();
        String newISIL = a.getISIL();
        if ((oldISIL != null &&  newISIL!= null && oldISIL.length() > newISIL.length()) || newISIL == null) {
            a.setISIL(oldISIL);
        }
        if (a.getCollection() == null && b.getCollection() != null) {
            a.setCollection(b.getCollection());
        }
        if (a.getCallnumber() == null && b.getCallnumber() != null) {
            a.setCallnumber(b.getCallnumber());
        }
        if (a.getCallnumber2() == null && b.getCallnumber2() != null) {
            a.setCallnumber2(b.getCallnumber2());
        }
        if (a.getDescription() == null && b.getDescription() != null) {
            a.setDescription(b.getDescription());
        }
        if (a.getUrls().size() < b.getUrls().size()) {
            b.getUrls().forEach(u -> a.getUrls().add(u));
        }
        if (a.getTextualHoldings() == null && b.getTextualHoldings() != null) {
            a.setTextualHoldings(b.getTextualHoldings());
        }
        // cleanup description if info is duplicated (e.g. HT020038521, DE-386)
        if (a.getCallnumber() != null && a.getCallnumber().equals(a.getDescription())) {
            a.setDescription(null);
        }
    }

    public static abstract class Builder implements Holdings.Builder {

        protected Configuration configuration;

        protected Record record;

        protected MarcField marcField;

        protected String identifier;

        protected boolean isDeleted;

        protected boolean isMerged;

        protected boolean isGreen;

        protected Material material;

        /**
         * The main library identifier. Must not be a sublibrary.
         */
        protected String isil;

        /**
         * The specific library identifier for this holding. Can be a sublibrary.
         */
        protected String fullIsil;

        protected String sigel;
        /**
         * The domain of the library.
         */
        protected String domain;

        protected Collection<Service> services;

        protected String movingWall;

        protected final List<MarcField.Subfield> subfieldList;

        protected final List<Product> products;

        protected Builder(Configuration configuration,
                          Record record,
                          MarcField marcField) {
            this.configuration = configuration;
            this.record = record;
            this.marcField = marcField;
            this.services = new LinkedHashSet<>();
            this.subfieldList = new ArrayList<>();
            this.products = new ArrayList<>();
            if (marcField != null) {
                setISIL(configuration, marcField);
                setIdentifier("(" + isil + ")" + getIdentifier(marcField));
                if (identifier == null) {
                    setIdentifier("(" + isil + ")" + record.getIdentifier());
                }
            }
            detectMaterial(record.getMarcRecord());
        }

        public void setIdentifier(String identifier) {
            this.identifier = identifier;
        }

        public String getIdentifier() {
            return identifier;
        }

        public String getIdentifier(MarcField marcField) {
            return marcField.getFirstSubfieldValue("a");
        }

        @Override
        public Builder setDeleted(boolean isDeleted) {
            this.isDeleted = isDeleted;
            return this;
        }

        protected void detectMaterial(MarcRecord marcRecord) {
            MarcField mf = marcRecord.getFirst("007");
            if (mf != null) {
                String value = mf.recoverControlFieldValue();
                this.material = switch (value.substring(0, 1)) {
                    case "a" -> MAP;
                    case "c" -> value.startsWith("cr") ? ONLINE_RESOURCE :ELECTRONIC_RESOURCE;
                    case "d" -> GLOBE;
                    case "f" -> TACTILE;
                    case "g" -> PROJECTED_GRAPHIC;
                    case "h" -> MICROFORM;
                    case "k" -> NON_PROJECTED_GRAPHIC;
                    case "m" -> MOTION_PICTURE;
                    case "o" -> KIT;
                    case "q" -> NOTATED_MUSIC;
                    case "r" -> REMOTE_SENSING_IMAGE;
                    case "s" -> SOUND_RECORDING;
                    case "t" -> TEXT;
                    case "v" -> VIDEO_RECORDING;
                    default -> UNSPECIFIED;
                };
            } else {
                this.material = UNSPECIFIED;
            }
        }

        @Override
        public Builder setMerged(boolean merged) {
            this.isMerged = merged;
            return this;
        }

        protected void setISIL(Configuration configuration, MarcField marcfield) {
            String candidate = findISILCandidate(record.getMarcRecord(), marcfield);
            if (candidate == null || candidate.isEmpty()) {
                setDeleted(true);
                return;
            }
            String fullIsil = null;
            // map candidate to ISIL and check for black list
            List<String> mapped = configuration.mapLibraryIdentifier(candidate);
            if (mapped != null && !mapped.isEmpty()) {
                fullIsil = mapped.getFirst();
            }
            if (fullIsil == null) {
                setDeleted(true);
                return;
            }
            // create a main ISIL and a full ISIL
            String mainIsil = getMainIsil(fullIsil);
            if (!fullIsil.equals(mainIsil)) {
                this.fullIsil = fullIsil;
            }
            IndexLibrary library = configuration.getLibraries().get(mainIsil);
            if (library == null) {
                setDeleted(true);
                return;
            }
            // fix library if additional info in the marc field and in the library config is available
            library = fixLibrary(library, record.getMarcRecord(), marcfield);
            this.isil = library.getISIL();
            this.domain = library.getDomain().getName();
            // first alias is the "sigel" (original legacy identifier). Might not be present.
            // The "sigel" must be written in MARC 924 $9 (for whatever reason)
            this.sigel = library.getAliases() != null && !library.getAliases().isEmpty() ?
                    library.getAliases().iterator().next() : null;
        }

        protected void addAlias(String alias) {
            this.sigel = alias;
        }

        protected String findISILCandidate(MarcRecord marcRecord, MarcField marcfield) {
            return marcfield.getLastSubfieldValue("a");
        }

        protected IndexLibrary fixLibrary(IndexLibrary library,
                                          MarcRecord marcRecord,
                                          MarcField marcField) {
            return library;
        }

        protected String getMainIsil(String isil) {
            if (isil == null) {
                return null;
            }
            int pos1 = isil.indexOf('-');
            int pos2 = isil.lastIndexOf('-');
            if (pos1 > 0 && pos1 == pos2) {
                return isil;
            } else if (pos2 > 0) {
                return isil.substring(0, pos2);
            } else {
                return isil;
            }
        }

        /**
         * Reconstruct sublibrary information from a location information.
         */
        protected void setSubLibrary(Service service, String location) {
            if (location == null) {
                return;
            }
            IndexLibrary library = configuration.getLibraries().get(isil);
            if (library != null) {
                Map<String, Object> map = library.getSubLibraries();
                if (map != null) {
                    if (map.containsKey(location)) {
                        String subISIL = (String) map.get(location);
                        service.setISIL(subISIL);
                    }
                }
            }
        }

        /**
         * Map status text to our codes.
         * @param service the service
         * @param statusText the status text given in catalog
         * @param restrictionText the status restriction text given in the catalog
         * @param comment the status text comment as give in the catalog
         */
        protected void setStatusCode(Service service,
                                     String statusText,
                                     String restrictionText,
                                     String comment) {
            if (service == null) {
                return;
            }
            String isil = service.getISIL();
            if (isil == null) {
                return;
            }
            IndexLibrary library = configuration.getLibraries().get(isil);
            if (library == null) {
                return;
            }
            // "p" = "postal delivery", we ignore "n" = "domestic loan only"
            if (restrictionText.contains("p")) {
                service.setSecondStatusCode("p");
            }
            // comment
            if (comment != null && !comment.isEmpty()) {
                service.setThirdStatusCode(comment);
            }
            if (statusText == null || statusText.isEmpty()) {
                // allow nothing if nothing is given in the catalog
                service.setFirstStatusCode("d");
                updateMode(service);
                return;
            }
            // we maybe a code mapping
            Map<String, Object> statusCodeMap = library.getStatusCodes();
            if (statusCodeMap != null && statusCodeMap.containsKey(statusText)) {
                service.setFirstStatusCode(statusCodeMap.get(statusText).toString());
            } else {
                switch (statusText) {
                    case "a", "b", "c", "d", "e" ->
                            service.setFirstStatusCode(statusText); // direct code
                    // the following cases are combinations observed in EZB or ZDB, some may be legacy/obsolete
                    case "k" -> service.setFirstStatusCode("b");
                    case "l" -> service.setFirstStatusCode("c");
                    case "n" -> service.setFirstStatusCode("d");
                    default -> {
                        library.addUnmappedStatusText(statusText);
                        service.setFirstStatusCode(service.getMaterial() == ONLINE_RESOURCE ? "b" : "c");
                    }
                }
            }
            updateMode(service);
        }

        protected void updateMode(Service service) {
            if (service.getFirstStatusCode() == null) {
                logger.log(Level.WARNING, "logic is broken, no first status code found for service " + service);
                return;
            }
            // set mode according to first status code
            service.setMode(switch (service.getFirstStatusCode()) {
                case "a" -> LOAN;
                case "b" -> COPY;
                case "c" -> service.getMaterial() == ONLINE_RESOURCE ? COPY : LOAN_COPY;
                case "d" -> NONE;
                case "e" -> ELECTRONIC_COPY;
                default -> Mode.UNSPECIFIED;
            });
        }

        protected void updateModeByInterlibraryRelevance(Service service, MarcRecord marcRecord) {
            if (service.getMode() != null) {
                return;
            }
            // In rare cases, we were unable to compute interlibrary mode (from MARC 092 $op).
            // Possible reasons: libraries abroad, outside the domains of national interlibrary services.
            // Then, we look at MARC 092 $l and try to assume interlibrary mode from there.
            // This mode is expected to be rarely processed, except for international interlibrary loan.
            // 092 $l
            // a = leihverkehrsrelevant
            // b = nicht leihverkehrsrelevant
            // c = keine Angabe
            marcRecord.getAll("092").forEach(marcField -> {
                String relevance = marcField.getFirstSubfieldValue("l");
                if (relevance != null) {
                    switch (relevance) {
                        case "a" -> service.setMode(LOAN_COPY);
                        case "b", "c" -> service.setMode(NONE);
                    }
                }
            });
        }
    }
}
