package org.xbib.interlibrary.catalog.elasticsearch.lvi.record;

import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

import org.xbib.interlibrary.catalog.elasticsearch.lvi.Holdings;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.api.product.Product;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Record;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Service;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;
import org.xbib.marc.label.RecordLabel;
import org.xbib.marc.label.RecordStatus;

import static org.xbib.interlibrary.api.library.Material.ELECTRONIC_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.GLOBE;
import static org.xbib.interlibrary.api.library.Material.KIT;
import static org.xbib.interlibrary.api.library.Material.MAP;
import static org.xbib.interlibrary.api.library.Material.MICROFORM;
import static org.xbib.interlibrary.api.library.Material.MOTION_PICTURE;
import static org.xbib.interlibrary.api.library.Material.NON_PROJECTED_GRAPHIC;
import static org.xbib.interlibrary.api.library.Material.NOTATED_MUSIC;
import static org.xbib.interlibrary.api.library.Material.ONLINE_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.PROJECTED_GRAPHIC;
import static org.xbib.interlibrary.api.library.Material.REMOTE_SENSING_IMAGE;
import static org.xbib.interlibrary.api.library.Material.SOUND_RECORDING;
import static org.xbib.interlibrary.api.library.Material.TACTILE;
import static org.xbib.interlibrary.api.library.Material.TEXT;
import static org.xbib.interlibrary.api.library.Material.UNSPECIFIED;
import static org.xbib.interlibrary.api.library.Material.VIDEO_RECORDING;

public abstract class AbstractRecord implements Record {

    protected Builder builder;

    protected final Collection<Holdings> holdings;

    /**
     * Foreign system identifiers for MARC 035.
     */
    protected final Collection<String> alias;

    protected LocalDate creationDate;

    protected LocalDate modificationDate;

    protected AbstractRecord(Builder builder) {
        this.builder = builder;
        this.holdings = new ArrayList<>();
        this.alias = new ArrayList<>();
    }

    @Override
    public MarcRecord getMarcRecord() {
        return builder.marcRecord;
    }

    @Override
    public String getCategory() {
        return builder.database;
    }

    public void setCreationDate(LocalDate creationDate) {
        this.creationDate = creationDate;
    }

    @Override
    public LocalDate getCreationDate() {
        return creationDate;
    }

    public void setModificationDate(LocalDate modificationDate) {
        this.modificationDate = modificationDate;
    }

    public LocalDate getModificationDate() {
        return modificationDate;
    }

    @Override
    public Holdings addHoldings(Holdings.Builder holdingsBuilder) {
        Holdings holdings = build(holdingsBuilder);
        if (!holdings.isDeleted()) {
            if (!tryMerge(holdings)) {
                this.holdings.add(holdings);
            }
        }
        return holdings;
    }

    @Override
    public void addAlias(String alias) {
        this.alias.add(alias);
    }

    @Override
    public Collection<String> getAlias() {
        return alias;
    }

    @Override
    public Collection<Holdings> getHoldings() {
        return holdings;
    }

    @Override
    public boolean isDeleted() {
        return builder.isDeleted;
    }

    @Override
    public List<Material> getMaterials() {
        return builder.materials;
    }

    @Override
    public boolean isSerialMirror() {
        return builder.isSerialMirror;
    }

    @Override
    public boolean isGreen() {
        return builder.isGreen;
    }

    @Override
    public boolean isOpenAccess() {
        return builder.isOpenAccess;
    }

    @Override
    public String getIdentifier() {
        return builder.identifier;
    }

    @Override
    public String getExternalIdentifier() {
        return builder.externalIdentifier;
    }

    @Override
    public String getKey() {
        return builder.identifier != null ? builder.identifier : "";
    }

    @Override
    public String getBaseUrl() {
        return builder.baseUrl;
    }

    @Override
    public Integer getPublicationDate() {
        return builder.publicationDate;
    }

    @Override
    public String getChronology() {
        return builder.chronology;
    }

    @Override
    public Collection<Product> getProducts() {
        return builder.products;
    }

    @Override
    public Collection<Service.LabeledUrl> getGlobalServices() {
        return builder.globalServices;
    }

    @Override
    public String toString() {
        return getKey();
    }

    @Override
    public boolean equals(Object other) {
        return other instanceof Record && getKey() != null && getKey().equals(((Record) other).getKey());
    }

    @Override
    public int hashCode() {
        return getKey().hashCode();
    }

    @Override
    public int compareTo(Record record) {
        return getKey().compareTo(record.getKey());
    }

    /**
     * Somewhat hacky. We create Holdings and check for what holdings class we have.
     * The holdings can be marked for deletion here.
     * @param holdingsBuilder the holdings builder
     * @return the holdings
     */
    protected Holdings build(Holdings.Builder holdingsBuilder) {
        if (holdingsBuilder instanceof ZdbHoldings.Builder) {
            Holdings holdings = holdingsBuilder.build();
            if (holdings.isGreen()) {
                builder.isGreen = true;
            }
            return holdings;
        } else if (holdingsBuilder instanceof HbzAlmaItems.Builder || holdingsBuilder instanceof HbzAlmaHoldings.Builder) {
            Holdings holdings = holdingsBuilder.build();
            String isil = holdings.getISIL();
            if (isil == null) {
                holdingsBuilder.setDeleted(true);
            } else {
                // Alma over Aleph. We must be lucky to match physical items exactly.
                // But it is possible.
                this.holdings.stream()
                        .filter(it -> it instanceof HbzAlephHoldings)
                        .filter(it -> isil.equals(it.getISIL()))
                        .filter(it -> holdings.getServices().equals(it.getServices()))
                        .findFirst()
                        .ifPresent(h -> holdingsBuilder.setDeleted(true));
            }
            return holdingsBuilder.build();
        } else if (holdingsBuilder instanceof EzbHoldings.Builder) {
            Holdings holdings = holdingsBuilder.build();
            // propagate "green" from holdings to the record
            if (holdings.isGreen()) {
                builder.isGreen = true;
            }
            String isil = holdings.getISIL();
            if (isil == null) {
                holdingsBuilder.setDeleted(true);
            } else {
                // drop EZB holding if a ZDB holding of that ISIL is already present
                this.holdings.stream()
                        .filter(it -> it instanceof ZdbHoldings)
                        .filter(it -> isil.equals(it.getISIL()))
                        .findFirst()
                        .ifPresent(h -> holdingsBuilder.setDeleted(true));
            }
            return holdingsBuilder.build();
        } else if (holdingsBuilder instanceof HbzAlmaOnlineResource.Builder) {
            Holdings holdings = holdingsBuilder.build();
            if (holdings.isGreen()) {
                builder.isGreen = true;
            }
            String isil = holdings.getISIL();
            if (isil == null) {
                holdingsBuilder.setDeleted(true);
            }
            return holdingsBuilder.build();
       } else if (holdingsBuilder instanceof HbzGviHoldings.Builder) {
            // we want no GVI holdings, we ignore them
            holdingsBuilder.setDeleted(true);
            return holdingsBuilder.build();
        } else if (holdingsBuilder instanceof HbzAlephHoldings.Builder) {
            Holdings holdings = holdingsBuilder.build();
            if (holdings.isGreen()) {
                builder.isGreen = true;
            }
            String isil = holdings.getISIL();
            if (isil == null) {
                holdingsBuilder.setDeleted(true);
            } else {
                // we assume there is no EZB in Aleph which is indeed the case, so
                // physical matching is enough
                this.holdings.stream()
                        .filter(it -> it instanceof ZdbHoldings)
                        .filter(it -> isil.equals(it.getISIL()))
                        .filter(it -> holdings.getServices().equals(it.getServices()))
                        .findFirst()
                        .ifPresent(h -> holdingsBuilder.setDeleted(true));
                if (holdings.getServices().isEmpty()) {
                    holdingsBuilder.setDeleted(true);
                }
            }
            return holdingsBuilder.build();
        } else if (holdingsBuilder instanceof HbzfixHoldings.Builder) {
            return holdingsBuilder.build();
        }
        throw new UnsupportedOperationException("unknown holdings builder " + holdingsBuilder.getClass().getName());
    }

    protected boolean tryMerge(Holdings holdings) {
        if (holdings.getISIL() == null) {
            return false;
        }
        AtomicBoolean result = new AtomicBoolean();
        this.holdings.stream()
                .filter(it -> holdings.getISIL().equals(it.getISIL()))
                .findFirst()
                .ifPresent(h -> {
                    h.merge(holdings);
                    result.set(true);
                });
        return result.get();
    }

    public static abstract class Builder implements Record.Builder {

        protected Configuration configuration;

        protected MarcRecord marcRecord;

        protected String database;

        protected String identifier;

        protected String externalIdentifier;

        protected Collection<String> issns;

        protected boolean isDeleted;

        protected List<Material> materials;

        protected boolean isSerialMirror;

        protected boolean isGreen;

        protected boolean isOpenAccess;

        protected String baseUrl;

        protected Integer publicationDate;

        protected String chronology;

        protected final Collection<Product> products;

        protected final Collection<Service.LabeledUrl> globalServices;

        protected Builder(Configuration configuration,
                          String database,
                          MarcRecord marcRecord) {
            this.configuration = configuration;
            this.database = database;
            this.marcRecord = marcRecord;
            this.materials = new ArrayList<>();
            this.products = new ArrayList<>();
            this.globalServices = new ArrayList<>();
            this.issns = new ArrayList<>();
            findDeleted();
            this.materials = extractMaterial();
            if (!isDeleted) {
                findIdentifier();
                findPublicationDate();
                findProducts();
            }
        }

        public Builder setDeleted(boolean isDeleted) {
            this.isDeleted = isDeleted;
            return this;
        }

        public Builder setMaterials(List<Material> materials) {
            this.materials = materials;
            return this;
        }

        public Builder setGreen(boolean isGreen) {
            this.isGreen = isGreen;
            return this;
        }

        public Builder setOpenAccess(boolean isOpenAccess) {
            this.isOpenAccess = isOpenAccess;
            return this;
        }

        protected void findDeleted() {
            if (marcRecord != null) {
                RecordLabel recordLabel = marcRecord.getRecordLabel();
                if (recordLabel != null && recordLabel.getRecordStatus() == RecordStatus.DELETED) {
                    this.isDeleted = true;
                }
            }
        }

        protected List<Material> extractMaterial() {
            if (marcRecord == null) {
                return List.of();
            }
            return marcRecord.getAll("007").stream()
                    .map(MarcField::recoverControlFieldValue)
                    .map(this::toMaterial)
                    .toList();
        }

        private Material toMaterial(String value) {
            return switch (value.substring(0, 1)) {
                case "a" -> MAP;
                case "c" -> value.startsWith("cr") ? ONLINE_RESOURCE :ELECTRONIC_RESOURCE;
                case "d" -> GLOBE;
                case "f" -> TACTILE;
                case "g" -> PROJECTED_GRAPHIC;
                case "h" -> MICROFORM;
                case "k" -> NON_PROJECTED_GRAPHIC;
                case "m" -> MOTION_PICTURE;
                case "o" -> KIT;
                case "q" -> NOTATED_MUSIC;
                case "r" -> REMOTE_SENSING_IMAGE;
                case "s" -> SOUND_RECORDING;
                case "t" -> TEXT;
                case "v" -> VIDEO_RECORDING;
                default -> UNSPECIFIED;
            };
        }

        protected void setExternalIdentifier(String externalIdentifier) {
            this.externalIdentifier = externalIdentifier;
        }

        protected void setBaseUrl(String baseUrl) {
            this.baseUrl = baseUrl;
        }

        protected void findIdentifier() {
            if (marcRecord != null) {
                MarcField marcField = marcRecord.getFirst("001");
                if (marcField != null) {
                    this.identifier = marcField.getFirstSubfieldValue("_");
                    marcField = marcRecord.getFirst("003");
                    if (marcField != null) {
                        this.identifier = "(" + marcField.getFirstSubfieldValue("_") + ")" + identifier;
                    }
                }
            }
        }

        protected void findPublicationDate() {
            if (marcRecord != null) {
                MarcField marcField = marcRecord.getFirst("264");
                if (marcField != null) {
                    String string = marcField.getFirstSubfieldValue("c");
                    this.publicationDate = toIntegerDate(string);
                }
            }
        }

        protected void findProducts() {
            if (marcRecord != null) {
                marcRecord.getAll("912").forEach(mf -> {
                    mf.getSubfields().forEach(sf -> {
                        if ("a".equals(sf.getId())) {
                            String productName = sf.getValue();
                            Product product = configuration.getProducts().get(productName);
                            if (product != null) {
                                products.add(product);
                            }
                        }
                    });
                });
            }
        }

        protected Integer toIntegerDate(String publicationDate) {
            try {
                String string = publicationDate;
                if (string != null) {
                    string = string.replaceAll("[^0-9]+", "");
                    if (string.length() >= 4) {
                        return Integer.parseInt(string.substring(0, 4));
                    }
                }
            } catch (Exception e) {
                return null;
            }
            return null;
        }
    }
}
