package org.xbib.interlibrary.catalog.elasticsearch.lvi.record;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;

import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Service;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;

import static org.xbib.interlibrary.api.library.Material.ELECTRONIC_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.GLOBE;
import static org.xbib.interlibrary.api.library.Material.KIT;
import static org.xbib.interlibrary.api.library.Material.MAP;
import static org.xbib.interlibrary.api.library.Material.MICROFORM;
import static org.xbib.interlibrary.api.library.Material.MOTION_PICTURE;
import static org.xbib.interlibrary.api.library.Material.NON_PROJECTED_GRAPHIC;
import static org.xbib.interlibrary.api.library.Material.NOTATED_MUSIC;
import static org.xbib.interlibrary.api.library.Material.ONLINE_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.PROJECTED_GRAPHIC;
import static org.xbib.interlibrary.api.library.Material.REMOTE_SENSING_IMAGE;
import static org.xbib.interlibrary.api.library.Material.SOUND_RECORDING;
import static org.xbib.interlibrary.api.library.Material.TACTILE;
import static org.xbib.interlibrary.api.library.Material.TEXT;
import static org.xbib.interlibrary.api.library.Material.UNSPECIFIED;
import static org.xbib.interlibrary.api.library.Material.VIDEO_RECORDING;

public class HbzAlephRecord extends AbstractRecord {

    private final Builder builder;

    protected HbzAlephRecord(Builder builder) {
        super(builder);
        this.builder = builder;
    }

    public static Builder builder(Configuration configuration,
                                  MarcRecord marcRecord) {
        return new Builder(configuration, marcRecord);
    }

    public void handleResources(Consumer<MarcField> consumer) {
        getMarcRecord().all("088", consumer::accept);
    }

    public boolean isUState() {
        return builder.isUState;
    }

    public static class Builder extends AbstractRecord.Builder {

        private boolean isUState;

        protected Builder(Configuration configuration,
                          MarcRecord marcRecord) {
            super(configuration, "hbzaleph", marcRecord);
            // check if we are a mirror record of ZDB
            for (MarcField marcField : marcRecord.getAll("025")) {
                if ((marcField.getIndicator().startsWith("z"))) {
                    this.isSerialMirror = true;
                }
            }
            // check if we have "U state"
            for (MarcField marcField : marcRecord.getAll("FMT")) {
                for (MarcField.Subfield subfield : marcField.getSubfields()) {
                    if ("MU".equals(subfield.getValue())) {
                        this.isUState = true;
                    }
                }
            }
        }

        @Override
        public HbzAlephRecord build() {
            return new HbzAlephRecord(this);
        }

        @Override
        protected void findIdentifier() {
            if (marcRecord != null) {
                MarcField marcField = marcRecord.getFirst("001");
                if (marcField != null) {
                    this.identifier = "(DE-605)" + marcField.getFirstSubfieldValue("_");
                }
                this.externalIdentifier = this.identifier;
            }
        }

        @Override
        protected List<Material> extractMaterial() {
            List<Material> materials = new ArrayList<>();
            marcRecord.getAll("334").forEach(marcField -> {
                if ("Elektronische Ressource".equals(marcField.getFirstSubfieldValue("a"))) {
                    materials.add(Material.ELECTRONIC_RESOURCE);
                }
            });
            marcRecord.getAll("062").forEach(marcField -> {
                String value = marcField.getFirstSubfieldValue("b");
                materials.add(switch (value.substring(0, 1)) {
                    case "a" -> MAP;
                    case "c" -> value.startsWith("cr") ? ONLINE_RESOURCE :ELECTRONIC_RESOURCE;
                    case "d" -> GLOBE;
                    case "f" -> TACTILE;
                    case "g" -> PROJECTED_GRAPHIC;
                    case "h" -> MICROFORM;
                    case "k" -> NON_PROJECTED_GRAPHIC;
                    case "m" -> MOTION_PICTURE;
                    case "o" -> KIT;
                    case "q" -> NOTATED_MUSIC;
                    case "r" -> REMOTE_SENSING_IMAGE;
                    case "s" -> SOUND_RECORDING;
                    case "t" -> TEXT;
                    case "v" -> VIDEO_RECORDING;
                    default -> UNSPECIFIED;
                });
            });
            marcRecord.getAll("655").forEach(marcField -> {
                String label = marcField.getFirstSubfieldValue("x");
                // 655e1 = resolving, DOI etc.
                if ("e1".equals(marcField.getIndicator())) {
                    String url = marcField.getFirstSubfieldValue("u");
                    if (url != null) {
                        Service.LabeledUrl onlineService = new Service.LabeledUrl();
                        onlineService.setUrl(url);
                        onlineService.setLabel(label);
                        globalServices.add(onlineService);
                        // online = true?
                    }
                }
                // 655e9 = enrichment via publishing
                // Ebsco etc. Example: HT019117887
                if ("e9".equals(marcField.getIndicator())) {
                    String url = marcField.getFirstSubfieldValue("u");
                    if (url != null) {
                        Service.LabeledUrl onlineService = new Service.LabeledUrl();
                        onlineService.setUrl(url);
                        label = marcField.getFirstSubfieldValue("z");
                        onlineService.setLabel(label);
                        globalServices.add(onlineService);
                        materials.add(ONLINE_RESOURCE);
                    }
                }
                if ("EZB".equals(label)) {
                    materials.add(ONLINE_RESOURCE);
                }
            });
            return materials;
        }

        @Override
        protected void findPublicationDate() {
            // MAB (pre-RDA)
            // 425a Erscheinungsdatum des Titels
            marcRecord.getAll("425").forEach(marcField -> {
                if ("a1".equals(marcField.getIndicator())) {
                    this.publicationDate = toIntegerDate(marcField.getFirstSubfieldValue("a"));
                }
            });
            if (publicationDate == null) {
                // MAB 425b
                // 425a Erscheinungsdatum des ersten Bandes
                marcRecord.getAll("425").forEach(marcField -> {
                    if ("b1".equals(marcField.getIndicator())) {
                        this.publicationDate = toIntegerDate(marcField.getFirstSubfieldValue("a"));
                    }
                });
            }
            if (publicationDate == null) {
                // RDA 264
                marcRecord.getAll("264").forEach(marcField -> {
                    this.publicationDate = toIntegerDate(marcField.getFirstSubfieldValue("c"));
                });
            }
            if (materials.contains(ONLINE_RESOURCE)) {
                // MAB 405 Erscheinungsverlauf
                marcRecord.getAll("405", " 1").forEach(marcField -> {
                    this.chronology = marcField.getFirstSubfieldValue("a");
                });
            }
        }
    }
}
