package org.xbib.interlibrary.catalog.elasticsearch.lvi.record;

import java.util.Objects;

import org.xbib.datastructures.common.LinkedHashSetMultiMap;
import org.xbib.datastructures.common.MultiMap;
import org.xbib.interlibrary.api.library.IndexLibrary;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Service;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.service.DefaultService;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;

public final class HbzAlmaItems extends AbstractHoldings {

    private static final MultiMap<String, String> unknownSublibraries = new LinkedHashSetMultiMap<>();

    private HbzAlmaItems(Builder builder) {
        super();
        this.builder = builder;
    }

    public static MultiMap<String, String> getUnknownSublibraries() {
        return unknownSublibraries;
    }

    public static Builder builder(Configuration configuration,
                                  HbzAlmaRecord hbzAlmaRecord,
                                  MarcField marcField) {
        return new Builder(configuration, hbzAlmaRecord, marcField);
    }

    @Override
    protected String getIndicator() {
        return "0 ";
    }

    @Override
    protected void customSubfields(MarcField.Builder marcFieldBuilder) {
    }

    public static class Builder extends AbstractHoldings.Builder {

        protected Builder(Configuration configuration,
                          HbzAlmaRecord hbzAlmaRecord,
                          MarcField marcField) {
            super(configuration, hbzAlmaRecord, marcField);
            Service service = new DefaultService();
            service.setMaterial(material);
            if (fullIsil != null) {
                service.setISIL(fullIsil);
            } else {
                service.setISIL(isil);
            }
            // evaluate "u", "v", "c", "n"
            // Example: "EFB" should map service to DE-290-EFB
            String location = marcField.getFirstSubfieldValue("u");
            setSubLibrary(service, location);
            String collection = marcField.getFirstSubfieldValue("v");
            service.setCollection(collection);
            String callnumber = marcField.getFirstSubfieldValue("c");
            service.setCallnumber(callnumber);
            String secondCallNumber = marcField.getFirstSubfieldValue("n"); // zweite Signatur
            if (callnumber == null && secondCallNumber != null) {
                // move to first call number if possible
                service.setCallnumber(secondCallNumber);
            } else {
                // move to description otherwise
                service.setDescription(secondCallNumber);
            }
            StringBuilder statusText = new StringBuilder();
            StringBuilder restrictionText = new StringBuilder();
            StringBuilder comment = new StringBuilder();
            statusText.append(marcField.getFirstSubfieldValue("p"));
            // try to find "H92" value
            final String id = marcField.getFirstSubfieldValue("H");
            if (id != null) {
                hbzAlmaRecord.getMarcRecord().all("H92", mf -> {
                    if (id.equals(mf.getFirstSubfieldValue("8"))) {
                        mf.getSubfields().forEach(sf -> {
                            if (sf.getValue() != null) {
                                if ("o".equals(sf.getId())) {
                                    statusText.append(sf.getValue());
                                } else if ("p".equals(sf.getId())) {
                                    restrictionText.append(sf.getValue());
                                } else if ("q".equals(sf.getId())) {
                                    comment.append(sf.getValue());
                                }
                            }
                        });
                    }
                });
            }
            setStatusCode(service, statusText.toString(),
                    restrictionText.toString(), comment.toString());
            // q = Bestandsverlauf
            String textualHoldings = marcField.getFirstSubfieldValue("q");
            if (textualHoldings != null) {
                service.setTextualHoldings(textualHoldings);
            }
            // Alma (monograph) ITM does not have textual holdings, we use record publication date if present
            if (service.getTextualHoldings() == null) {
                if (record.getPublicationDate() != null) {
                    service.setTextualHoldings(record.getPublicationDate().toString());
                }
            }
            String processType = marcField.getFirstSubfieldValue("r");
            if (processType != null) {
                service.setItemProcessCode(processType); // for statistics
                String s = configuration.mapItemProcess(processType);
                String oldDescription = service.getDescription();
                if (oldDescription == null) {
                    service.setDescription(s);
                } else {
                    service.setDescription(oldDescription + ", " + s);
                }
            }
            service.computeYears();
            services.add(service);
        }

        @Override
        protected String findISILCandidate(MarcRecord marcRecord, MarcField marcfield) {
            String code = marcfield.getLastSubfieldValue("M");
            // special code for libraries without cataloging
            // "49" Exlibris code for Germany, "HBZ" Hochschulbibliothekszentrum, "FIZ" = "frozen inventory zone"
            if ("49HBZ_FIZ".contentEquals(code) || "49HBZ_BRIDGE_FIZ".contentEquals(code)) {
                // u = Aleph Owner, take precedence over "M"
                String u = marcfield.getLastSubfieldValue("u");
                if (u == null) {
                    // try H52 $b to take precedence over "M"
                    StringBuilder sb = new StringBuilder();
                    marcRecord.getAll("H52").forEach(mf -> {
                        String b = mf.getLastSubfieldValue("b");
                        if (b != null) {
                            sb.append(b);
                        }
                    });
                    if (!sb.isEmpty()) {
                        return sb.toString();
                    }
                } else {
                    return u;
                }
            }
            return code;
        }

        @Override
        protected IndexLibrary fixLibrary(IndexLibrary library,
                                          MarcRecord marcRecord,
                                          MarcField marcfield) {
            String extraAlmaCode = marcfield.getLastSubfieldValue("u");
            if (extraAlmaCode == null) {
                return library;
            }
            // we have a custom ISIL mapping configured in "sublibraries"
            if (library.getSubLibraries() != null) {
                if (library.getSubLibraries().containsKey(extraAlmaCode)) {
                    IndexLibrary extraLibrary = configuration.getLibraries().get(library.getSubLibraries().get(extraAlmaCode).toString());
                    if (extraLibrary != null) {
                        return extraLibrary;
                    }
                } else {
                    unknownSublibraries.put(library.getISIL(), extraAlmaCode);
                }
            }
            IndexLibrary extraLibrary = configuration.getLibraries().get(extraAlmaCode);
            return Objects.requireNonNullElse(extraLibrary, library);
        }

        @Override
        public HbzAlmaItems build() {
            return new HbzAlmaItems(this);
        }
    }
}
