package org.xbib.interlibrary.catalog.elasticsearch.lvi.record;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.TreeSet;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.interlibrary.catalog.entities.ResourceLoader;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;
import org.xbib.standardnumber.ISBN;

public class HbzAlmaRecord extends AbstractRecord {

    private static final Pattern FIELD_KEYS = Pattern.compile("^\\d\\d\\d\\$[ 0-9a-z]+\\$[0-9a-zA-Z_]+$");

    private static final Map<String, Object> relatorCodes =
            new ResourceLoader().loadResourceMap("org/xbib/interlibrary/catalog/elasticsearch/hbzalma/relator-codes-de.json");

    private static final Set<String> unknownRelatorCodes = new HashSet<>();

    private static final Set<String> unknownRdaCodes = new HashSet<>();

    protected HbzAlmaRecord(Builder builder) {
        super(builder);
    }

    public static Builder builder(Configuration configuration,
                                  MarcRecord marcRecord) {
        return new Builder(configuration, marcRecord);
    }

    @Override
    public String getKey() {
        return getIdentifier() != null ? getIdentifier() : "";
    }

    public void handleItmResources(Consumer<MarcField> consumer) {
        getMarcRecord().all("ITM", consumer::accept);
    }

    public void handleHolResources(Consumer<MarcField> consumer) {
        getMarcRecord().all("HOL", consumer::accept);
    }

    public MarcRecord getFilteredMarcRecord(List<MarcField> list,
                                            Comparator<MarcField> fieldComparator) {
        MarcRecord marcRecord = getMarcRecord();
        MarcField marcField = marcRecord.getFirst("003");
        String prefix = marcField != null ? marcField.getFirstSubfieldValue("_") : "DE-655";
        Collection<MarcField> newList = Stream.concat(marcRecord.getFields().stream()
                                .filter(predicate())
                                .map(mf -> mapField(marcRecord, mf))
                                .filter(Objects::nonNull),
                        extraFields(marcRecord, list, prefix).stream())
                .collect(Collectors.toCollection(() -> new TreeSet<>(fieldComparator)));
        return new MarcRecord(marcRecord.getFormat(),
                marcRecord.getType(),
                marcRecord.getRecordLabel(),
                newList, false, Comparator.naturalOrder());
    }

    public static Set<String> getUnknownRelatorCodes() {
        return unknownRelatorCodes;
    }

    public static Set<String> getUnknownRdaCodes() {
        return unknownRdaCodes;
    }

    private static Predicate<MarcField> predicate() {
        return marcField -> {
            String key = marcField.toKey();
            return FIELD_KEYS.matcher(key).matches();
        };
    }

    private static MarcField mapField(MarcRecord marcRecord, MarcField marcField) {
        switch (marcField.getTag()) {
            case "001" -> {
                // move "HT Nummer" up from 035 to 001
                for (MarcField mf : marcRecord.getAll("035")) {
                    for (MarcField.Subfield sf : mf.getSubfields()) {
                        String id = sf.getId();
                        String value = sf.getValue();
                        if ("a".equals(id) && value != null && value.startsWith("(DE-605)")) {
                            return MarcField.builder()
                                    .tag("001").indicator("  ").subfield(" ", stripPrefixIfPresent(value))
                                    .build();
                        }
                    }
                }
                // if no HT-Nummer found, keep ALma MMS-ID in 001
                return MarcField.builder()
                        .tag("001").indicator("  ").subfield(" ", marcField.getFirstSubfieldValue("_"))
                        .build();
            }
            case "003" -> {
                // delete prefix
                return null;
            }
            case "005", "007" -> {
                // fix indicator/subfield
                return MarcField.builder()
                        .tag(marcField.getTag()).indicator("  ").subfield(" ", marcField.getFirstSubfieldValue("_"))
                        .build();
            }
            case "008" -> {
                // we do not tolerate '#' in 008
                return MarcField.builder()
                        .tag("008").indicator("  ").subfield(" ", marcField.getFirstSubfieldValue("_").replace('#',' '))
                        .build();
            }
            case "100", "110", "111", "700", "710", "751" -> {
                // clean "a" subfield from interpunctuation
                // undo Alma expansions
                // expand "4" to "e" subfield
                // keep first "0" subfield, but delete all others
                String name = marcField.getFirstSubfieldValue("a");
                String gndIdentifier = marcField.getFirstSubfieldValue("0");
                String relatorCode = relatorCode(marcField.getFirstSubfieldValue("4"));
                return MarcField.builder().marcField(marcField)
                        .replace("a", cleanOffendingCharacters(name))
                        .subfieldIfNotNull("e", relatorCode)
                        .removeSubfields(Set.of("0", "9", "B", "M"))
                        .subfieldIfNotNull("0", gndIdentifier)
                        .build();
            }
            case "240", "245" -> {
                // clean "a" subfield from interpunctuation
                // undo Alma expansions
                String title = marcField.getFirstSubfieldValue("a");
                String gndIdentifier = marcField.getFirstSubfieldValue("0");
                return MarcField.builder().marcField(marcField)
                        .replace("a", cleanOffendingCharacters(title))
                        .removeSubfields(Set.of("0", "9", "B", "M"))
                        .subfieldIfNotNull("0", gndIdentifier)
                        .build();
            }
            case "336" -> {
                String rdaCode = marcField.getFirstSubfieldValue("b");
                String rdaTranslation = translateRDACode(rdaCode);
                if (rdaTranslation != null) {
                    return MarcField.builder()
                            .tag(marcField.getTag())
                            .indicator(marcField.getIndicator())
                            .subfield("2", "rdacontent")
                            .subfield("a", rdaTranslation)
                            .subfield("b", rdaCode)
                            .build();
                } else {
                    unknownRdaCodes.add(rdaCode);
                    return marcField;
                }
            }
            case "337" -> {
                String rdaCode = marcField.getFirstSubfieldValue("b");
                String rdaTranslation = translateRDACode(rdaCode);
                if (rdaTranslation != null) {
                    return MarcField.builder()
                            .tag(marcField.getTag())
                            .indicator(marcField.getIndicator())
                            .subfield("2", "rdamedia")
                            .subfield("a", rdaTranslation)
                            .subfield("b", rdaCode)
                            .build();
                } else {
                    unknownRdaCodes.add(rdaCode);
                    return marcField;
                }
            }
            case "338" -> {
                String rdaCode = marcField.getFirstSubfieldValue("b");
                String rdaTranslation = translateRDACode(rdaCode);
                if (rdaTranslation != null) {
                    return MarcField.builder()
                            .tag(marcField.getTag())
                            .indicator(marcField.getIndicator())
                            .subfield("2", "rdacarrier")
                            .subfield("a", rdaTranslation)
                            .subfield("b", rdaCode)
                            .build();
                } else {
                    unknownRdaCodes.add(rdaCode);
                    return marcField;
                }
            }
            case "500", "830", "856" -> {
                return MarcField.builder().marcField(marcField)
                        .removeSubfields(Set.of("9", "M"))
                        .build();
            }
            case "590", "912", "935", "960", "962", "964", "980", "981", "983", "987", "991", "995", "996", "997" -> {
                // drop "local fields"
                return null;
            }
            case "689" -> {
                // keep first "0" subfield, but delete all others
                String gndIdentifier = marcField.getFirstSubfieldValue("0");
                return MarcField.builder().marcField(marcField)
                        .removeSubfields(Set.of("0", "B", "M"))
                        .subfieldIfNotNull("0", gndIdentifier)
                        .build();
            }
            default -> {
                return MarcField.builder().marcField(marcField)
                        .removeSubfields(Set.of("M"))
                        .build();
            }
        }
    }

    private static String cleanOffendingCharacters(String string) {
        // do not return null
        // u00ac = ¬
        return string == null ? EMPTY : string.trim()
                .replaceAll("<<(.*?)>>", "\u0098$1\u009C")
                .replaceAll("<(.*?)>", "[$1]")
                .replaceAll("¬(.*?)¬", "\u0098$1\u009C");
    }

    private static final String EMPTY = "";

    private static String relatorCode(String code) {
        if (relatorCodes.containsKey(code)) {
            return relatorCodes.get(code).toString();
        } else if (code != null) {
            synchronized (unknownRelatorCodes) {
                unknownRelatorCodes.add(code);
            }
        }
        return null;
    }

    private static List<MarcField> extraFields(MarcRecord marcRecord,
                                               List<MarcField> marcFields,
                                               String prefix) {
        List<MarcField> list = new ArrayList<>(marcFields);
        String id = marcRecord.getFirst("001").getFirstSubfieldValue("_");
        // move ID to foreign identifiers, prefixed by "(DE-655)"
        list.add(MarcField.builder()
                .tag("035").indicator("  ").subfield("a", "(DE-655)" + id)
                .build());
        list.add(MarcField.builder()
                .tag("003").indicator("  ").subfield(" ", prefix)
                .build());
        // add ISBN-10/ISBN-13 formats if not exists
        Collection<String> isbns = extract(marcRecord, "020", null, List.of("a"));
        for (String isbn : createExtraISBNs(isbns)) {
            list.add(MarcField.builder()
                    .tag("020").indicator("  ").subfield("a", isbn)
                    .build());
        }
        return list;
    }

    private static String stripPrefixIfPresent(String identifier) {
        if (identifier == null) {
            return null;
        }
        if (identifier.indexOf('(') == 0) {
            int pos = identifier.lastIndexOf(')');
            return identifier.substring(pos + 1);
        }
        return identifier;
    }

    public void handlePortfolioResources(Consumer<MarcField> consumer) {
        getMarcRecord().all("POR", marcField -> {
            if (marcField.getSubfieldIds().contains("A")) {
                for (String value : marcField.getSubfieldValues(List.of("A"))) {
                    MarcField expanded = MarcField.builder()
                            .marcField(marcField)
                            .subfield("M", value)
                            .build();
                    consumer.accept(expanded);
                }
            } else {
                consumer.accept(marcField);
            }
        });
    }

    public void handlePortfolioCoverage(Consumer<MarcField> chronologyConsumer) {
        getMarcRecord().all("POC", chronologyConsumer::accept);
    }

    private static String translateRDACode(String code) {
        if (code == null) {
            return null;
        }
        return switch (code) {
            case "s" -> "Audio";
            case "c" -> "Computermedien";
            case "h" -> "Mikroform";
            case "p" -> "mikroskopisch";
            case "n" -> "ohne Hilfsmittel zu benutzen";
            case "g" -> "projizierbar";
            case "e" -> "stereografisch";
            case "v" -> "Video";
            case "x" -> "sonstige Medien";
            case "z" -> "nicht spezifiziert";
            case "sg" -> "Audio-Cartridge";
            case "sd" -> "Audiodisk";
            case "ss" -> "Audiokassette";
            case "sq" -> "Notenrolle";
            case "se" -> "Phonographenzylinder";
            case "st" -> "Tonbandspule";
            case "si" -> "Tonspurspule";
            case "sz" -> "Sonstige Audiomedien";
            case "cb" -> "Computerchip-Cartridge";
            case "cd" -> "Computerdisk";
            case "ce" -> "Computerdisk-Cartridge";
            case "ca" -> "Magnetbandcartridge";
            case "cf" -> "Magnetbandkassette";
            case "ch" -> "Magnetbandspule";
            case "cr" -> "Online-Ressource";
            case "ck" -> "Speicherkarte";
            case "cz" -> "Sonstige Computermedien";
            case "hg" -> "Lichtundurchlässiger Mikrofiche";
            case "he" -> "Mikrofiche";
            case "hf" -> "Mikrofichekassette";
            case "hb" -> "Mikrofilm-Cartridge";
            case "hc" -> "Mikrofilmkassette";
            case "ha" -> "Mikrofilmlochkarte";
            case "hj" -> "Mikrofilmrolle";
            case "hd" -> "Mikrofilmspule";
            case "hh" -> "Mikrofilmstreifen";
            case "hz" -> "Sonstige Mikroformen";
            case "pp" -> "Objektträger";
            case "pz" -> "Sonstige Mikroskop-Anwendungen";
            case "gs" -> "Dia";
            case "mc" -> "Filmdose";
            case "mf" -> "Filmkassette";
            case "mo" -> "Filmrolle";
            case "mr" -> "Filmspule";
            case "gf" -> "Filmstreifen";
            case "gd" -> "Filmstreifen für Einzelbildvorführung";
            case "gc" -> "Filmstreifen-Cartridge";
            case "gt" -> "Overheadfolie";
            case "mz" -> "Sonstige projizierbare Bilder";
            case "eh" -> "Stereobild";
            case "es" -> "Stereografische Disk";
            case "ez" -> "Sonstige stereografische Datenträger";
            case "nc" -> "Band";
            case "nb" -> "Blatt";
            case "nn" -> "Flipchart";
            case "nr" -> "Gegenstand";
            case "no" -> "Karte";
            case "na" -> "Rolle";
            case "nz" -> "Sonstige Datenträger ohne Hilfsmittel";
            case "vr" -> "Videobandspule";
            case "vc" -> "Videocartridge";
            case "vd" -> "Videodisk";
            case "vf" -> "Videokassette";
            case "vz" -> "sonstige Videomedien";
            case "zu" -> "nicht spezifiziert";
            case "prm" -> "aufgeführte Musik";
            case "ntv" -> "Bewegungsnotation";
            case "cod" -> "Computerdaten";
            case "cop" -> "Computerprogramm";
            case "tdf" -> "dreidimensionale Form";
            case "tdm" -> "dreidimensionales bewegtes Bild";
            case "snd" -> "Geräusche";
            case "spw" -> "gesprochenes Wort";
            case "crf" -> "kartografische dreidimensionale Form";
            case "crn" -> "kartografische taktile dreidimensionale Form";
            case "crd" -> "kartografischer Datensatz";
            case "crm" -> "kartografisches bewegtes Bild";
            case "cri" -> "kartografisches Bild";
            case "crt" -> "kartografisches taktiles Bild";
            case "ntm" -> "Noten";
            case "tcn" -> "taktile Bewegungsnotation";
            case "tcf" -> "taktile dreidimensionale Form";
            case "tcm" -> "taktile Noten";
            case "tct" -> "taktiler Text";
            case "tci" -> "taktiles Bild";
            case "txt" -> "Text";
            case "sti" -> "unbewegtes Bild";
            case "tdi" -> "zweidimensionales bewegtes Bild";
            case "xxx" -> "Sonstige";
            case "zzz" -> "nicht spezifiziert";
            default -> "";
        };
    }

    private static Collection<String> extract(MarcRecord marcRecord,
                                              String field,
                                              List<String> indicators,
                                              List<String> subfieldIds) {
        Set<String> set = new LinkedHashSet<>();
        marcRecord.getAll(field).stream()
                .filter(mf -> indicators == null || indicators.contains(mf.getIndicator())).forEach(f ->
                        f.getSubfields().stream()
                                .filter(sf -> subfieldIds == null || subfieldIds.contains(sf.getId())).forEach(sf ->
                                        set.add(sf.getValue())));
        return set;
    }

    private static Collection<String> createExtraISBNs(Collection<String> isbns) {
        if (isbns == null) {
            return List.of();
        }
        List<String> list = new ArrayList<>();
        for (String isbnString : isbns) {
            if (isbnString == null || isbnString.isEmpty()) {
                continue;
            }
            ISBN isbn = new ISBN();
            isbn.ean(false).set(isbnString).normalize();
            if (isbn.isValid()) {
                String isbn10 = isbn.normalizedValue();
                if (!isbns.contains(isbn10)) {
                    list.add(isbn10);
                }
                isbn10 = isbn.format();
                if (!isbns.contains(isbn10)) {
                    list.add(isbn10);
                }
                if (!isbn.isEAN()) {
                    isbn = isbn.ean(true);
                    isbn.set(isbnString).normalize();
                    String isbn13 = isbn.normalizedValue();
                    if (!isbns.contains(isbn13)) {
                        list.add(isbn13);
                    }
                    isbn13 = isbn.format();
                    if (!isbns.contains(isbn13)) {
                        list.add(isbn13);
                    }
                }
            }
        }
        return list;
    }

    public static class Builder extends AbstractRecord.Builder {

        protected Builder(Configuration configuration,
                          MarcRecord marcRecord) {
            super(configuration, "hbzalma", marcRecord);
        }

        @Override
        public HbzAlmaRecord build() {
            return new HbzAlmaRecord(this);
        }

        @Override
        protected void findIdentifier() {
            MarcField marcField = marcRecord.getFirst("001");
            if (marcField != null) {
                // enforce DE-655 prefix. 003 contains "DE-605" which is wrong.
                this.identifier = "(DE-655)" + marcField.getFirstSubfieldValue("_");
            }
            // now, set the external identifier
            // first, try to find ZDB ID
            marcRecord.getAll("035").stream()
                    .map(mf -> mf.getFirstSubfieldValue("a"))
                    .filter(Objects::nonNull)
                    .filter(s -> s.startsWith("(DE-599)ZDB"))
                    .findFirst()
                    .ifPresent(this::setExternalIdentifier);
            // second, set external identifier to the old Aleph HBZ-ID if exists
            if (externalIdentifier == null) {
                marcRecord.getAll("035").stream()
                        .map(mf -> mf.getFirstSubfieldValue("a"))
                        .filter(Objects::nonNull)
                        .filter(s -> s.startsWith("(DE-605)"))
                        .map(s -> s.substring(8))
                        .findFirst()
                        .ifPresent(this::setExternalIdentifier);
            }
        }
    }
}
