package org.xbib.interlibrary.catalog.elasticsearch.lvi.record;

import java.util.List;

import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Service;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.service.DefaultService;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;
import org.xbib.marc.label.RecordStatus;

public final class ZdbHoldings extends AbstractHoldings {

    private ZdbHoldings(Builder builder) {
        super();
        this.builder = builder;
    }

    public static Builder builder(Configuration configuration,
                                  ZdbRecord zdbRecord,
                                  MarcRecord marcRecord) {
        return new Builder(configuration, zdbRecord, marcRecord);
    }

    @Override
    protected String getIndicator() {
        return builder.material == Material.ONLINE_RESOURCE ? "1 " : "0 ";
    }

    @Override
    protected void customSubfields(MarcField.Builder marcFieldBuilder) {
        builder.subfieldList.forEach(marcFieldBuilder::subfield);
    }

    public static class Builder extends AbstractHoldings.Builder {

        protected Builder(Configuration configuration,
                          ZdbRecord zdbRecord,
                          MarcRecord marcRecord) {
            super(configuration, zdbRecord, null);
            if (marcRecord.getRecordLabel().getRecordStatus() == RecordStatus.DELETED) {
                isDeleted = true;
            }
            if (!isDeleted) {
                Service service = new DefaultService();
                service.setMaterial(material);
                StringBuilder sb = new StringBuilder();
                marcRecord.getAll("852").forEach(marcField -> {
                    if (marcField.getIndicator().equals("  ")) {
                        String candidate = marcField.getFirstSubfieldValue("a");
                        if (candidate != null) {
                            setISIL(configuration, marcField);
                            if (fullIsil != null) {
                                service.setISIL(fullIsil);
                            } else {
                                service.setISIL(isil);
                            }
                        }
                    } else if (marcField.getIndicator().equals(" 1")) {
                        String collection = marcField.getFirstSubfieldValue("b");
                        if (collection != null) {
                            service.setCollection(collection);
                            setSubLibrary(service, collection);
                        }
                        String callnumber = marcField.getFirstSubfieldValue("c");
                        if (callnumber != null) {
                            service.setCallnumber(callnumber);
                        }
                    } else if (marcField.getIndicator().equals(" 2")) {
                        sb.setLength(0);
                        // indicator " 2" = Sonderstandortsignatur
                        String collection = marcField.getFirstSubfieldValue("b");
                        if (collection != null) {
                            sb.append(collection);
                        }
                        // second callnumber (rarely used)
                        String callnumber = marcField.getFirstSubfieldValue("c");
                        if (callnumber != null) {
                            if (!sb.isEmpty()) {
                                sb.append(" ");
                            }
                            sb.append(callnumber);
                        }
                        // copy to collection if no collection present (for matching ZDB and Aleph entries)
                        if (service.getCollection() == null) {
                            service.setCollection(sb.toString());
                        } else {
                            service.setCallnumber2(sb.toString());
                        }
                    }
                });
                MarcField mf = marcRecord.getFirst("001");
                if (mf != null) {
                    setIdentifier("(" + isil + ")" + mf.getFirstSubfield().getValue());
                }
                // Example: "KR Süd" should map service to DE-829-KR-Sued
                setSubLibrary(service, sb.toString());
                // electronic resources, URLs
                marcRecord.getAll("856").forEach(marcField -> {
                    String url = marcField.getFirstSubfieldValue("u");
                    String label = marcField.getFirstSubfieldValue("x");
                    // shorten label
                    if ("Zugang via EZB".equals(label) || "EZB Link".equals(label)) {
                        label = "EZB";
                    }
                    Service.LabeledUrl labeledUrl = new Service.LabeledUrl();
                    labeledUrl.setUrl(url);
                    labeledUrl.setLabel(label);
                    service.addUrl(labeledUrl);
                });
                // structured holdings
                marcRecord.getAll("859").forEach(marcField -> {
                    // moving wall e.g. "-003Y"
                    String movingWall = marcField.getFirstSubfieldValue("y");
                    if (movingWall != null) {
                        service.setMovingWall(movingWall);
                    }
                    if (marcField.getIndicator().startsWith("0")) { // begin group
                        if (!subfieldList.isEmpty()) {
                            subfieldList.add(MarcField.subfield("w", ";"));
                        }
                        // begin volume
                        String beginVolume = marcField.getFirstSubfieldValue("a");
                        if (beginVolume != null) {
                            // m = Bandzählung in Beginngruppe
                            subfieldList.add(MarcField.subfield("m", beginVolume));
                        }
                        // begin year
                        String beginYear = marcField.getFirstSubfieldValue("i");
                        if (beginYear != null) {
                            // q = Berichtsjahr
                            subfieldList.add(MarcField.subfield("q", beginYear));
                        }
                        if (marcField.getIndicator().endsWith("1")) { // continuing
                            subfieldList.add(MarcField.subfield("x", "-"));
                        }
                    }
                    if (marcField.getIndicator().startsWith("1")) { // end group
                        // end volume
                        String endVolume = marcField.getFirstSubfieldValue("a");
                        if (endVolume != null) {
                            // r = Bandzählung in Endegruppe
                            subfieldList.add(MarcField.subfield("r", endVolume));
                        }
                        // end year
                        String endYear = marcField.getFirstSubfieldValue("i");
                        if (endYear != null) {
                            // v = Berichtsjahr in Endegruppe
                            subfieldList.add(MarcField.subfield("v", endYear));
                        }
                    }
                });
                // textual holdings. We may have null or many occurences of 866, look for indicator 30
                StringBuilder textualholdings = new StringBuilder();
                StringBuilder description = new StringBuilder();
                marcRecord.getAll("866").forEach(marcField -> {
                    if ("30".equals(marcField.getIndicator())) {
                        // 9 = "einleitender Text" ("Index", "Supplement", "Sonderheft zu", ...)
                        String type = marcField.getFirstSubfieldValue("9");
                        if (type != null) {
                            textualholdings.append(type).append(" ");
                        }
                        String string = marcField.getFirstSubfieldValue("a");
                        if (string != null) {
                            textualholdings.append(string);
                        }
                        string = marcField.getFirstSubfieldValue("z");
                        if (string != null) {
                            description.append(string);
                        }
                    } else if (marcField.getIndicator().endsWith("0")) {
                        String string = marcField.getFirstSubfieldValue("z");
                        if (string != null) {
                            description.append(string);
                        }
                    }
                    String license = marcField.getFirstSubfieldValue("x");
                    if (license != null) {
                        service.setCollection(license);
                    }
                });
                if (!textualholdings.isEmpty()) {
                    service.setTextualHoldings(textualholdings.toString());
                }
                if (!description.isEmpty()) {
                    service.setDescription(description.toString());
                }
                // maybe textual holdings not set, so we try to reconstruct it from 859
                if (service.getTextualHoldings() == null) {
                    String string = printHoldings(marcRecord.getAll("859"));
                    service.setTextualHoldings(string);
                }
                // maybe status not set, look up 092 $o
                if (service.getFirstStatusCode() == null) {
                    StringBuilder statusText = new StringBuilder();
                    StringBuilder restrictionText = new StringBuilder();
                    StringBuilder comment = new StringBuilder();
                    marcRecord.getAll("092").forEach(marcField -> {
                        marcField.getSubfields().forEach(sf -> {
                            if (sf.getValue() != null) {
                                // $o = Fernleihindikator (NW)
                                //    a = Fernleihe (nur Ausleihe)
                                //    b = Fernleihe (nur Kopie)
                                //    c = Fernleihe (Kopie und Ausleihe)
                                //    d = keine Fernleihe
                                //    e = Fernleihe (Kopie, elektronischer Versand an Endnutzer möglich)
                                //    x = Füllzeichen : keine Einschränkung der Fernleih-Verfügbarkeit
                                // $p = Vertragsrechtl. Einschränkungen bei der Fernleihe (W)
                                //    n = Fernleihe nur im Inland
                                //    p = elektron. Übertragungsweg zwischen Bibliotheken vertraglich ausgeschlossen
                                //    x = keine vertragl. Einschränkung
                                if ("o".equals(sf.getId())) {
                                    statusText.append(sf.getValue());
                                } else if ("p".equals(sf.getId())) {
                                    restrictionText.append(sf.getValue());
                                } else if ("q".equals(sf.getId())) {
                                    comment.append(sf.getValue());
                                }
                            }
                        });
                    });
                    setStatusCode(service, statusText.toString(),
                            restrictionText.toString(), comment.toString());
                }
                service.computeYears();
                updateModeByInterlibraryRelevance(service, marcRecord);
                services.add(service);
            }
        }

        @Override
        public ZdbHoldings build() {
            return new ZdbHoldings(this);
        }

        private String printHoldings(List<MarcField> list) {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < list.size(); i++) {
                if (!sb.isEmpty()) {
                    sb.append("; ");
                }
                MarcField mf = list.get(i);
                String volume = mf.getFirstSubfieldValue("a");
                String year = mf.getFirstSubfieldValue("i");
                format(sb, volume, year);
                if (mf.getIndicator().endsWith("1")) { // begin group with continuation
                    sb.append(" - ");
                }
                if (i + 1 < list.size()) {
                    mf = list.get(i + 1);
                    if (mf.getIndicator().startsWith("1")) { // end group
                        volume = mf.getFirstSubfieldValue("a");
                        year = mf.getFirstSubfieldValue("i");
                        sb.append(" - ");
                        format(sb, volume, year);
                        i++;
                    }
                }
            }
            return sb.toString();
        }

        private void format(StringBuilder sb, String volume, String year) {
            if (volume != null && year != null) {
                sb.append(volume).append(".").append(year);
            } else {
                if (year != null) {
                    sb.append(year);
                }
            }
        }
    }
}
