package org.xbib.interlibrary.catalog.elasticsearch.lvi.record;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.TreeSet;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Service;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;

public class ZdbRecord extends AbstractRecord {

    private boolean isOpenAccess;

    private ZdbRecord(Builder builder) {
        super(builder);
    }

    public static Builder builder(Configuration configuration,
                                  MarcRecord marcRecord) {
        return new Builder(configuration, marcRecord);
    }

    public MarcRecord getFilteredMarcRecord(List<MarcField> list,
                                            Comparator<MarcField> fieldComparator) {
        MarcRecord marcRecord = getMarcRecord();
        Collection<MarcField> newList = Stream.concat(marcRecord.getFields().stream()
                                .map(this::mapField)
                                .filter(Objects::nonNull),
                        extraFields(marcRecord, list).stream())
                .collect(Collectors.toCollection(() -> new TreeSet<>(fieldComparator)));
        return new MarcRecord(marcRecord.getFormat(),
                marcRecord.getType(),
                marcRecord.getRecordLabel(),
                newList, false, Comparator.naturalOrder());
    }

    public Collection<String> getISSNs() {
        return builder.issns;
    }

    public void setOpenAccess(boolean isOpenAccess) {
        this.isOpenAccess = isOpenAccess;
    }

    @Override
    public boolean isOpenAccess() {
        return isOpenAccess;
    }

    private MarcField mapField(MarcField marcField) {
        if ("001".equals(marcField.getTag())) {
            // move ZDB-ID up from 016 to 001
            return MarcField.builder()
                    .tag("001").indicator("  ").subfield(" ", getExternalIdentifier())
                    .build();
        } else if ("003".equals(marcField.getTag())) {
            // delete "003", we use our own
            return null;
        } else if (marcField.getTag().startsWith("00")) {
            // we have a single blank indicator and "_" subfield
            // replace indicator and subfield with ones that are compatible with Elasticsearch
            MarcField.Builder builder = MarcField.builder()
                    .tag(marcField.getTag()).indicator("  ");
            for (MarcField.Subfield subfield : marcField.getSubfields()) {
                builder.subfield(" ", subfield.getValue());
            }
            return builder.build();
        }
        return marcField;
    }

    private List<MarcField> extraFields(MarcRecord marcRecord, List<MarcField> marcFields) {
        // move 001 to 035 as DNB-ID, change 003 to DE-600 = "ZDB"
        List<MarcField> list = new ArrayList<>(marcFields);
        String dnbIdentifier = marcRecord.getFirst("001").getFirstSubfieldValue("_");
        if (dnbIdentifier != null) {
            list.add(MarcField.builder()
                    .tag("003").indicator("  ").subfield(" ", "DE-600") // DE-600 = "ZDB"
                    .build());
            list.add(MarcField.builder()
                    .tag("035").indicator("  ").subfield("a", "(DE-101)" + dnbIdentifier)
                    .build());
            list.add(MarcField.builder()
                    .tag("035").indicator("  ").subfield("a", "(DE-600)" + getExternalIdentifier())
                    .build());
        }
        return list;
    }

    public static class Builder extends AbstractRecord.Builder {

        protected Builder(Configuration configuration,
                          MarcRecord marcRecord) {
            super(configuration, "zdb", marcRecord);
        }

        @Override
        public ZdbRecord build() {
            return new ZdbRecord(this);
        }

        @Override
        protected void findIdentifier() {
            // lookup external identifier (ZDB-ID) from MARC 016. We may have no ZDB-ID.
            marcRecord.getAll("016").stream()
                    .filter(mf -> "DE-600".equals(mf.getFirstSubfieldValue("2")))
                    .map(mf -> mf.getFirstSubfieldValue("a"))
                    .filter(Objects::nonNull)
                    .findAny()
                    .ifPresent(this::setExternalIdentifier);
            Objects.requireNonNull(externalIdentifier);
            this.identifier = externalIdentifier;
            marcRecord.getAll("856")
                .forEach(mf -> {
                    Service.LabeledUrl labeledUrl = new Service.LabeledUrl();
                    String url = mf.getFirstSubfieldValue("u");
                    if (url != null) {
                        labeledUrl.setUrl(url);
                        List<String> list = new LinkedList<>();
                        String s = mf.getFirstSubfieldValue("x");
                        if (s != null) {
                            list.add(s);
                        }
                        s = mf.getFirstSubfieldValue("z");
                        if (s != null) {
                            list.add(s);
                        }
                        labeledUrl.setLabel(String.join(", ", list));
                        globalServices.add(labeledUrl);
                    }
                });
            // lookup EZB base URL
            marcRecord.getAll("856").stream()
                    .filter(mf -> "EZB".equals(mf.getFirstSubfieldValue("x")))
                    .map(mf -> mf.getFirstSubfieldValue("u"))
                    .filter(Objects::nonNull)
                    .findAny()
                    .ifPresent(this::setBaseUrl);
            // find EZB green
            marcRecord.getAll("856").stream()
                    .filter(mf -> "kostenfrei".equals(mf.getFirstSubfieldValue("z")))
                    .findAny()
                    .ifPresent(mf -> setGreen(true));
            // find open access
            marcRecord.getAll("506").stream()
                    .filter(mf -> "Unrestricted online access".equals(mf.getFirstSubfieldValue("f")))
                    .findAny()
                    .ifPresent(mf -> setOpenAccess(true));
            marcRecord.getAll("540").stream()
                    .filter(mf -> "DOAJ".equals(mf.getFirstSubfieldValue("q")))
                    .findAny()
                    .ifPresent(mf -> setOpenAccess(true));
            // lookup ISSNs for DOAJ
            marcRecord.getAll("022").forEach(mf -> mf.getSubfields()
                    .stream().filter(sf -> "a".equals(sf.getId()))
                    .forEach(sf -> issns.add(sf.getValue())));
        }
    }
}
