package org.xbib.interlibrary.catalog.elasticsearch.lvi.service;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.tiny.TinyMultiMap;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.ElxOutputHelper;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Holdings;
import org.xbib.interlibrary.api.library.IndexLibrary;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Record;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Service;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.marc.json.JsonBuilder;
import org.xbib.settings.Settings;

public class ServiceManager {

    private static final Logger logger = Logger.getLogger(ServiceManager.class.getName());

    private final Settings settings;

    private final Configuration configuration;

    private final Record record;

    private final TinyMultiMap<Integer, String> yearMap;

    private final boolean enabled;

    public ServiceManager(Settings settings,
                          Configuration configuration,
                          Record record) {
        this.settings = settings;
        this.configuration = configuration;
        this.record = record;
        this.enabled = settings.getAsBoolean("services.enabled", false);
        this.yearMap = new TinyMultiMap<>();
        if (enabled) {
            record.getHoldings().forEach(holding -> {
                if (checked(holding)) {
                    yearMap.put(-1, holding.getKey());
                    findAllYears(holding).forEach(d -> {
                        holding.addYear(d);
                        yearMap.put(d, holding.getKey());
                    });
                }
            });
        }
    }

    private Collection<Integer> findAllYears(Holdings holdings) {
        Set<Integer> years = new LinkedHashSet<>();
        for (Service service : holdings.getServices()) {
            if (service.getYears() != null) {
                years.addAll(service.getYears());
            }
        }
        return years;
    }

    /**
     * We have three styles of record that overlap.
     *
     * 1. the "record" style with the record key as document ID
     * 2. the "holdings" style with the holdings key as document ID
     * 3. and the "year" style with the holdings key plus the year as document ID
     *
     * @param elxOutputHelper the ELX output
     * @param indexDefinition the index definition of the LVI service index
     */
    public void index(ElxOutputHelper elxOutputHelper,
                      IndexDefinition indexDefinition) {
        try {
            String content = createRecordContent(record);
            if (!content.isEmpty()) {
                if (enabled && settings.getAsBoolean("enabled", false)) {
                    elxOutputHelper.index(indexDefinition, record.getKey(), true, content);
                } else {
                    logger.log(Level.FINE, MessageFormat.format("not indexing {0}/{1}/{2} {3}",
                            indexDefinition.getFullIndexName(),
                            indexDefinition.getType(),
                            record.getKey(), content));
                }
            }
        } catch (IOException e) {
            throw new IllegalStateException(e);
        }
        record.getHoldings().forEach(holdings -> {
            if (checked(holdings)) {
                try {
                    String content = createHoldingsContent(record, holdings);
                    if (!content.isEmpty()) {
                        if (enabled && settings.getAsBoolean("enabled", false)) {
                            elxOutputHelper.index(indexDefinition, holdings.getKey(), true, content);
                        } else {
                            logger.log(Level.FINE, MessageFormat.format("not indexing holdings {0}/{1}/{2} {3}",
                                    indexDefinition.getFullIndexName(), indexDefinition.getType(), holdings.getKey(), content));
                        }
                    }
                } catch (IOException e) {
                    throw new IllegalStateException(e);
                }
            }
        });
        yearMap.asMap().forEach((d, ids) -> {
            if (d != null && d != -1) {
                try {
                    String content = createYearContent(record, ids);
                    if (!content.isEmpty()) {
                        String key = record.getKey() + "." + d;
                        if (enabled && settings.getAsBoolean("enabled", false)) {
                            elxOutputHelper.index(indexDefinition, key, true, content);
                        } else {
                            logger.log(Level.FINE, MessageFormat.format("not indexing year {0}/{1}/{2} {3}",
                                    indexDefinition.getFullIndexName(), indexDefinition.getType(), key, content));
                        }
                    }
                } catch (IOException e) {
                    throw new IllegalStateException(e);
                }
            }
        });
    }

    private boolean checked(Holdings holdings) {
        if (holdings == null) {
            return false;
        }
        return !holdings.isDeleted() || holdings.isMerged();
    }

    private String createRecordContent(Record record) throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        builder.buildKey("parent")
                .buildValue(record.getKey());
        builder.buildKey("green")
                .buildValue(record.isGreen());
        builder.buildKey("openaccess")
                .buildValue(record.isOpenAccess());
        builder.buildKey("globalservices");
        builder.beginCollection();
        for (Service.LabeledUrl url : record.getGlobalServices()) {
            builder.beginMap();
            builder.buildKey("url")
                            .buildValue(url.getUrl());
            builder.buildKey("label")
                            .buildValue(url.getLabel());
            builder.endMap();
        }
        builder.endCollection();
        // record all years (except fallback value -1)
        Set<Integer> years = new HashSet<>(yearMap.keySet());
        years.remove(-1);
        builder.buildKey("years")
                .buildValue(years);
        // ids without year
        builder.buildKey("ids")
                .buildValue(yearMap.asMap().get(-1));
        builder.endMap();
        return builder.build();
    }

    private String createHoldingsContent(Record record,
                                         Holdings holdings) throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        builder.buildKey("parent")
                .buildValue(record.getKey());
        builder.buildKey("green")
                .buildValue(record.isGreen());
        builder.buildKey("openaccess")
                .buildValue(record.isOpenAccess());
        // it's more convenient to replicate the global services here
        if (!record.getGlobalServices().isEmpty()) {
            builder.buildKey("globalservices");
            builder.beginCollection();
            for (Service.LabeledUrl url : record.getGlobalServices()) {
                builder.beginMap();
                builder.buildKey("url")
                        .buildValue(url.getUrl());
                builder.buildKey("label")
                        .buildValue(url.getLabel());
                builder.endMap();
            }
            builder.endCollection();
        }
        builder.buildKey("isil")
                .buildValue(holdings.getISIL());
        IndexLibrary library = configuration.getLibraries().get(holdings.getISIL());
        if (library != null) {
            builder.buildKey("name")
                    .buildValue(library.getName());
            builder.buildKey("domain")
                    .buildValue(library.getDomain().getName());
            builder.buildKey("organization")
                    .buildValue(library.getDomain().getOrganization());
        }
        if (!holdings.getServices().isEmpty()) {
            builder.buildKey("services");
            builder.beginCollection();
            for (Service service : holdings.getServices()) {
                builder.beginMap();
                if (service.getMaterial() != null) {
                    builder.buildKey("material")
                            .buildValue(service.getMaterial().toValue());
                }
                if (service.getMode() != null) {
                    builder.buildKey("mode")
                            .buildValue(service.getMode().toValue());
                }
                builder.buildKey("priority")
                        .buildValue(getPriority(service));
                if (service.getISIL() != null && !service.getISIL().equals(holdings.getISIL())) {
                    builder.buildKey("library")
                            .buildValue(service.getISIL());
                    library = configuration.getLibraries().get(service.getISIL());
                    if (library != null) {
                        builder.buildKey("libraryname")
                                .buildValue(library.getName());
                    }
                }
                if (service.getCollection() != null) {
                    builder.buildKey("collection")
                            .buildValue(service.getCollection());
                }
                if (service.getCallnumber() != null) {
                    builder.buildKey("callnumber")
                            .buildValue(service.getCallnumber());
                }
                if (service.getCallnumber2() != null) {
                    builder.buildKey("callnumber2")
                            .buildValue(service.getCallnumber2());
                }
                if (service.getDescription() != null) {
                    builder.buildKey("description")
                            .buildValue(service.getDescription());
                }
                if (service.getTextualHoldings() != null) {
                    builder.buildKey("textualholdings")
                            .buildValue(service.getTextualHoldings());
                }
                if (service.getYears() != null && !service.getYears().isEmpty()) {
                    builder.buildKey("years")
                            .buildValue(service.getYears());
                }
                if (!service.getUrls().isEmpty()) {
                    builder.buildKey("links");
                    builder.beginCollection();
                    for (Service.LabeledUrl labeledUrl : service.getUrls()) {
                        builder.beginMap();
                        builder.buildKey("url")
                                .buildValue(labeledUrl.getUrl());
                        if (labeledUrl.getLabel() != null) {
                            builder.buildKey("label")
                                    .buildValue(labeledUrl.getLabel());
                        }
                        builder.endMap();
                    }
                    builder.endCollection();
                }
                builder.endMap();
            }
            builder.endCollection();
        }
        builder.endMap();
        return builder.build();
    }

    private String createYearContent(Record record,
                                     Collection<String> ids) throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        builder.buildKey("green")
                .buildValue(record.isGreen());
        builder.buildKey("openaccess")
                .buildValue(record.isOpenAccess());
        builder.buildKey("ids")
                .buildValue(ids);
        builder.endMap();
        return builder.build();
    }

    /**
     * Get the priority for delivery.
     * @param service the service
     * @return the prioirty as an integer value from 1 to 9
     */
    private Integer getPriority(Service service) {
        if (service.getFirstStatusCode() == null) {
            return 9; // delivery is unspecified or invalid
        }
        switch (service.getFirstStatusCode()) {
            case "a", "b", "c", "e" -> {
                if ("p".equals(service.getSecondStatusCode()) || "p".equals(service.getThirdStatusCode())) {
                    return 3; // "non-electronic", i.e. postal delivery
                }
                return switch (service.getMaterial()) {
                    case ONLINE_RESOURCE -> 1; // upload online material
                    case TEXT -> 2; // scan material
                    case ELECTRONIC_RESOURCE, MICROFORM -> 4; // maybe copy and upload from data carrier, maybe not
                    default -> 5; // all other material, delivery speed is unknown
                };
            }
            default -> {
                return 9; // delivery is unspecified or invalid
            }
        }
    }
}
