package org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics;

import java.io.IOException;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.MessageFormat;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.ElxOutputHelper;
import org.xbib.marc.json.JsonBuilder;
import org.xbib.settings.Settings;

public class ConfigurationTable {

    private static final Logger logger = Logger.getLogger(ConfigurationTable.class.getName());

    private final Settings settings;

    private final boolean enabled;

    private final ResourceStatistics resourceStatistics;

    private final ProductStatistics productStatistics;

    private final StatusCodeMapStatistics statusCodeMapStatistics;

    public ConfigurationTable(Settings settings,
                              ResourceStatistics resourceStatistics,
                              ProductStatistics productStatistics,
                              StatusCodeMapStatistics statusCodeMapStatistics) {
        this.settings = settings;
        this.enabled = settings.getAsBoolean("statistics.configurations.enabled", false);
        this.resourceStatistics = resourceStatistics;
        this.productStatistics = productStatistics;
        this.statusCodeMapStatistics = statusCodeMapStatistics;
    }

    public void write(String fileName) {
        if (enabled && fileName != null && !fileName.isEmpty()) {
            if (settings.containsSetting("statistics.path")) {
                String statisticsPath = settings.get("statistics.path");
                resourceStatistics.getISILs().forEach(isil -> {
                    try {
                        String content = getLibraryContent(isil,
                                resourceStatistics.getCollectionMapAsConfig(isil),
                                resourceStatistics.getStatusCodeMapAsConfig(isil),
                                resourceStatistics.getItemProcessCodeMapAsConfig(isil));
                        if (!content.isEmpty()) {
                            Files.createDirectories(Paths.get(statisticsPath, fileName));
                            try (Writer writer = Files.newBufferedWriter(Paths.get(statisticsPath, fileName, isil + ".json"))) {
                                writer.write(content);
                            }
                        }
                    } catch (IOException e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                    }
                });
                productStatistics.getCodeMap().forEach((product, members) -> {
                    try {
                        String content = getProductContent(product, members.objectMap());
                        if (!content.isEmpty()) {
                            Files.createDirectories(Paths.get(statisticsPath, fileName));
                            try (Writer writer = Files.newBufferedWriter(Paths.get(statisticsPath, fileName, clean(product) + ".json"))) {
                                writer.write(content);
                            }
                        }
                    } catch (IOException e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                    }
                });
            }
        }
    }

    public void index(Settings settings,
                      ElxOutputHelper elxOutputHelper,
                      IndexDefinition indexDefinition) {
        resourceStatistics.getISILs().forEach(isil -> {
            try {
                String content = getLibraryContent(isil,
                        resourceStatistics.getCollectionMapAsConfig(isil),
                        resourceStatistics.getStatusCodeMapAsConfig(isil),
                        resourceStatistics.getItemProcessCodeMapAsConfig(isil));
                if (!content.isEmpty()) {
                    if (enabled && settings.getAsBoolean("enabled", false) ) {
                        elxOutputHelper.index(indexDefinition,
                                isil, false, content);
                    } else {
                        logger.log(Level.FINE, MessageFormat.format("not indexing {0}/{1}/{2} {3}",
                                indexDefinition.getFullIndexName(),
                                indexDefinition.getType(),
                                isil, content));
                    }
                }
            } catch (IOException e) {
                throw new IllegalStateException(e);
            }
        });
        productStatistics.getCodeMap().forEach((product, members) -> {
            try {
                String content = getProductContent(product, members.objectMap());
                if (!content.isEmpty()) {
                    if (enabled && settings.getAsBoolean("enabled", false) ) {
                        elxOutputHelper.index(indexDefinition,
                                clean(product), false, content);
                    } else {
                        logger.log(Level.FINE, MessageFormat.format("not indexing {0}/{1}/{2} {3}",
                                indexDefinition.getFullIndexName(),
                                indexDefinition.getType(),
                                clean(product), content));
                    }
                }
            } catch (IOException e) {
                throw new IllegalStateException(e);
            }
        });
    }

    private String getLibraryContent(String name,
                                     Map<String, Object> collectionMap,
                                     Map<String, Object> statusCodeMap,
                                     Map<String, Object> processCodeMap) throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        builder.buildKey("name")
                .buildValue(name);
        builder.buildKey("enabled")
                .buildValue(true);
        builder.buildKey("collections").buildValue(collectionMap);
        builder.buildKey("statuscodes").buildValue(statusCodeMap);
        builder.buildKey("statuscodemap").buildValue(statusCodeMapStatistics.getCodeMaps(name));
        builder.buildKey("processcodes").buildValue(processCodeMap);
        builder.endMap();
        return builder.build();
    }

    private String getProductContent(String name, Map<String, Object> collectionMap) throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        builder.buildKey("name")
                .buildValue(name);
        builder.buildKey("enabled")
                .buildValue(true);
        builder.buildKey("collections").buildValue(collectionMap);
        builder.endMap();
        return builder.build();
    }

    private static String clean(String string) {
        if (string == null){
            return null;
        }
        return string.replaceAll("\\p{C}|\\p{IsPunctuation}|\\p{IsWhite_Space}", "");
    }

}
