package org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics;

import java.io.IOException;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.MessageFormat;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.ElxOutputHelper;
import org.xbib.interlibrary.api.product.Product;
import org.xbib.marc.json.JsonBuilder;
import org.xbib.settings.Settings;

public class ProductStatistics {

    private static final Logger logger = Logger.getLogger(ProductStatistics.class.getName());

    private final TinyMap.Builder<String, Integer> integerMapBuilder;

    private final TinyMap.Builder<String, Statistics.CodeMap> codeMapBuilder;

    private final Settings settings;

    private final boolean enabled;

    private Map<String, Integer> countMap;

    private Map<String, Statistics.CodeMap> codeMap;

    private boolean closed;

    public ProductStatistics(Settings settings) {
        this.settings = settings;
        this.enabled = settings.getAsBoolean("statistics.products.enabled", false);
        this.integerMapBuilder = TinyMap.builder();
        this.codeMapBuilder = TinyMap.builder();
    }

    public synchronized void count(Product product, String member) {
        if (!enabled) {
            return;
        }
        if (closed) {
            throw new ConcurrentModificationException();
        }
        if (product == null) {
            return;
        }
        String isil = product.getISIL();
        if (isil == null) {
            return;
        }
        Integer i = integerMapBuilder.getOrDefault(isil, 0);
        integerMapBuilder.put(isil, i + 1);
        Statistics.CodeMap codeMap = codeMapBuilder.getOrDefault(isil, new Statistics.CodeMap());
        if (member != null) {
            codeMap.add(member);
        }
        codeMapBuilder.put(isil, codeMap);
    }

    public Map<String, Integer> getCountMap() {
        if (countMap == null) {
            countMap = integerMapBuilder.build();
            closed = true;
        }
        return countMap;
    }

    public Integer getCountMapSummary() {
        return getCountMap().values().stream().reduce(0, Integer::sum);
    }

    public Map<String, Statistics.CodeMap> getCodeMap() {
        if (codeMap == null) {
            codeMap = codeMapBuilder.build();
            closed = true;
        }
        return codeMap;
    }

    public Map<String, Integer> getCodeMapSummary() {
        return getCodeMap()
                .values()
                .stream()
                .map(Statistics.CodeMap::map)
                .flatMap(m -> m.entrySet().stream())
                .collect(Collectors.groupingBy(Entry::getKey, Collectors.summingInt(Entry::getValue)));
    }

    public Collection<String> getProducts(String isil) {
        return getCodeMap().entrySet().stream()
                .filter(e -> e.getValue().map().containsKey(isil))
                .map(Entry::getKey)
                .collect(Collectors.toList());
    }

    public void write(String fileName, String suffix) throws IOException {
        if (!enabled) {
            return;
        }
        if (settings.containsSetting("statistics.path")) {
            String statisticsPath = settings.get("statistics.path");
            Files.createDirectories(Paths.get(statisticsPath, fileName));
            getCodeMap().forEach((product, members) -> {
                try (Writer writer = Files.newBufferedWriter(Paths.get(statisticsPath, fileName,
                        clean(product) + "." + suffix))) {
                    writer.write(getContent(product, members));
                } catch (IOException e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                }
            });
        }
    }

    public void index(Settings settings,
                      ElxOutputHelper elxOutputHelper,
                      IndexDefinition indexDefinition) {
        getCodeMap().forEach((product, members) -> {
            try {
                String content = getContent(product, members);
                if (!content.isEmpty()) {
                    if (enabled && settings.getAsBoolean("enabled", false) ) {
                        elxOutputHelper.index(indexDefinition,
                                clean(product), false, content);
                    } else {
                        logger.log(Level.FINE, MessageFormat.format("not indexing {0}/{1}/{2} {3}",
                                indexDefinition.getFullIndexName(),
                                indexDefinition.getType(),
                                clean(product), content));
                    }
                }
            } catch (IOException e) {
                throw new IllegalStateException(e);
            }
        });
    }

    private String getContent(String product, Statistics.CodeMap members) throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        builder.buildKey("product")
                .buildValue(product);
        builder.buildKey("counts");
        builder.beginMap();
        for (Entry<String, Integer> e : members.map().entrySet()) {
            builder.buildKey(e.getKey());
            builder.buildValue(e.getValue());
        }
        builder.endMap();
        builder.endMap();
        return builder.build();
    }

    private static String clean(String string) {
        if (string == null){
            return null;
        }
        return string.replaceAll("\\p{C}|\\p{IsPunctuation}|\\p{IsWhite_Space}", "");
    }
}
