package org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics;

import java.io.IOException;
import java.text.MessageFormat;
import java.time.Instant;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.ElxOutputHelper;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Record;
import org.xbib.marc.json.JsonBuilder;
import org.xbib.settings.Settings;

public class RecordStatistics {

    private static final Logger logger = Logger.getLogger(RecordStatistics.class.getName());

    private final boolean enabled;

    private final RecordCountStatistics recordCountStatistics;

    private final RecordDeletedCountStatistics recordDeletedCountStatistics;

    private final RecordMaterialCountStatistics recordMaterialCountStatistics;

    private final RecordGreenCountStatistics recordGreenCountStatistics;

    private final RecordOpenAccessCountStatistics recordOpenAccessCountStatistics;

    public RecordStatistics(Settings settings,
                            String prefix) {
        this.enabled = settings.getAsBoolean("statistics.records.enabled", false);
        this.recordCountStatistics = new RecordCountStatistics(settings, prefix);
        this.recordDeletedCountStatistics = new RecordDeletedCountStatistics(settings, prefix);
        this.recordMaterialCountStatistics = new RecordMaterialCountStatistics(settings, prefix);
        this.recordGreenCountStatistics = new RecordGreenCountStatistics(settings, prefix);
        this.recordOpenAccessCountStatistics = new RecordOpenAccessCountStatistics(settings, prefix);
    }

    public synchronized void count(Record record) {
        if (!enabled) {
            return;
        }
        recordCountStatistics.count(record);
        recordDeletedCountStatistics.count(record);
        recordMaterialCountStatistics.count(record);
        recordGreenCountStatistics.count(record);
        recordOpenAccessCountStatistics.count(record);
    }

    public Map<String, Integer> getCountMap() {
        return recordCountStatistics.map();
    }

    public Integer getCountSummary() {
        return recordCountStatistics.summary();
    }

    public Map<String, Integer> getDeletedCountMap() {
        return recordDeletedCountStatistics.map();
    }

    public Integer getDeletedCountSummary() {
        return recordDeletedCountStatistics.summary();
    }

    public Map<String, Integer> getMaterialCountMap() {
        return recordMaterialCountStatistics.map();
    }

    public Integer getMaterialCountSummary() {
        return recordMaterialCountStatistics.summary();
    }

    public Map<String, Integer> getGreenCountMap() {
        return recordGreenCountStatistics.map();
    }

    public Integer getGreenCountSummary() {
        return recordGreenCountStatistics.summary();
    }

    public Map<String, Integer> getOpenAccessCountMap() {
        return recordOpenAccessCountStatistics.map();
    }

    public Integer getOpenAccessCountSummary() {
        return recordOpenAccessCountStatistics.summary();
    }

    public void write(String fileName, String suffix) throws IOException {
        if (!enabled) {
            return;
        }
        recordCountStatistics.write(fileName + "-record-count." + suffix);
        recordDeletedCountStatistics.write(fileName + "-record-deleted-count." + suffix);
        recordMaterialCountStatistics.write(fileName + "-record-material-count." + suffix);
        recordGreenCountStatistics.write(fileName + "-record-green-count." + suffix);
        recordOpenAccessCountStatistics.write(fileName + "-record-open-access-count." + suffix);
    }

    public void index(Settings settings,
            ElxOutputHelper elxOutputHelper,
            IndexDefinition indexDefinition) {
        // overview
        try {
            JsonBuilder builder = JsonBuilder.builder();
            builder.beginMap();
            builder.buildKey("completed")
                    .buildValue(Instant.now().toString());
            builder.buildKey("count")
                    .buildValue(recordCountStatistics.objectMap());
            builder.buildKey("deletedcount")
                    .buildValue(recordDeletedCountStatistics.objectMap());
            builder.buildKey("materialcount")
                    .buildValue(recordMaterialCountStatistics.objectMap());
            builder.buildKey("greencount")
                    .buildValue(recordGreenCountStatistics.objectMap());
            builder.buildKey("openaccesscount")
                    .buildValue(recordOpenAccessCountStatistics.objectMap());
            builder.endMap();
            String content = builder.build();
            if (!content.isEmpty()) {
                if (enabled && settings.getAsBoolean("enabled", false)) {
                    elxOutputHelper.index(indexDefinition,
                            "recordoverview", false, content);
                } else {
                    logger.log(Level.FINE, MessageFormat.format("not indexing {0}/{1}/{2} {3}",
                            indexDefinition.getFullIndexName(),
                            indexDefinition.getType(),
                            "recordoverview", content));
                }
            }
        } catch (IOException e) {
            throw new IllegalStateException(e);
        }
    }
}
