package org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics;

import java.io.IOException;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.Holdings;
import org.xbib.marc.json.JsonBuilder;
import org.xbib.settings.Settings;

public class ResourceStatistics {

    private final boolean enabled;

    private final String domain;

    private final ResourceCountStatistics resourceCountStatistics;

    private final ResourceDeletedCountStatistics resourceDeletedCountStatistics;

    private final ResourceMaterialCountStatistics resourceMaterialCountStatistics;

    private final ResourceCollectionStatistics resourceCollectionStatistics;

    private final ResourceStatusCodeStatistics resourceStatusCodeStatistics;

    private final ResourceItemProcessStatistics resourceItemProcessStatistics;

    public ResourceStatistics(Settings settings, String prefix) {
        this.enabled = settings.getAsBoolean("statistics.resources.enabled", false);
        this.domain = settings.get("statistics.resources.domain", null);
        this.resourceCountStatistics = new ResourceCountStatistics(settings, prefix);
        this.resourceDeletedCountStatistics = new ResourceDeletedCountStatistics(settings, prefix);
        this.resourceMaterialCountStatistics = new ResourceMaterialCountStatistics(settings, prefix);
        this.resourceCollectionStatistics = new ResourceCollectionStatistics(settings, prefix);
        this.resourceStatusCodeStatistics = new ResourceStatusCodeStatistics(settings, prefix);
        this.resourceItemProcessStatistics = new ResourceItemProcessStatistics(settings, prefix);
    }

    public synchronized void count(Holdings holdings) {
        if (!enabled) {
            return;
        }
        if (domain == null || domain.equals(holdings.getDomain())) {
            resourceCountStatistics.count(holdings);
            resourceMaterialCountStatistics.count(holdings);
            resourceCollectionStatistics.count(holdings);
            resourceStatusCodeStatistics.count(holdings);
            resourceItemProcessStatistics.count(holdings);
        }
    }

    public Set<String> getISILs() {
        return resourceCountStatistics.map().keySet();
    }

    public Map<String, Integer> getCountMap() {
        return resourceCountStatistics.map();
    }

    public Integer getCountSummary() {
        return resourceCountStatistics.summary();
    }

    public Map<String, Integer> getDeletedCountMap() {
        return resourceDeletedCountStatistics.map();
    }

    public Integer getDeletedCountSummary() {
        return resourceDeletedCountStatistics.summary();
    }

    public Map<String, Integer> getMaterialCountMap() {
        return resourceMaterialCountStatistics.map();
    }

    public Integer getMaterialCountSummary() {
        return resourceMaterialCountStatistics.summary();
    }

    public Map<String, Statistics.CodeMap> getCollectionMap() {
        return resourceCollectionStatistics.map();
    }

    public String getCollectionMapAsJson() throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.buildMap(resourceCollectionStatistics.objectMap());
        return builder.build();
    }

    public Map<String, Object> getCollectionMap(String isil) {
        Map<String, Statistics.CodeMap> map = getCollectionMap();
        return map.containsKey(isil) ? map.get(isil).objectMap() : Map.of();
    }

    public Map<String, Object> getCollectionMapAsConfig(String isil) {
        Map<String, Statistics.CodeMap> map = getCollectionMap();
        return map.containsKey(isil) ? map.get(isil).objectMap().keySet().stream()
                .map(o -> Map.entry(o, true))
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue))
                : Map.of();
    }

    public Map<String, Object> getStatusCodeMap() {
        return resourceStatusCodeStatistics.objectMap();
    }

    public String getStatusCodeMapAsJson() throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.buildMap(resourceStatusCodeStatistics.objectMap());
        return builder.build();
    }

    public Map<String, Object> getStatusCodeMap(String isil) {
        Map<String, Statistics.CodeMap> map = resourceStatusCodeStatistics.map();
        return map.containsKey(isil) ? map.get(isil).objectMap() : Map.of();
    }

    public Map<String, Object> getStatusCodeMapAsConfig(String isil) {
        Map<String, Statistics.CodeMap> map = resourceStatusCodeStatistics.map();
        return map.containsKey(isil) ? map.get(isil).objectMap().keySet().stream()
                .map(o -> Map.entry(o, o))
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue))
                : Map.of();
    }

    public Map<String, Integer> getStatusCodeMapSummary() {
        return resourceStatusCodeStatistics.getSummary();
    }

    public Map<String, Statistics.CodeMap> getItemProcessCodeMap() {
        return resourceItemProcessStatistics.map();
    }

    public String getItemProcessCodeMapAsJson() throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.buildMap(resourceItemProcessStatistics.objectMap());
        return builder.build();
    }

    public Map<String, Object> getItemProcessCodeMap(String isil) {
        Map<String, Statistics.CodeMap> map = getItemProcessCodeMap();
        return map.containsKey(isil) ? map.get(isil).objectMap() : Map.of();
    }

    public Map<String, Object> getItemProcessCodeMapAsConfig(String isil) {
        Map<String, Statistics.CodeMap> map = getItemProcessCodeMap();
        return map.containsKey(isil) ? map.get(isil).objectMap().keySet().stream()
                .map(o -> Map.entry(o, o))
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue))
                : Map.of();
    }

    public Map<String, Integer> getItemProcessCodeMapSummary() {
        return resourceItemProcessStatistics.getSummary();
    }

    public void write(String fileName, String suffix) throws IOException {
        if (!enabled) {
            return;
        }
        resourceCountStatistics.write(fileName + "-item-count." + suffix);
        resourceDeletedCountStatistics.write(fileName + "-item-deleted-count." + suffix);
        resourceMaterialCountStatistics.write(fileName + "-item-material-count." + suffix);
        resourceCollectionStatistics.write(fileName + "-item-collection." + suffix);
        resourceStatusCodeStatistics.write(fileName + "-item-statuscodes." + suffix);
        resourceItemProcessStatistics.write(fileName + "-item-processcodes." + suffix);
    }
}
