package org.xbib.interlibrary.catalog.elasticsearch.lvi.statistics;

import java.io.IOException;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.MessageFormat;
import java.time.Instant;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.csv.Generator;
import org.xbib.elx.api.IndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.ElxOutputHelper;
import org.xbib.interlibrary.catalog.elasticsearch.lvi.config.Configuration;
import org.xbib.marc.json.JsonBuilder;
import org.xbib.settings.Settings;

public class ResourceStatisticsTable {

    private static final Logger logger = Logger.getLogger(ResourceStatisticsTable.class.getName());

    private final Configuration configuration;

    private final Settings settings;

    private final boolean enabled;

    private final ResourceStatistics lvi;

    private final ResourceStatistics zdb;

    private final ResourceStatistics hbzfix;

    private final ResourceStatistics hbzalma;

    private final StatusCodeMapStatistics statusCodeMapStatistics;

    public ResourceStatisticsTable(Configuration configuration,
                                   Settings settings,
                                   ResourceStatistics lvi,
                                   ResourceStatistics zdb,
                                   ResourceStatistics hbzfix,
                                   ResourceStatistics hbzaleph,
                                   ResourceStatistics hbzalma,
                                   StatusCodeMapStatistics statusCodeMapStatistics) {
        this.configuration = configuration;
        this.settings = settings;
        this.enabled = settings.getAsBoolean("statistics.enabled", false);
        this.lvi = lvi;
        this.zdb = zdb;
        this.hbzfix = hbzfix;
        this.hbzalma = hbzalma;
        this.statusCodeMapStatistics = statusCodeMapStatistics;
    }

    public Set<String> getKeys() {
        SortedSet<String> set = new TreeSet<>();
        set.addAll(lvi.getCountMap().keySet());
        set.addAll(zdb.getCountMap().keySet());
        set.addAll(hbzfix.getCountMap().keySet());
        set.addAll(hbzalma.getCountMap().keySet());
        return set;
    }

    public void write(String fileName) throws IOException {
        if (enabled && fileName != null && !fileName.isEmpty()) {
            if (settings.containsSetting("statistics.path")) {
                String statisticsPath = settings.get("statistics.path");
                Files.createDirectories(Paths.get(statisticsPath));
                try (Writer writer = Files.newBufferedWriter(Paths.get(statisticsPath, fileName))) {
                    Generator csvGenerator = new Generator(writer);
                    csvGenerator.keys(List.of("REGION", "ISIL", "ZDB", "ALMA", "HBZFIX", "LVI"));
                    csvGenerator.writeKeys();
                    for (String key : getKeys()) {
                        csvGenerator.write(configuration.getLibraries().get(key).getDomain().getName());
                        csvGenerator.write(key);
                        Integer i;
                        i = zdb.getCountMap().get(key);
                        csvGenerator.write(i != null ? i.toString() : "");
                        i = hbzalma.getCountMap().get(key);
                        csvGenerator.write(i != null ? i.toString() : "");
                        i = hbzfix.getCountMap().get(key);
                        csvGenerator.write(i != null ? i.toString() : "");
                        i = lvi.getCountMap().get(key);
                        csvGenerator.write(i != null ? i.toString() : "");
                    }
                    csvGenerator.close();
                }
            }
        }
    }

    public void index(Settings settings,
                      ElxOutputHelper elxOutputHelper,
                      IndexDefinition indexDefinition) {
        // overview
        try {
            JsonBuilder builder = JsonBuilder.builder();
            builder.beginMap();
            builder.buildKey("completed");
            builder.buildValue(Instant.now().toString());
            builder.buildKey("count");
            builder.beginMap();
            builder.buildKey("zdb").buildValue(zdb.getCountSummary());
            builder.buildKey("hbzalma").buildValue(hbzalma.getCountSummary());
            builder.buildKey("hbzfix").buildValue(hbzfix.getCountSummary());
            builder.buildKey("lvi").buildValue(lvi.getCountSummary());
            builder.endMap();
            builder.buildKey("materialcount");
            builder.beginMap();
            builder.buildKey("zdb").buildValue(zdb.getMaterialCountSummary());
            builder.buildKey("hbzalma").buildValue(hbzalma.getMaterialCountSummary());
            builder.buildKey("hbzfix").buildValue(hbzfix.getMaterialCountSummary());
            builder.buildKey("lvi").buildValue(lvi.getMaterialCountSummary());
            builder.endMap();
            builder.buildKey("materialmaps");
            builder.beginMap();
            builder.buildKey("zdb").buildValue(zdb.getMaterialCountMap());
            builder.buildKey("hbzalma").buildValue(hbzalma.getMaterialCountMap());
            builder.buildKey("hbzfix").buildValue(hbzfix.getMaterialCountMap());
            builder.buildKey("lvi").buildValue(lvi.getMaterialCountMap());
            builder.endMap();
            builder.buildKey("statuscode");
            builder.beginMap();
            builder.buildKey("zdb").buildValue(zdb.getStatusCodeMapSummary());
            builder.buildKey("hbzalma").buildValue(hbzalma.getStatusCodeMapSummary());
            builder.buildKey("hbzfix").buildValue(hbzfix.getStatusCodeMapSummary());
            builder.buildKey("lvi").buildValue(lvi.getStatusCodeMapSummary());
            builder.endMap();
            builder.buildKey("statuscodemaps");
            builder.beginMap();
            builder.buildKey("zdb").buildValue(zdb.getStatusCodeMap());
            builder.buildKey("hbzalma").buildValue(hbzalma.getStatusCodeMap());
            builder.buildKey("hbzfix").buildValue(hbzfix.getStatusCodeMap());
            builder.buildKey("lvi").buildValue(lvi.getStatusCodeMap());
            builder.endMap();
            builder.buildKey("processcode");
            builder.beginMap();
            builder.buildKey("zdb").buildValue(zdb.getItemProcessCodeMapSummary());
            builder.buildKey("hbzalma").buildValue(hbzalma.getItemProcessCodeMapSummary());
            builder.buildKey("hbzfix").buildValue(hbzfix.getItemProcessCodeMapSummary());
            builder.buildKey("lvi").buildValue(lvi.getItemProcessCodeMapSummary());
            builder.endMap();
            builder.endMap();
            String content = builder.build();
            if (!content.isEmpty()) {
                if (enabled && settings.getAsBoolean("enabled", false)) {
                    elxOutputHelper.index(indexDefinition,
                            "resourcesoverview", false, content);
                } else {
                    logger.log(Level.FINE, MessageFormat.format("not indexing {0}/{1}/{2} {3}",
                            indexDefinition.getFullIndexName(),
                            indexDefinition.getType(),
                            "resourcesoverview", content));
                }
            }
        } catch (IOException e) {
            throw new IllegalStateException(e);
        }
        // per ISIL
        getKeys().forEach(isil -> {
            try {
                JsonBuilder builder = JsonBuilder.builder();
                builder.beginMap();
                builder.buildKey("isil")
                        .buildValue(isil);
                builder.buildKey("count");
                builder.beginMap();
                builder.buildKey("zdb").buildValue(zdb.getCountMap().get(isil));
                builder.buildKey("hbzalma").buildValue(hbzalma.getCountMap().get(isil));
                builder.buildKey("hbzfix").buildValue(hbzfix.getCountMap().get(isil));
                builder.buildKey("lvi").buildValue(lvi.getCountMap().get(isil));
                builder.endMap();
                builder.buildKey("materialcount");
                builder.beginMap();
                builder.buildKey("zdb").buildValue(zdb.getMaterialCountMap().get(isil));
                builder.buildKey("hbzalma").buildValue(hbzalma.getMaterialCountMap().get(isil));
                builder.buildKey("hbzfix").buildValue(hbzfix.getMaterialCountMap().get(isil));
                builder.buildKey("lvi").buildValue(lvi.getMaterialCountMap().get(isil));
                builder.endMap();
                builder.buildKey("collection");
                builder.beginMap();
                builder.buildKey("zdb").buildValue(zdb.getCollectionMap(isil));
                builder.buildKey("hbzalma").buildValue(hbzalma.getCollectionMap(isil));
                builder.buildKey("hbzfix").buildValue(hbzfix.getCollectionMap(isil));
                builder.buildKey("lvi").buildValue(lvi.getCollectionMap(isil));
                builder.endMap();
                builder.buildKey("statuscode");
                builder.beginMap();
                builder.buildKey("zdb").buildValue(zdb.getStatusCodeMap(isil));
                builder.buildKey("hbzalma").buildValue(hbzalma.getStatusCodeMap(isil));
                builder.buildKey("hbzfix").buildValue(hbzfix.getStatusCodeMap(isil));
                builder.buildKey("lvi").buildValue(lvi.getStatusCodeMap(isil));
                builder.endMap();
                builder.buildKey("statuscodemap").buildValue(statusCodeMapStatistics.getCodeMaps(isil));
                builder.buildKey("processcode");
                builder.beginMap();
                builder.buildKey("zdb").buildValue(zdb.getItemProcessCodeMap(isil));
                builder.buildKey("hbzalma").buildValue(hbzalma.getItemProcessCodeMap(isil));
                builder.buildKey("hbzfix").buildValue(hbzfix.getItemProcessCodeMap(isil));
                builder.buildKey("lvi").buildValue(lvi.getItemProcessCodeMap(isil));
                builder.endMap();
                builder.endMap();
                String content = builder.build();
                if (!content.isEmpty()) {
                    if (enabled && settings.getAsBoolean("enabled", false)) {
                        elxOutputHelper.index(indexDefinition,
                                isil, false, content);
                    } else {
                        logger.log(Level.FINE, MessageFormat.format("not indexing {0}/{1}/{2} {3}",
                                indexDefinition.getFullIndexName(),
                                indexDefinition.getType(),
                                isil, content));
                    }
                }
            } catch (IOException e) {
                throw new IllegalStateException(e);
            }
        });
    }
}
