package org.xbib.interlibrary.catalog.elasticsearch.marc;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.Normalizer;
import java.time.Instant;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.ConverterOutput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.marc.Marc;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.settings.Settings;

/**
 * Elasticsearch indexer tool for MARC bibliographic data input streams.
 */
public class MarcBibliographic extends ElasticsearchTool<String> {

    public MarcBibliographic() {
    }

    @Override
    public MarcBibliographicWorkerPool newPool(Settings settings) throws IOException {
        return new MarcBibliographicWorkerPool(settings);
    }

    public class MarcBibliographicWorkerPool extends ElasticsearchWorkerPool<String> {

        public MarcBibliographicWorkerPool(Settings settings) throws IOException {
            super("marc-bib", settings);
        }

        @Override
        public MarcBibliographicWorker newWorker(String name, int count) throws IOException {
            return new MarcBibliographicWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }

    public class MarcBibliographicWorker extends ElasticsearchWorker<String, MarcBibliographicWorkerPool> {

        private final Logger logger = Logger.getLogger(MarcBibliographicWorker.class.getName());

        protected final MarcBibliographicWorkerPool pool;

        private CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        public MarcBibliographicWorker(Settings settings, MarcBibliographicWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            MarcValueTransformers marcValueTransformers = newMarcValueTransformers();
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings)) {
                catalogEntityWorkerPool = new CatalogEntityWorkerPool<>(settings, new DefaultWorkerPoolMarcRecordListener(),
                        MarcBibliographicToolCatalogEntityBuilder::new, pool);
                catalogEntityWorkerPool.setRequestOrigin(uri);
                doExecute(inputStream, catalogEntityWorkerPool, marcValueTransformers);
                // we must close the entity pool between URI executions
                catalogEntityWorkerPool.close();
            }
        }

        protected MarcValueTransformers newMarcValueTransformers() {
            MarcValueTransformers marcValueTransformers = new MarcValueTransformers();
            marcValueTransformers.setMarcValueTransformer(value -> Normalizer.normalize(value, Normalizer.Form.NFC));
            return marcValueTransformers;
        }

        protected void doExecute(InputStream inputStream,
                                 CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool,
                                 MarcValueTransformers marcValueTransformers) throws IOException {
            String dialect = settings.get("dialect");
            String charset = settings.get("charset", "UTF-8");
            if ("bibliomondo".equals(dialect)) {
                Marc marc = Marc.builder()
                        .setInputStream(inputStream)
                        .setMarcListener(catalogEntityWorkerPool)
                        .setMarcValueTransformers(marcValueTransformers)
                        .setCharset(Charset.forName(charset))
                        .build();
                marc.wrapIntoCollection(marc.bibliomondo());
            } else {
                Marc.builder()
                        .setInputStream(inputStream)
                        .setMarcRecordListener(catalogEntityWorkerPool)
                        .setMarcValueTransformers(marcValueTransformers)
                        .setCharset(Charset.forName(charset))
                        .build()
                        .writeRecords();
            }
        }

        /**
         *
         */
        protected class MarcBibliographicToolCatalogEntityBuilder extends CatalogEntityBuilder {

            private final ElasticsearchIndexDefinition elasticsearchIndexDefinition;

            protected MarcBibliographicToolCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
                elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
                if (elasticsearchIndexDefinition == null) {
                    throw new IOException("no 'bib' index definition configured");
                }
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                RouteRdfXContentParams routeRdfXContentParams =
                        new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                                elasticsearchIndexDefinition.getFullIndexName(),
                                elasticsearchIndexDefinition.getType());
                routeRdfXContentParams.setHandler((content, p) -> {
                    pool.bulkIndex(elasticsearchIndexDefinition, state.getRecordIdentifier(), false, content);
                    for (Map.Entry<String, ConverterOutput.Entry> entry : pool.getConverterOutput().getMap().entrySet()) {
                        try {
                            synchronized (pool) {
                                OutputStream outputStream = entry.getValue().getOut();
                                outputStream.write(content.getBytes(StandardCharsets.UTF_8));
                                outputStream.write('\n');
                                outputStream.flush();
                            }
                        } catch (IOException e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                        }
                    }
                });
                if (settings.get("format") != null) {
                    state.getResource().add("format", settings.get("format"));
                }
                if (settings.get("type") != null) {
                    state.getResource().add("type", settings.get("type"));
                }
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("xbib")
                        .add("uid", state.getRecordIdentifier());
                state.getResource().newResource("rec")
                        .add("created", catalogEntityWorkerPool.getCreatedFormatted())
                        .add("modified", catalogEntityWorkerPool.getLastModifiedFormatted())
                        .add("indexed", Instant.now().toString())
                        .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(state.getResource());
                }
            }
        }
    }
}
