package org.xbib.interlibrary.catalog.elasticsearch.marc;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.Normalizer;
import java.time.Instant;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;

import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.ConverterOutput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.marc.Marc;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.settings.Settings;

/**
 * Elasticsearch indexer tool for MARC holdings input streams.
 */
public class MarcHoldings extends ElasticsearchTool<String> {

    public MarcHoldings() {
    }

    @Override
    public MarcHoldingsWorkerPool newPool(Settings settings) throws IOException {
        return new MarcHoldingsWorkerPool(settings);
    }

    public class MarcHoldingsWorkerPool extends ElasticsearchWorkerPool<String> {

        public MarcHoldingsWorkerPool(Settings settings) throws IOException {
            super("marc-hol", settings);
        }

        @Override
        public MarcHoldingsWorker newWorker(String name, int count) {
            return new MarcHoldingsWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }

    public class MarcHoldingsWorker extends ElasticsearchWorker<String, MarcHoldingsWorkerPool> {

        private CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        protected final MarcHoldingsWorkerPool pool;

        public MarcHoldingsWorker(Settings settings, MarcHoldingsWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            MarcValueTransformers marcValueTransformers = newMarcValueTransformers();
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings)) {
                 catalogEntityWorkerPool = new CatalogEntityWorkerPool<>(settings, new DefaultWorkerPoolMarcRecordListener(),
                                 MarcHoldingsToolCatalogEntityBuilder::new, pool);
                 catalogEntityWorkerPool.setRequestOrigin(uri);
                 doExecute(inputStream, catalogEntityWorkerPool, marcValueTransformers);
                 catalogEntityWorkerPool.close();
            }
        }

        protected MarcValueTransformers newMarcValueTransformers() {
            MarcValueTransformers marcValueTransformers = new MarcValueTransformers();
            marcValueTransformers.setMarcValueTransformer(value -> Normalizer.normalize(value, Normalizer.Form.NFC));
            return marcValueTransformers;
        }

        protected void doExecute(InputStream inputStream,
                                 CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityBuilder,
                                 MarcValueTransformers marcValueTransformers) throws IOException {
            String dialect = settings.get("dialect");
            String charset = settings.get("charset", "UTF-8");
            if ("bibliomondo".equals(dialect)) {
                Marc marc = Marc.builder()
                        .setInputStream(inputStream)
                        .setMarcListener(catalogEntityBuilder)
                        .setMarcValueTransformers(marcValueTransformers)
                        .setCharset(Charset.forName(charset))
                        .build();
                marc.wrapIntoCollection(marc.bibliomondo());
            } else {
                Marc.builder()
                        .setInputStream(inputStream)
                        .setMarcRecordListener(catalogEntityBuilder)
                        .setMarcValueTransformers(marcValueTransformers)
                        .setCharset(Charset.forName(charset))
                        .build()
                        .writeRecords();
            }
        }

        /**
         *
         */
        private class MarcHoldingsToolCatalogEntityBuilder extends CatalogEntityBuilder {

            private final ElasticsearchIndexDefinition elasticsearchIndexDefinition;

            private MarcHoldingsToolCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
                this. elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("hol");
                if (elasticsearchIndexDefinition == null) {
                    throw new IOException("no 'hol' index definition configured");
                }
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                RouteRdfXContentParams routeRdfXContentParams =
                        new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                                elasticsearchIndexDefinition.getFullIndexName(),
                                elasticsearchIndexDefinition.getType());
                routeRdfXContentParams.setHandler((content, p) -> {
                    pool.bulkIndex(elasticsearchIndexDefinition, state.getRecordIdentifier(), false, content);
                    for (Map.Entry<String, ConverterOutput.Entry> entry : pool.getConverterOutput().getMap().entrySet()) {
                        if (entry.getKey().startsWith("file")) {
                            synchronized (pool) {
                                entry.getValue().getOut().write(content.getBytes(StandardCharsets.UTF_8));
                                entry.getValue().getOut().write('\n');
                                entry.getValue().getOut().flush();
                            }
                        }
                    }
                });
                if (settings.get("format") != null) {
                    state.getResource().add("format", settings.get("format"));
                }
                if (settings.get("type") != null) {
                    state.getResource().add("type", settings.get("type"));
                }
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("xbib")
                        .add("uid", state.getRecordIdentifier());
                state.getResource().newResource("rec")
                        .add("created", catalogEntityWorkerPool.getCreatedFormatted())
                        .add("modified", catalogEntityWorkerPool.getLastModifiedFormatted())
                        .add("indexed", Instant.now().toString())
                        .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(state.getResource());
                }
            }
        }
    }
}
