package org.xbib.interlibrary.catalog.elasticsearch.marc;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;
import java.text.Normalizer;
import java.time.Instant;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.oai.ElasticsearchToolOAI;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcListener;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.marc.xml.MarcXchangeWriter;
import org.xbib.settings.Settings;

/**
 * Harvest MARC Holdings records from OAI service and index into Elasticsearch.
 */
public class MarcHoldingsOAI extends ElasticsearchToolOAI {

    @Override
    public MarcHoldingsOAIWorkerPool newPool(Settings settings) throws IOException {
        return new MarcHoldingsOAIWorkerPool(settings);
    }

    public class MarcHoldingsOAIWorkerPool extends ElasticsearchOAIWorkerPool {

        public MarcHoldingsOAIWorkerPool(Settings settings) throws IOException {
            super(settings);
        }

        @Override
        public MarcHoldingsOAIWorker newWorker(String name, int count) {
            return new MarcHoldingsOAIWorker(settings, this);
        }
    }

    public class MarcHoldingsOAIWorker extends ElasticsearchOAIWorker {

        private final Logger logger = Logger.getLogger(MarcHoldingsOAIWorker.class.getName());

        private CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        private MarcListener marcListener;

        private MarcValueTransformers marcValueTransformers;

        public MarcHoldingsOAIWorker(Settings settings, ElasticsearchOAIWorkerPool pool) {
            super(settings, pool);
            try {
                marcValueTransformers = new MarcValueTransformers();
                marcValueTransformers.setMarcValueTransformer(value ->
                        Normalizer.normalize(new String(value.getBytes(StandardCharsets.ISO_8859_1), StandardCharsets.UTF_8),
                                Normalizer.Form.NFKC));
                catalogEntityWorkerPool =
                        new CatalogEntityWorkerPool<>(settings, new DefaultWorkerPoolMarcRecordListener(),
                                MarcHoldingsToolCatalogEntityBuilder::new, pool);
                if (pool.getConverterOutput().getMap().get("output") != null) {
                    marcListener = newMarcWriter(pool.getConverterOutput().getMap().get("output").getOut());
                }
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        @Override
        public ElasticsearchIndexDefinition getIndexDefinition() {
            return pool.getOutputIndexDefinitionMap().get("hol");
        }

        @Override
        protected void process(InputStream inputStream) {
            try {
                Marc.builder()
                        .setInputStream(inputStream)
                        .setMarcRecordListener(catalogEntityWorkerPool)
                        .setMarcValueTransformers(marcValueTransformers)
                        .setContentHandler(newMarcContentHandler(marcListener))
                        .build()
                        .xmlReader()
                        .parse();
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        @Override
        public void close() {
        }

        protected MarcContentHandler newMarcContentHandler(MarcListener marcListener) {
            return new MarcContentHandler()
                    .setFormat("MarcXML")
                    .setType("Holdings")
                    .addNamespace("http://www.loc.gov/MARC21/slim")
                    .setMarcListener(marcListener);
        }

        protected MarcListener newMarcWriter(OutputStream outputStream) throws IOException {
            return new MarcXchangeWriter(outputStream, true)
                    .setFormat("MARC21")
                    .setType("Holdings");
        }

        private class MarcHoldingsToolCatalogEntityBuilder extends CatalogEntityBuilder {

            private final ElasticsearchIndexDefinition elasticsearchIndexDefinition;

            private MarcHoldingsToolCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
                this.elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("hol");
                if (elasticsearchIndexDefinition == null) {
                    throw new IOException("no 'hol' index definition configured");
                }
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                RouteRdfXContentParams routeRdfXContentParams =
                        new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(),
                                elasticsearchIndexDefinition.getFullIndexName(),
                                elasticsearchIndexDefinition.getType());
                routeRdfXContentParams.setHandler((content, p) ->
                        pool.bulkIndex(elasticsearchIndexDefinition, state.getRecordIdentifier(), false, content));
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("xbib")
                        .add("uid", state.getRecordIdentifier());
                state.getResource().newResource("rec")
                        .add("created", catalogEntityWorkerPool.getCreatedFormatted())
                        .add("modified", catalogEntityWorkerPool.getLastModifiedFormatted())
                        .add("indexed", Instant.now().toString())
                        .add("origin", catalogEntityWorkerPool.getRequestOrigin());
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(state.getResource());
                }
            }
        }
    }
}
