package org.xbib.interlibrary.catalog.elasticsearch.oai;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.io.ntriple.NTripleContentParams;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.oai.OAIExceptionHandler;
import org.xbib.oai.client.OAIClient;
import org.xbib.oai.client.identify.IdentifyResponse;
import org.xbib.oai.exceptions.OAIException;
import org.xbib.oai.rdf.RdfResourceHandler;
import org.xbib.oai.xml.MetadataHandler;
import org.xbib.oai.xml.SimpleMetadataHandler;
import org.xbib.settings.Settings;
import org.xbib.time.chronic.Chronic;
import org.xbib.time.chronic.Span;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Harvest from OAI and index into Elasticsearch.
 */
public class ElasticsearchToolOAI extends ElasticsearchTool<String> {

    @Override
    public ElasticsearchOAIWorkerPool newPool(Settings settings) throws IOException {
        return new ElasticsearchOAIWorkerPool(settings);
    }

    public class ElasticsearchOAIWorkerPool extends ElasticsearchWorkerPool<String> {

        private final Settings oaiSettings;

        private final IRINamespaceContext namespaceContext;

        private final AtomicLong errorCounter;

        public ElasticsearchOAIWorkerPool(Settings settings) throws IOException {
            super("elasticsearch-oai", settings);
            this.oaiSettings = settings.getGroups("input").get("oai");
            this.namespaceContext = IRINamespaceContext.newInstance(ElasticsearchToolOAI.class);
            this.errorCounter = new AtomicLong(0);
        }

        @Override
        public ElasticsearchOAIWorker newWorker(String name, int count) throws IOException {
            return new ElasticsearchOAIWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        @Override
        protected void openInput() throws IOException, InterruptedException {
            if (oaiSettings != null) {
                submit(newWorkerRequest(oaiSettings.get("uri"), "oai"));
            }
            super.openInput();
        }

        @Override
        public IRINamespaceContext getNamespaceContext() {
            return namespaceContext;
        }

        public long getErrorCounter() {
            return errorCounter.get();
        }
    }

    public class ElasticsearchOAIWorker extends ElasticsearchWorker<String, ElasticsearchOAIWorkerPool> {

        private final Logger logger = Logger.getLogger(ElasticsearchOAIWorker.class.getName());

        private final Settings oaiSettings;

        protected final ElasticsearchOAIWorkerPool pool;

        public ElasticsearchOAIWorker(Settings settings, ElasticsearchOAIWorkerPool pool) {
            super(settings);
            this.pool = pool;
            this.oaiSettings = settings.getGroups("input").get("oai");
        }

        public ElasticsearchIndexDefinition getIndexDefinition() {
            return pool.getOutputIndexDefinitionMap().get("bib");
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            if (oaiSettings == null) {
                doExecuteFromFile(uri);
                return;
            }
            String metadataPrefix = oaiSettings.get("metadataPrefix");
            String set = oaiSettings.get("set");
            String granularity = oaiSettings.get("granularity");
            Instant from = fromChrono(oaiSettings.get("from"));
            Instant until = fromChrono(oaiSettings.get("until"));
            Instant base = fromChrono(oaiSettings.get("base"));
            OAIClient oaiClient = new OAIClient(uri);
            oaiClient.setUserAgent("hbz OAI client");
            if (granularity == null) {
                try {
                    IdentifyResponse identifyResponse = oaiClient.identify();
                    granularity = identifyResponse.getGranularity();
                } catch (OAIException e) {
                    logger.log(Level.WARNING, e.getMessage(), e);
                } catch (InterruptedException e) {
                    logger.log(Level.WARNING, "interrupted");
                }
            }
            if (granularity == null) {
                granularity = "YYYY-MM-DD";
            }
            logger.log(Level.INFO, "using granularity = " + granularity);
            DateTimeFormatter dateTimeFormatter = switch (granularity) {
                case "yyyy-MM-dd'T'HH:mm:ss'Z'", "yyyy-MM-ddTHH:mm:ssZ" ->
                        DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss'Z'").withZone(ZoneId.of("UTC"));
                default -> DateTimeFormatter.ofPattern("yyyy-MM-dd").withZone(ZoneId.of("UTC"));
            };
            MetadataHandler handler = getMetadataHandler();
            OAIExceptionHandler exceptionHandler = getExceptionHandler();
            try {
                if (handler != null) {
                    oaiClient.listRecords(metadataPrefix, set,
                            dateTimeFormatter, from, until, base, handler, exceptionHandler, null);
                } else {
                    oaiClient.listRecords(metadataPrefix, set,
                            dateTimeFormatter, from, until, base, null, exceptionHandler, this::process);
                }
            } catch (OAIException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                pool.errorCounter.incrementAndGet();
                // drop out with illegal state
                throw new IllegalStateException(e);
            }
        }

        private void doExecuteFromFile(String uri) throws IOException {
            ElasticsearchIndexDefinition indexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings);
                 BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(inputStream,
                         StandardCharsets.UTF_8))) {
                bufferedReader.lines().forEach(line -> {
                    if (line.startsWith("oai:")) {
                        logger.log(Level.INFO, line);
                    } else {
                        pool.bulkIndex(indexDefinition, null, false, line);
                    }
                });
            }
        }

        protected void process(InputStream inputStream) {
            // override
        }

        private Instant fromChrono(String chronoString) {
            if (chronoString == null) {
                logger.log(Level.WARNING, "chrono is null, assuming current time");
                return Instant.now();
            }
            try {
                Span span = Chronic.parse(chronoString);
                logger.log(Level.INFO, "found instant from chrono: " + DateTimeFormatter.ISO_INSTANT.format(span.getBeginCalendar()));
                return span.getBeginCalendar().toInstant();
            } catch (ParseException e) {
                logger.log(Level.WARNING, "no chrono found in " + chronoString + ", parsing instant");
                return Instant.parse(chronoString);
            }
        }

        protected MetadataHandler getMetadataHandler() {
            return null; // override if required
        }

        protected OAIExceptionHandler getExceptionHandler() {
            return null;
        }

        protected String convert(String id, String content) throws IOException {
            return content;
        }

        private RdfResourceHandler rdfResourceHandler() {
            NTripleContentParams params = new NTripleContentParams(pool.namespaceContext);
            return new RdfResourceHandler(params);
        }

        public class OAISimpleMetadataHandler extends SimpleMetadataHandler {

            protected final ElasticsearchOAIWorker worker;

            protected final IRINamespaceContext namespaceContext;

            RdfResourceHandler handler;

            public OAISimpleMetadataHandler(ElasticsearchOAIWorker worker) {
                this.worker = worker;
                namespaceContext = IRINamespaceContext.newInstance();
                namespaceContext.addNamespace("", "http://www.openarchives.org/OAI/2.0/oai_dc/");
                namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
            }

            @Override
            public void startDocument() throws SAXException {
                this.handler = worker.rdfResourceHandler();
                handler.setDefaultNamespace("", "http://www.openarchives.org/OAI/2.0/oai_dc/");
                handler.startDocument();
            }

            @Override
            public void endDocument() throws SAXException {
                handler.endDocument();
                try {
                    RouteRdfXContentParams params = new RouteRdfXContentParams(namespaceContext,
                            worker.getIndexDefinition().getFullIndexName(),
                            worker.getIndexDefinition().getType());
                    params.setHandler((content, p) -> {
                        // convert OAI content
                        String data = convert(getHeader().getIdentifier(), content);
                        if (data != null) {
                            worker.pool.bulkIndex(worker.getIndexDefinition(), getHeader().getIdentifier(), false, data);
                        } else {
                            logger.log(Level.WARNING, "unable to index null");
                        }
                    });
                    try (RdfContentBuilder<?> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                        builder.receive(handler.getResource());
                    }
                } catch (IOException e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                    throw new SAXException(e);
                }
            }

            @Override
            public void startPrefixMapping(String string, String string1) throws SAXException {
                handler.startPrefixMapping(string, string1);
            }

            @Override
            public void endPrefixMapping(String string) throws SAXException {
                handler.endPrefixMapping(string);
            }

            @Override
            public void startElement(String ns, String localname, String string2, Attributes atrbts) throws SAXException {
                handler.startElement(ns, localname, string2, atrbts);
            }

            @Override
            public void endElement(String ns, String localname, String string2) throws SAXException {
                handler.endElement(ns, localname, string2);
            }

            @Override
            public void characters(char[] chars, int i, int i1) throws SAXException {
                handler.characters(chars, i, i1);
            }
        }
    }
}
