package org.xbib.interlibrary.catalog.elasticsearch.openalex;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.Map;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.common.bytes.BytesArray;
import org.elasticsearch.common.xcontent.XContentType;
import org.xbib.datastructures.common.Pair;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.interlibrary.api.action.worker.Worker;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.api.util.Maps;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.settings.Settings;

public class OpenAlex extends ElasticsearchTool<String> {

    @Override
    public OpenAlexPool newPool(Settings settings) throws IOException {
        return new OpenAlexPool(settings);
    }

    public class OpenAlexPool extends ElasticsearchWorkerPool<String> {

        protected OpenAlexPool(Settings settings) throws IOException {
            super("openalex", settings);
        }

        @Override
        public Worker<String> newWorker(String name, int count) {
            return new OpenAlexWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }

    public class OpenAlexWorker extends ElasticsearchWorker<String, OpenAlexPool> {

        private final Logger logger = Logger.getLogger(OpenAlexWorker.class.getName());

        private final OpenAlexPool pool;

        private final ElasticsearchIndexDefinition elasticsearchIndexDefinition;

        protected OpenAlexWorker(Settings settings, OpenAlexPool pool) {
            super(settings);
            this.pool = pool;
            this.elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("openalex");
            Objects.requireNonNull(elasticsearchIndexDefinition.getFullIndexName());
        }

        @Override
        protected void doExecute(String request) throws IOException {
            logger.log(Level.INFO, "processing " + request);
            try (InputStream in = ConverterInput.getInputStream(request, settings);
                 BufferedReader reader = new BufferedReader(new InputStreamReader(in, StandardCharsets.UTF_8))) {
                reader.lines().forEach(line -> {
                    Pair<String, String> pair = transform(line);
                    if (pair != null) {
                        pool.bulkIndex(new IndexRequest()
                                .index(elasticsearchIndexDefinition.getFullIndexName())
                                .id(pair.getKey())
                                .source(new BytesArray(pair.getValue()), XContentType.JSON)
                                .create(false));
                    } else {
                        logger.log(Level.WARNING, "id not found in " + line);
                    }
                });
            }
        }

        private static Pair<String, String> transform(String content) {
            Map<String, Object> map = Json.toMap(content);
            String id = Maps.getString(map, "ids.openalex");
            if (id != null && id.startsWith("https://openalex.org/")) {
                id = id.substring(21);
                map.put("id", id);
            }
            if (id == null) {
                return null;
            } else {
                String doi = Maps.getString(map, "ids.doi");
                if (doi != null && doi.startsWith("https://doi.org/")) {
                    doi = doi.substring(16);
                    map.put("doi", doi);
                }
                String pmid = Maps.getString(map, "ids.pmid");
                if (pmid != null && pmid.startsWith("https://pubmed.ncbi.nlm.nih.gov/")) {
                    pmid = pmid.substring(32);
                    map.put("pmid", pmid);
                }
                try {
                    return Pair.of(id, Json.toString(map));
                } catch (IOException e) {
                    // never happens
                    return Pair.of(id, content);
                }
            }
        }
    }
}
