package org.xbib.interlibrary.catalog.elasticsearch.openalex;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Instant;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.CRC32;
import java.util.zip.GZIPInputStream;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.common.bytes.BytesArray;
import org.elasticsearch.common.xcontent.XContentType;
import org.xbib.datastructures.common.Pair;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.interlibrary.api.action.worker.Worker;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.api.util.Maps;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.settings.Settings;
import software.amazon.awssdk.services.s3.S3AsyncClient;
import software.amazon.awssdk.services.s3.model.S3Object;

public class OpenAlexFromAmazonS3 extends ElasticsearchTool<S3BucketObject> {

    @Override
    public ElasticsearchWorkerPool<S3BucketObject> newPool(Settings settings) throws IOException {
        return new OpenAlexPool(settings);
    }

    public class OpenAlexPool extends ElasticsearchWorkerPool<S3BucketObject> {

        private final Logger logger = Logger.getLogger(OpenAlexPool.class.getName());

        private final S3InputHelper s3InputHelper;

        private final S3AsyncClient s3AsyncClient;

        protected OpenAlexPool(Settings settings) throws IOException {
            super("openalex-s3", settings);
            Settings inputSettings = settings.getGroups("input").get("s3");
            boolean isInput = inputSettings != null && !inputSettings.getAsMap().isEmpty();
            s3InputHelper = isInput ? new S3InputHelper(inputSettings) : null;
            s3AsyncClient = s3InputHelper != null ? s3InputHelper.createClient() : null;
        }

        @Override
        public Worker<S3BucketObject> newWorker(String name, int count) {
            return new OpenAlexWorker(settings, this);
        }

        @Override
        protected void openInput() throws IOException, InterruptedException {
            super.openInput();
            if (s3InputHelper == null) {
                logger.log(Level.WARNING, "no S3 input helper, aborting");
                return;
            }
            Settings inputSettings = settings.getGroups("input").get("s3");
            if (inputSettings == null) {
                logger.log(Level.WARNING, "no s3 input definitions?");
                return;
            }
            s3InputHelper.setIndexDefinitions(inputSettings.getGroups("bucket"));
            logger.log(Level.INFO, "s3 input definitions set up: " + s3InputHelper.getIndexDefinitionMap());
            final AtomicBoolean error = new AtomicBoolean(false);
            s3InputHelper.getIndexDefinitionMap().values().forEach(def -> {
                List<S3Object> list = new CopyOnWriteArrayList<>();
                if (!s3InputHelper.list(s3AsyncClient, def, s3Object -> {
                    logger.log(Level.INFO, "found S3 object: key = " + s3Object.key() +
                            " size = " + s3Object.size() +
                            " last modified = " + s3Object.lastModified());
                    list.add(s3Object);
                })) {
                    error.set(true);
                }
                logger.log(Level.INFO, def.getBucket() + ": found " + list.size() + " S3 objects");
                for (S3Object s3Object : list) {
                    S3BucketObject s3BucketObject = new S3BucketObject(s3InputHelper, def.getBucket(), s3Object);
                    submit(newWorkerRequest(s3BucketObject, null));
                }
            });
            if (error.get()) {
                throw new IllegalStateException("there were errors listing S3 objects");
            }
        }

        @Override
        public WorkerRequest<S3BucketObject> newWorkerRequest(Object s3BucketObject, String origin) {
            return s3BucketObject != null ? new S3BucketObjectWorkerRequest((S3BucketObject) s3BucketObject) : null;
        }

        @Override
        protected void closeInput(int returncode) throws IOException {
            try {
                s3AsyncClient.close();
            } finally {
                super.closeInput(returncode);
            }
        }
    }

    public class OpenAlexWorker extends ElasticsearchWorker<S3BucketObject, OpenAlexPool> {

        private final Logger logger = Logger.getLogger(OpenAlexWorker.class.getName());

        private final OpenAlexPool pool;

        private final ElasticsearchIndexDefinition elasticsearchIndexDefinition;

        protected OpenAlexWorker(Settings settings, OpenAlexPool pool) {
            super(settings);
            this.pool = pool;
            this.elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("openalex");
            Objects.requireNonNull(elasticsearchIndexDefinition.getFullIndexName());
        }

        @Override
        protected void doExecute(S3BucketObject s3BucketObject) throws IOException {
            String bucket = s3BucketObject.getBucket();
            String key = s3BucketObject.getS3Object().key();
            // build a temp file name that derives from bucket and key and the current time
            CRC32 crc32 = new CRC32();
            crc32.update((bucket + "_" + key + "_" + Instant.now()).getBytes(StandardCharsets.UTF_8));
            String prefix = Long.toHexString(crc32.getValue());
            String suffix = ".s3object";
            Path path = Files.createTempFile(prefix, suffix);
            logger.log(Level.INFO, "downloading " + key + " prefix = " + prefix + " suffix = " + suffix);
            if (!s3BucketObject.getS3InputHelper().getObject(pool.s3AsyncClient, bucket, key, path)) {
                Files.deleteIfExists(path);
                throw new IOException("there were errors while getting S3 object " + bucket + " " + key);
            }
            logger.log(Level.INFO, "indexing " + bucket + "_" + key + " from " + path);
            try (BufferedReader reader = new BufferedReader(new InputStreamReader(
                    new GZIPInputStream(Files.newInputStream(path))))) {
                reader.lines().forEach(line -> {
                    if (!line.trim().isEmpty()) {
                        Pair<String, BytesArray> pair = transform(line);
                        if (pair != null) {
                            pool.bulkIndex(new IndexRequest()
                                    .index(elasticsearchIndexDefinition.getFullIndexName())
                                    .id(pair.getKey())
                                    .source(pair.getValue(), XContentType.JSON)
                                    .create(false));
                        } else {
                            logger.log(Level.WARNING, "problems with document, skipping: " + line);
                        }
                    }
                });
            } catch (IOException e) {
                logger.log(Level.WARNING, "JSON I/O error " + path + " " + e.getMessage(), e);
            } finally {
                if (Files.deleteIfExists(path)) {
                    logger.log(Level.INFO, "deleted " + path);
                } else {
                    logger.log(Level.WARNING, "file is gone: " + path);
                }
            }
        }

        private static Pair<String, BytesArray> transform(String content) {
            Map<String, Object> map = Json.toMap(content);
            String id = Maps.getString(map, "ids.openalex");
            if (id != null && id.startsWith("https://openalex.org/")) {
                id = id.substring(21);
                map.put("id", id);
            }
            if (id == null) {
                return null;
            } else {
                String doi = Maps.getString(map, "ids.doi");
                if (doi != null && doi.startsWith("https://doi.org/")) {
                    doi = doi.substring(16);
                    map.put("doi", doi);
                }
                String pmid = Maps.getString(map, "ids.pmid");
                if (pmid != null && pmid.startsWith("https://pubmed.ncbi.nlm.nih.gov/")) {
                    pmid = pmid.substring(32);
                    map.put("pmid", pmid);
                }
                // we use Json.toString(), not the Elasticsearch JSON converter, because it can not convert StringBuilder:
                // cannot write xcontent for unknown value of type class java.lang.StringBuilder :
                // java.lang.IllegalArgumentException: cannot write xcontent for unknown value of type class java.lang.StringBuilder
                try {
                    return Pair.of(id, new BytesArray(Json.toString(map)));
                } catch (Exception e) {
                    return null;
                }
            }
        }
    }
}
