package org.xbib.interlibrary.catalog.elasticsearch.openalex;

import java.nio.file.Path;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import org.xbib.settings.Settings;
import software.amazon.awssdk.auth.credentials.AnonymousCredentialsProvider;
import software.amazon.awssdk.http.nio.netty.NettyNioAsyncHttpClient;
import software.amazon.awssdk.regions.Region;
import software.amazon.awssdk.services.s3.S3AsyncClient;
import software.amazon.awssdk.services.s3.model.ChecksumMode;
import software.amazon.awssdk.services.s3.model.GetObjectRequest;
import software.amazon.awssdk.services.s3.model.GetObjectResponse;
import software.amazon.awssdk.services.s3.model.ListObjectsV2Request;
import software.amazon.awssdk.services.s3.model.S3Object;
import software.amazon.awssdk.services.s3.paginators.ListObjectsV2Publisher;

public class S3InputHelper {

    private static final Logger logger = Logger.getLogger(S3InputHelper.class.getName());

    private final Settings settings;

    private Map<String, S3BucketInputDefinition> indexDefinitionMap;

    public S3InputHelper(Settings settings) {
        this.settings = settings;
    }

    public S3AsyncClient createClient() {
        // use long access and read timeouts of 300 seconds (5 minutes) to try to avoid data loss and subsequent re-runs
        Duration connectionAcquisitionTimeout = Duration.ofSeconds(settings.getAsInt("timeout.acquire", 300)); // default 10 seconds
        Duration connectionTimeout = Duration.ofSeconds(settings.getAsInt("timeout.connection", 300)); // default 10 seconds
        Duration readTimeout = Duration.ofSeconds(settings.getAsInt("timeout.read", 300)); // default 10 seconds
        AnonymousCredentialsProvider anonymousCredentialsProvider = AnonymousCredentialsProvider.create();
        return S3AsyncClient.builder()
                .region(Region.US_EAST_1)
                .httpClient(NettyNioAsyncHttpClient.builder()
                        .connectionAcquisitionTimeout(connectionAcquisitionTimeout)
                        .connectionTimeout(connectionTimeout)
                        .readTimeout(readTimeout)
                        .build())
                .credentialsProvider(anonymousCredentialsProvider)
                .build();
    }

    public void setIndexDefinitions(Map<String, Settings> indexDefinitions) {
        this.indexDefinitionMap = new LinkedHashMap<>();
        for (Map.Entry<String, Settings> entry : indexDefinitions.entrySet()) {
            Settings settings = entry.getValue();
            String bucket = settings.get("name");
            Pattern pattern = Pattern.compile(settings.get("pattern", ".*"));
            Instant lastmodified = settings.get("lastmodified") != null ?
                    LocalDate.parse(settings.get("lastmodified")).atStartOfDay(ZoneId.systemDefault()).toInstant() :
                    Instant.EPOCH;
            S3BucketInputDefinition def = new S3BucketInputDefinition(bucket, pattern, lastmodified);
            indexDefinitionMap.put(entry.getKey(), def);
        }
    }

    public Map<String, S3BucketInputDefinition> getIndexDefinitionMap() {
        return indexDefinitionMap;
    }

    public boolean list(S3AsyncClient s3AsyncClient,
                     S3BucketInputDefinition s3BucketInputDefinition,
                     Consumer<S3Object> consumer) {
        try {
            ListObjectsV2Request listObjectsV2Request = ListObjectsV2Request.builder()
                    .bucket(s3BucketInputDefinition.getBucket())
                    .build();
            ListObjectsV2Publisher listObjectsV2Publisher =
                    s3AsyncClient.listObjectsV2Paginator(listObjectsV2Request);
            listObjectsV2Publisher.contents().subscribe(s3Object -> {
                if (s3BucketInputDefinition.getPattern() == null || s3BucketInputDefinition.getPattern().matcher(s3Object.key()).matches()) {
                    consumer.accept(s3Object);
                }
            }).get(1L, TimeUnit.MINUTES);
            return true;
        } catch (Exception e) {
            logger.log(Level.SEVERE, "listing S3 objects failed");
            return false;
        }
    }

    public boolean getObject(S3AsyncClient s3AsyncClient,
                          String bucket,
                          String key,
                          Path path) {
        try {
            GetObjectRequest objectRequest = GetObjectRequest.builder()
                    .bucket(bucket)
                    .key(key)
                    .checksumMode(ChecksumMode.ENABLED)
                    .build();
            GetObjectResponse getObjectResponse = s3AsyncClient
                    .getObject(objectRequest, path)
                    .get(1L, TimeUnit.HOURS);
            logger.log(Level.INFO, "got response, length " + getObjectResponse.contentLength() +
                    " checksum type = " + getObjectResponse.checksumTypeAsString());
            return true;
        } catch (Exception e) {
            logger.log(Level.SEVERE, "error when getting S3 object: " + e.getMessage(), e);
            return false;
        }
    }
}
