package org.xbib.interlibrary.catalog.elasticsearch.pubmed;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfContentParams;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RdfXContentParams;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.internal.DefaultAnonymousResource;
import org.xbib.content.rdf.io.xml.XmlContentParser;
import org.xbib.content.rdf.io.xml.XmlHandler;
import org.xbib.content.resource.IRI;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.content.resource.NamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.settings.Settings;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Load Medical Subject heading files and index them into Elasticsearch.
 */
public class Mesh extends ElasticsearchTool<String> {

    public Mesh() {
    }

    @Override
    public MeshWorkerPool newPool(Settings settings) throws IOException {
        return new MeshWorkerPool(settings);
    }

    public class MeshWorkerPool extends ElasticsearchWorkerPool<String> {

        public MeshWorkerPool(Settings settings) throws IOException {
            super("mesh", settings);
        }

        @Override
        public MeshWorker newWorker(String name, int count) {
            return new MeshWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }

    public class MeshWorker extends ElasticsearchWorker<String, MeshWorkerPool> {

        private final Logger logger = Logger.getLogger(MeshWorker.class.getName());

        private final MeshWorkerPool pool;

        public MeshWorker(Settings settings, MeshWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings)) {
                RdfContentParams params = new RdfXContentParams(pool.getNamespaceContext());
                Handler handler = new Handler(params)
                        .setDefaultNamespace("", "http://xbib.org/ns/mesh/");
                InvalidXmlCharacterFilterReader reader = new InvalidXmlCharacterFilterReader(inputStream, StandardCharsets.UTF_8);
                XmlContentParser<RdfContentParams> parser = new XmlContentParser<>(params, reader);
                parser.setNamespaces(false).setHandler(handler);
                parser.parse();
            }
        }

        private class Handler extends DefaultHandler implements XmlHandler<RdfContentParams> {

            private final RdfContentParams params;

            private final StringBuilder content = new StringBuilder();

            private final Deque<QName> parents = new ArrayDeque<>();

            private Resource resource;

            private String defaultPrefix;

            private String defaultNamespace;

            private String descriptor;

            private final List<String> variants = new LinkedList<>();

            private Handler(RdfContentParams params) {
                this.params = params;
                this.resource = new DefaultAnonymousResource();
            }

            public RdfContentParams getParams() {
                return params;
            }

            public Resource getResource() {
                return resource;
            }

            @Override
            public XmlHandler<RdfContentParams> setNamespaceContext(NamespaceContext namespaceContext) {
                return this;
            }

            @Override
            public IRINamespaceContext getNamespaceContext() {
                return params.getNamespaceContext();
            }

            @Override
            public Handler setDefaultNamespace(String prefix, String namespaceURI) {
                this.defaultPrefix = prefix;
                this.defaultNamespace = namespaceURI;
                params.getNamespaceContext().addNamespace(prefix, namespaceURI);
                return this;
            }

            @Override
            public XmlHandler<RdfContentParams> setBuilder(RdfContentBuilder<RdfContentParams> builder) {
                // ignore
                return this;
            }

            @Override
            public void startDocument() throws SAXException {
                try {
                    openResource();
                } catch (IOException e) {
                    throw new SAXException(e);
                }
                parents.push(new QName("_"));
            }

            @Override
            public void endDocument() throws SAXException {
                try {
                    closeResource();
                } catch (IOException e) {
                    throw new SAXException(e);
                }
            }

            @Override
            public void startElement(String nsURI, String localname, String qname, Attributes atts) throws SAXException {
                try {
                    QName name = makeQName(nsURI, localname, qname);
                    boolean delimiter = isResourceDelimiter(name);
                    if (delimiter) {
                        closeResource();
                        openResource();
                    }
                    if (skip(name)) {
                        return;
                    }
                    parents.push(name);
                } catch (Exception e) {
                    throw new SAXException(e);
                }
            }

            @Override
            public void endElement(String nsURI, String localname, String qname) throws SAXException {
                QName name = makeQName(nsURI, localname, qname);
                if (skip(name)) {
                    content.setLength(0);
                    return;
                }
                parents.pop();
                if ("DescriptorUI".equals(name.getLocalPart())) {
                    resource.setId(IRI.create(content()));
                }
                if (!isResourceDelimiter(name) && !parents.isEmpty() &&
                        "String".equals(name.getLocalPart()) &&
                        "DescriptorName".equals(parents.peek().getLocalPart())) {
                    String content = content();
                    int pos = content.indexOf('[');
                    if (pos > 0) {
                        descriptor = content.substring(pos + 1, content.length() - 1); // without [ and ]
                        variants.add(content.substring(0, pos)); // german form
                    }
                }
                content.setLength(0);
            }

            @Override
            public void characters(char[] chars, int start, int length) throws SAXException {
                content.append(new String(chars, start, length));
            }

            @Override
            public void startPrefixMapping(String prefix, String uri) throws SAXException {
                // nothing to do here
            }

            @Override
            public void endPrefixMapping(String prefix) throws SAXException {
                // nothing to do here
            }

            private QName makeQName(String nsURI, String localname, String qname) {
                String prefix = params.getNamespaceContext().getPrefix(nsURI);
                return new QName(isNotEmpty(nsURI) ? nsURI : defaultNamespace,
                        isNotEmpty(localname) ? localname : qname,
                        isNotEmpty(prefix) ? prefix : defaultPrefix);
            }

            private String content() {
                String s = content.toString().trim();
                return s.length() > 0 ? s : null;
            }

            private void openResource() throws IOException {
                resource = new DefaultAnonymousResource();
            }

            private void closeResource() throws IOException {
                if (descriptor != null) {
                    resource.add("descriptor", descriptor);
                    variants.forEach(s -> resource.add("variant", s));
                    descriptor = null;
                    variants.clear();
                }
                boolean empty = resource.isEmpty();
                if (empty) {
                    logger.log(Level.WARNING, "resource is empty");
                    return;
                }
                ElasticsearchIndexDefinition indexDefinition = pool.getOutputIndexDefinitionMap().get("mesh");
                RouteRdfXContentParams params = new RouteRdfXContentParams(getNamespaceContext(),
                        indexDefinition.getFullIndexName(), indexDefinition.getType());
                params.setHandler((content, p) -> pool.bulkIndex(indexDefinition,
                        resource.id().toString(), false, content));
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                    builder.receive(resource);
                }
                if (settings.getAsBoolean("mock", false)) {
                    logger.log(Level.INFO, resource.id() + " " + params.getGenerator().get());
                }
            }

            private boolean isResourceDelimiter(QName name) {
                return "DescriptorRecord".equals(name.getLocalPart());
            }

            private boolean skip(QName name) {
                boolean isAttr = name.getLocalPart().startsWith("@");
                return "DescriptorRecordSet".equals(name.getLocalPart())
                        || "DescriptorRecord".equals(name.getLocalPart())
                        || isAttr;
            }

            private boolean isNotEmpty(String s) {
                return s != null && s.length() != 0;
            }
        }
    }
}
