package org.xbib.interlibrary.catalog.elasticsearch.pubmed;

import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.xbib.content.core.XContentHelper;
import org.xbib.content.rdf.RdfConstants;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfContentParams;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RdfXContentParams;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.io.xml.AbstractXmlHandler;
import org.xbib.content.rdf.io.xml.AbstractXmlResourceHandler;
import org.xbib.content.rdf.io.xml.XmlContentParser;
import org.xbib.content.rdf.io.xml.XmlHandler;
import org.xbib.content.resource.IRI;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.content.resource.NamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.matching.AuthoredWorkKey;
import org.xbib.interlibrary.catalog.matching.EncoderException;
import org.xbib.settings.Settings;

/**
 * Elasticsearch indexer tool for Pubmed XML files.
 */
public class Pubmed extends ElasticsearchTool<String> {

    public Pubmed() {
    }

    @Override
    public PubmedWorkerPool newPool(Settings settings) throws IOException {
        return new PubmedWorkerPool(settings);
    }

    public class PubmedWorkerPool extends ElasticsearchWorkerPool<String> {

        private final IRINamespaceContext namespaceContext;

        public PubmedWorkerPool(Settings settings) throws IOException {
            super("pubmed", settings);
            this.namespaceContext = createNamespaceContext();
        }

        @Override
        public PubmedWorker newWorker(String name, int count) {
            return new PubmedWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        @Override
        public IRINamespaceContext getNamespaceContext() {
            return namespaceContext;
        }

        private static IRINamespaceContext createNamespaceContext() {
            IRINamespaceContext namespaceContext = IRINamespaceContext.newInstance();
            namespaceContext.addNamespace(RdfConstants.NS_PREFIX, RdfConstants.NS_URI);
            namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
            namespaceContext.addNamespace("dcterms", "http://purl.org/dc/terms/");
            namespaceContext.addNamespace("foaf", "http://xmlns.com/foaf/0.1/");
            namespaceContext.addNamespace("frbr", "http://purl.org/vocab/frbr/core#");
            namespaceContext.addNamespace("fabio", "http://purl.org/spar/fabio/");
            namespaceContext.addNamespace("prism", "http://prismstandard.org/namespaces/basic/3.0/");
            return namespaceContext;
        }
    }

    public class PubmedWorker extends ElasticsearchWorker<String, PubmedWorkerPool> {

        private final Logger logger = Logger.getLogger(PubmedWorker.class.getName());

        private final PubmedWorkerPool pool;

        public PubmedWorker(Settings settings, PubmedWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings)) {
                RdfContentParams params = new RdfXContentParams(pool.getNamespaceContext());
                AbstractXmlHandler<RdfContentParams> handler = new Handler(params)
                        .setDefaultNamespace("ml", "http://www.nlm.nih.gov/medline");
                XmlContentParser<RdfContentParams> parser = new XmlContentParser<>(params, inputStream);
                parser.setNamespaces(false).setHandler(handler).parse();
            }
        }

        private class Handler extends AbstractXmlResourceHandler<RdfContentParams> {

            private final List<String> authors;

            private String id;

            private String work;

            private String forename;

            private String lastname;

            private String date;

            private String volume;

            private String issue;

            Handler(RdfContentParams params) {
                super(params);
                this.authors = new LinkedList<>();
            }

            @Override
            public void closeResource() throws IOException {
                super.closeResource();
                // create bibliographic key
                // there are works with "no authors listed" (e.g. PMID 5236443)
                if (work != null) {
                    AuthoredWorkKey authoredWorkKey = new AuthoredWorkKey()
                            .workName(work)
                            .authorName(authors)
                            .chronology(date)
                            .chronology(volume)
                            .chronology(issue);
                    try {
                        getResource().add("xbib:key", authoredWorkKey.createKey());
                    } catch (EncoderException e) {
                        throw new IOException(e);
                    }
                }
                ElasticsearchIndexDefinition elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
                RouteRdfXContentParams params = new RouteRdfXContentParams(getNamespaceContext(),
                        elasticsearchIndexDefinition.getFullIndexName(), elasticsearchIndexDefinition.getType());
                params.setHandler((content, p) ->
                        pool.bulkIndex(elasticsearchIndexDefinition, id, false, convert2fabio(content)));
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                    builder.receive(getResource());
                }
                id = null;
                authors.clear();
                work = null;
                date = null;
                volume = null;
                issue = null;
            }

            private String convert2fabio(String content) {
                Map<String, ?> map = XContentHelper.convertToMap(content);
                PubmedMapper mf = new PubmedMapper();
                RouteRdfXContentParams params = new RouteRdfXContentParams(pool.getNamespaceContext(), "index", "type");
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                    builder.receive(mf.map(map));
                    return params.getGenerator().get();
                } catch (IOException e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                }
                return null;
            }

            @Override
            public boolean isResourceDelimiter(QName name) {
                return "MedlineCitation".equals(name.getLocalPart());
            }

            @Override
            public void identify(QName name, String value, IRI identifier) {
                // important: there are many occurances of PMID.
                // We must only take the first occurance for the ID.
                LinkedList<QName> parents = getParents();
                if (id == null && "PMID".equals(name.getLocalPart())) {
                    this.id = value;
                } else if ("ArticleTitle".equals(name.getLocalPart())) {
                    this.work = value;
                } else if ("LastName".equals(name.getLocalPart())) {
                    this.lastname = value;
                } else if ("ForeName".equals(name.getLocalPart())) {
                    this.forename = value;
                    if (forename != null && lastname != null) {
                        authors.add(lastname + ", " + forename);
                        forename = null;
                        lastname = null;
                    }
                } else if ("Year".equals(name.getLocalPart())
                        && ("PubDate".equals(parents.peek().getLocalPart()) ||
                        "DateCreated".equals(parents.peek().getLocalPart()))) {
                    date = value;
                } else if ("Volume".equals(name.getLocalPart())) {
                    volume = value;
                } else if ("Issue".equals(name.getLocalPart())) {
                    // issue is needed, see e.g. PMID 5015805
                    issue = value;
                }
            }

            @Override
            public boolean skip(QName name) {
                return "MedlineCitationSet".equals(name.getLocalPart())
                        || "MedlineCitation".equals(name.getLocalPart())
                        || "PubmedArticleSet".equals(name.getLocalPart())
                        || "PubmedArticle".equals(name.getLocalPart())
                        || name.getLocalPart().startsWith("@");
            }

            @Override
            public XmlHandler<RdfContentParams> setNamespaceContext(NamespaceContext namespaceContext) {
                return this;
            }

            @Override
            public IRINamespaceContext getNamespaceContext() {
                return getParams().getNamespaceContext();
            }
        }
    }
}
