package org.xbib.interlibrary.catalog.elasticsearch.pubmed;

import java.io.IOException;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.xbib.content.rdf.Literal;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.internal.DefaultAnonymousResource;
import org.xbib.content.rdf.internal.DefaultLiteral;
import org.xbib.content.resource.IRI;
import org.xbib.interlibrary.catalog.entities.ArticleVocabulary;
import org.xbib.interlibrary.catalog.matching.AuthoredWorkKey;
import org.xbib.interlibrary.catalog.matching.EncoderException;

public class PubmedMapper implements ArticleVocabulary {

    private final Set<Author> authors = new LinkedHashSet<>();

    private Author author;

    private Datestamp pubdate;

    private Datestamp articledate;

    private Datestamp creationdate;

    private String medlinedate;

    private String title;

    private String journal;

    private String issn;

    private String shortTitle;

    PubmedMapper() {
    }

    public Resource map(Map<String, ?> map) throws IOException {
        Resource r = new DefaultAnonymousResource().a(FABIO_ARTICLE);
        author = new Author();
        map(r, null, map);
        String year = null;
        if (pubdate == null) {
            pubdate = articledate;
        }
        if (pubdate == null) {
            pubdate = creationdate;
        }
        if (pubdate != null) {
            if (pubdate.year.length() > 3) {
                // may throw NumberFormatException if pubdate starts with "Spring" or "Summer" etc.
                year = pubdate.year.substring(0, 4);
                try {
                    r.add(DC_DATE, new DefaultLiteral(year).type(Literal.GYEAR));
                } catch (NumberFormatException e) {
                    // skip
                }
            }
            StringBuilder sb = new StringBuilder();
            sb.append(pubdate.year);
            if (pubdate.month != null) {
                sb.append('-').append(pubdate.month);
            }
            if (pubdate.day != null) {
                sb.append('-').append(pubdate.day);
            }
            r.add(PRISM_PUBLICATIONDATE, sb.toString());
        } else if (medlinedate != null) {
            if (medlinedate.length() > 3) {
                year = medlinedate.substring(0, 4);
                // may throw NumberFormatException if pubdate starts with "Spring" or "Summer" etc.
                try {
                    r.add(DC_DATE, new DefaultLiteral(year).type(Literal.GYEAR));
                } catch (NumberFormatException e) {
                    // skip
                }
            }
            r.add(PRISM_PUBLICATIONDATE, medlinedate);
        }
        if (author != null && author.lastName != null) {
            authors.add(author);
            r.newResource(DC_CREATOR)
                    .a(FOAF_AGENT)
                    .add(FOAF_FAMILYNAME, author.lastName)
                    .add(FOAF_GIVENNAME, author.foreName);
        }
        r.newResource(FRBR_PARTOF)
                .a(FABIO_JOURNAL)
                .add(PRISM_PUBLICATION_NAME, journal)
                .add(PRISM_ISSN, issn)
                .add(FABIO_HAS_SHORT_TITLE, shortTitle);
        AuthoredWorkKey wa = new AuthoredWorkKey()
                .workName(title)
                .chronology(year);
        for (Author author : authors) {
            wa.authorNameWithForeNames(author.lastName, author.foreName);
        }
        try {
            r.add(XBIB_KEY, wa.createKey());
        } catch (EncoderException e) {
            throw new IOException(e);
        }
        return r;
    }

    @SuppressWarnings("unchecked")
    private void map(Resource r, String prefix, Map<String, ?> map) throws IOException {
        for (Map.Entry<String, ?> entry : map.entrySet()) {
            String key = entry.getKey();
            String p = prefix != null ? prefix + "." + key : key;
            Object value = entry.getValue();
            if (value instanceof Map) {
                map(r, p, (Map<String, Object>) value);
            } else if (value instanceof List) {
                for (Object o : (List) value) {
                    if (o instanceof Map) {
                        map(r, p, (Map<String, Object>) o);
                    } else {
                        map(r, p, o.toString());
                    }
                }
            } else {
                if (value != null) {
                    map(r, p, value.toString());
                }
            }
        }
    }

    public void map(Resource r, String key, String valueString) throws IOException {
        String value = valueString;
        switch (key) {
            case "ml:PMID":
                IRI dereferencable = IRI.builder().scheme("http").host("xbib.info")
                        .path("/pmid/").fragment(value).build();
                r.setId(dereferencable);
                r.add(FABIO_HAS_PUBMEDID, value);
                r.add(DC_IDENTIFIER, "PMID:" + value);
                break;
            case "ml:OtherID":
                r.add(DC_IDENTIFIER, value);
                if (value.startsWith("PMC")) {
                    r.add(FABIO_HAS_PUBMEDCENTRALID, value.substring(3));
                }
                break;
            case "ml:Article.ml:ArticleTitle":
                title = value;
                if (title.endsWith(".")) {
                    title = title.substring(0, value.length() - 1);
                }
                r.add(DC_TITLE, title);
                break;
            case "ml:Article.ml:Journal.ml:ISSN":
                issn = value;
                break;
            case "ml:Article.ml:Journal.ml:Title":
                journal = value;
                break;
            case "ml:Article.ml:Journal.ml:ISOAbbreviation":
                shortTitle = value;
                break;
            case "ml:Article.ml:AuthorList.ml:Author.ml:ForeName":
                if (author == null) {
                    author = new Author();
                }
                author.setForeName(value);
                break;
            case "ml:Article.ml:AuthorList.ml:Author.ml:LastName":
                if (author == null) {
                    author = new Author();
                } else if (author.lastName != null) {
                    authors.add(author);
                    r.newResource(DC_CREATOR)
                            .a(FOAF_AGENT)
                            .add(FOAF_FAMILYNAME, author.lastName)
                            .add(FOAF_GIVENNAME, author.foreName);
                    author = new Author();
                }
                author.setLastName(value);
                break;
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:PubDate.ml:MedlineDate":
                medlinedate = value;
                break;
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:PubDate.ml:Year":
                if (pubdate == null) {
                    pubdate = new Datestamp();
                }
                pubdate.year = value;
                break;
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:PubDate.ml:Month":
                if (pubdate == null) {
                    pubdate = new Datestamp();
                }
                pubdate.month = value;
                break;
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:PubDate.ml:Day":
                if (pubdate == null) {
                    pubdate = new Datestamp();
                }
                pubdate.day = value;
                break;
            case "ml:Article.ml:ArticleDate.ml:Year":
                if (articledate == null) {
                    articledate = new Datestamp();
                }
                articledate.year = value;
                break;
            case "ml:Article.ml:ArticleDate.ml:Month":
                if (articledate == null) {
                    articledate = new Datestamp();
                }
                articledate.month = value;
                break;
            case "ml:Article.ml:ArticleDate.ml:Day":
                if (articledate == null) {
                    articledate = new Datestamp();
                }
                articledate.day = value;
                break;
            case "ml:DateCreated.ml:Year":
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:DateCreated.ml:Year":
                if (creationdate == null) {
                    creationdate = new Datestamp();
                }
                creationdate.year = value;
                break;
            case "ml:DateCreated.ml:Month":
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:DateCreated.ml:Month":
                if (creationdate == null) {
                    creationdate = new Datestamp();
                }
                creationdate.month = value;
                break;
            case "ml:DateCreated.ml:Day":
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:DateCreated.ml:Day":
                if (creationdate == null) {
                    creationdate = new Datestamp();
                }
                creationdate.day = value;
                break;
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:Issue":
                r.newResource(FRBR_EMBODIMENT)
                        .a(FABIO_PERIODICAL_ISSUE)
                        .add(PRISM_NUMBER, value);
                break;
            case "ml:Article.ml:Journal.ml:JournalIssue.ml:Volume":
                r.newResource(FRBR_EMBODIMENT)
                        .a(FABIO_PERIODICAL_VOLUME)
                        .add(PRISM_VOLUME, value);
                break;
            case "ml:Article.ml:Pagination.ml:MedlinePgn":
                int pos = value.indexOf('-');
                if (pos > 0) {
                    String spage = value.substring(0, pos);
                    String epage = value.substring(pos + 1);
                    int l = spage.length() - epage.length();
                    if (l > 0) {
                        epage = spage.substring(0, l) + epage;
                    }
                    r.newResource(FRBR_EMBODIMENT)
                            .a(FABIO_PRINT_OBJECT)
                            .add(PRISM_STARTING_PAGE, spage)
                            .add(PRISM_ENDING_PAGE, epage);
                } else {
                    r.newResource(FRBR_EMBODIMENT)
                            .a(FABIO_PRINT_OBJECT)
                            .add(PRISM_PAGERANGE, value);
                }
                break;
            case "ml:Article.ml:ELocationID":
                r.add(PRISM_DOI, value);
                break;
            case "ml:Article.ml:Abstract.ml:AbstractText":
                r.newResource(FRBR_PART)
                        .a(FABIO_ABSTRACT)
                        .add(DCTERMS_ABSTRACT, value);
                break;
            case "ml:MeshHeadingList.ml:MeshHeading.ml:DescriptorName":
                r.newResource(FABIO_HAS_SUBJECT_TERM)
                        .a(FABIO_SUBJECT_TERM)
                        .add(DC_SOURCE, "MeSH")
                        .add(RDFS_LABEL, value);
                break;
            case "ml:CitationSubset":
                switch (value) {
                    case "AIM":
                        value = "Abridged Index Medicus";
                        break;
                    case "B":
                        value = "Biotechnology";
                        break;
                    case "C":
                        value = "Communication Disorders";
                        break;
                    case "D":
                        value = "Dental journals";
                        break;
                    case "E":
                        value = "Ethics journals";
                        break;
                    case "F":
                        value = "Older journals";
                        break;
                    case "H":
                        value = "Health administration journals";
                        break;
                    case "IM":
                        value = "Index Medicus";
                        break;
                    case "J":
                        value = "Population information journals";
                        break;
                    case "K":
                        value = "Consumer health journals";
                        break;
                    case "N":
                        value = "Nursing journals";
                        break;
                    case "OM":
                        value = "Old Medline";
                        break;
                    case "Q":
                        value = "History of medicine journals";
                        break;
                    case "QIS":
                        value = "non-Index Medicus History of medicine journals";
                        break;
                    case "QO":
                        value = "older History of medicine journals";
                        break;
                    case "R":
                        value = "Population and reproduction journals";
                        break;
                    case "S":
                        value = "Space life sciences journals";
                        break;
                    case "T":
                        value = "Health technology assessment journals";
                        break;
                    case "X":
                        value = "AIDS/HIV journals";
                        break;
                    default:
                        break;
                }
                r.add(DC_SOURCE, value);
                break;
            default:
                break;
        }
    }

    static class Author implements Comparable<Author> {
        private String lastName;
        private String foreName;
        private String normalized;

        void setLastName(String lastName) {
            this.lastName = lastName;
            this.normalized = normalize();
        }

        void setForeName(String foreName) {
            this.foreName = foreName;
            this.normalized = normalize();
        }

        private String normalize() {
            StringBuilder sb = new StringBuilder();
            if (lastName != null) {
                sb.append(lastName);
            }
            if (foreName != null) {
                sb.append(' ').append(foreName);
            }
            return sb.toString();
        }

        @Override
        public int compareTo(Author o) {
            return normalized.compareTo(o.normalized);
        }

        @Override
        public boolean equals(Object o) {
            return o instanceof Author && normalized.equals(((Author) o).normalized);
        }

        @Override
        public int hashCode() {
            return normalized.hashCode();
        }
    }

    private static class Datestamp {
        String year;
        String month;
        String day;
    }
}
