package org.xbib.interlibrary.catalog.elasticsearch.springer;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.Literal;
import org.xbib.content.rdf.RdfConstants;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.internal.DefaultLiteral;
import org.xbib.content.rdf.internal.DefaultResource;
import org.xbib.content.resource.IRI;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchTool;
import org.xbib.interlibrary.catalog.elasticsearch.ElasticsearchIndexDefinition;
import org.xbib.interlibrary.catalog.entities.ArticleVocabulary;
import org.xbib.interlibrary.catalog.matching.AuthoredWorkKey;
import org.xbib.interlibrary.catalog.matching.EncoderException;
import org.xbib.settings.Settings;

/**
 * Push Springer citations to Elasticsearch.
 * Example:
 * <a href="http://link.springer.com/export-citation/article/10.1007%2FBF01242025.txt">...</a>
 */
public class SpringerCitations extends ElasticsearchTool<String> {

    public SpringerCitations() {
    }

    @Override
    public SpringerCitationsWorkerPool newPool(Settings settings) throws IOException {
        return new SpringerCitationsWorkerPool(settings);
    }

    public class SpringerCitationsWorkerPool extends ElasticsearchWorkerPool<String> {

        public SpringerCitationsWorkerPool(Settings settings) throws IOException {
            super("springer", settings);
        }

        @Override
        public SpringerCitationsWorker newWorker(String name, int count) throws IOException {
            return new SpringerCitationsWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }

    public class SpringerCitationsWorker extends ElasticsearchWorker<String, SpringerCitationsWorkerPool> implements ArticleVocabulary {

        private final Logger logger = Logger.getLogger(SpringerCitationsWorker.class.getName());

        private final IRINamespaceContext namespaceContext;

        private final SpringerCitationsWorkerPool pool;

        public SpringerCitationsWorker(Settings settings, SpringerCitationsWorkerPool pool) {
            super(settings);
            this.pool = pool;
            this.namespaceContext = createNamespaceContext();
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            try (InputStream in = ConverterInput.getInputStream(uri, settings)) {
                BufferedReader reader = new BufferedReader(new InputStreamReader(in, StandardCharsets.UTF_8));
                process(uri, reader);
            }
        }

        private void process(String uri, BufferedReader reader) throws IOException {
            String title = null;
            List<String> author = new LinkedList<>();
            String year = null;
            String journal = null;
            String issn = null;
            String volume = null;
            String issue = null;
            String pagination = null;
            String doi = null;
            String publisher = null;
            String line;
            while ((line = reader.readLine()) != null) {
                if (line.isEmpty()) {
                    continue;
                }
                if ('%' != line.charAt(0)) {
                    continue;
                }
                char ch = line.charAt(1);
                switch (ch) {
                    case 'D' -> year = line.substring(3).trim();
                    case 'T' -> title = line.substring(3).trim();
                    case '@' -> issn = line.substring(3).trim();
                    case 'J' -> journal = line.substring(3).trim();
                    case 'A' -> author.add(line.substring(3).trim());
                    case 'V' -> volume = line.substring(3).trim();
                    case 'N' -> issue = line.substring(3).trim();
                    case 'P' -> pagination = line.substring(3).trim();
                    case 'R' -> doi = line.substring(3).trim();
                    case 'I' -> publisher = line.substring(3).trim();

                    // URL (DOI resolver)
                    // keywords
                    // record type
                    // day
                    // language
                    default -> {
                    }
                }
            }
            if (doi == null) {
                logger.log(Level.WARNING, "does not contain DOI: {}", uri);
            } else {
                doi = doi.toLowerCase();
            }
            IRI dereferencable = IRI.builder().scheme("http").host("xbib.info")
                    .path("/doi/").fragment(doi).build();
            Resource r = new DefaultResource(dereferencable)
                    .a(FABIO_ARTICLE)
                    .add(PRISM_DOI, doi)
                    .add(DC_TITLE, title);
            if (title != null) {
                String s = author.isEmpty() ? null : author.get(0);
                AuthoredWorkKey authoredWorkKey = new AuthoredWorkKey()
                        .authorName(s)
                        .workName(title)
                        .chronology(year)
                        .chronology(volume)
                        .chronology(issue);
                String key = null;
                try {
                    key = authoredWorkKey.createKey();
                } catch (EncoderException e) {
                    throw new IOException(e);
                }
                r.add(XBIB_KEY, key);
            }
            for (String a : author) {
                r.newResource(DC_CREATOR)
                        .a(FOAF_AGENT)
                        .add(FOAF_NAME, a);

            }
            r.add(PRISM_PUBLICATIONDATE, new DefaultLiteral(year).type(Literal.GYEAR));
            r.newResource(FRBR_EMBODIMENT)
                    .a(FABIO_PERIODICAL_VOLUME)
                    .add(PRISM_VOLUME, volume);
            r.newResource(FRBR_EMBODIMENT)
                    .a(FABIO_PERIODICAL_ISSUE)
                    .add(PRISM_NUMBER, issue);
            r.newResource(FRBR_EMBODIMENT)
                    .a(FABIO_PRINT_OBJECT)
                    .add(PRISM_PAGERANGE, pagination);
            r.newResource(FRBR_PARTOF)
                    .a(FABIO_JOURNAL)
                    .add(PRISM_PUBLICATION_NAME, journal)
                    .add(PRISM_ISSN, issn)
                    .add(DC_PUBLISHER, publisher);

            ElasticsearchIndexDefinition elasticsearchIndexDefinition = pool.getOutputIndexDefinitionMap().get("bib");
            RouteRdfXContentParams params = new RouteRdfXContentParams(namespaceContext,
                    elasticsearchIndexDefinition.getFullIndexName(), "type");
            params.setHandler((content, p) -> {
                if (!elasticsearchIndexDefinition.isEnabled()) {
                    logger.log(Level.INFO, "{}", content);
                } else {
                    pool.bulkIndex(elasticsearchIndexDefinition, r.id().toString(), false, content);
                }
            });
            try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                builder.receive(r);
            }
        }
    }

    private static IRINamespaceContext createNamespaceContext() {
        IRINamespaceContext namespaceContext = IRINamespaceContext.newInstance();
        namespaceContext.addNamespace(RdfConstants.NS_PREFIX, RdfConstants.NS_URI);
        namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
        namespaceContext.addNamespace("dcterms", "http://purl.org/dc/terms/");
        namespaceContext.addNamespace("foaf", "http://xmlns.com/foaf/0.1/");
        namespaceContext.addNamespace("frbr", "http://purl.org/vocab/frbr/core#");
        namespaceContext.addNamespace("fabio", "http://purl.org/spar/fabio/");
        namespaceContext.addNamespace("prism", "http://prismstandard.org/namespaces/basic/3.0/");
        return namespaceContext;
    }
}
